/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* backtrellis.c --- keep trellis at 1st pass and provide access to 2nd pass */

#include <julius.h>

#define WORD_PER_FRAME_STEP 200	/* step to increase buffer */

/* $B%H%l%j%99=B$BN$r=i4|2=!J5/F0;~$N$_!K(B */
void
bt_init(BACKTRELLIS *bt)
{
  bt->num  = NULL;
  bt->rw   = NULL;
  bt->list = NULL;
}

/* $B%H%l%j%93JG<=`Hw(B ($B2;@<(B1$BF~NO$4$H(B) */
void
bt_prepare(BACKTRELLIS *bt)
{
  TRELLIS_ATOM *tre, *tmp;
  int t;

  /* $BA02s$N$,;D$C$F$$$l$P2rJ|(B */
  if (bt->num != NULL) {
    for(t=0;t<bt->framelen;t++) {
      if (bt->num[t] > 0) {
	free(bt->rw[t]);
      }
    }
    free(bt->rw);
    free(bt->num);
  }
  bt->num = NULL;
  bt->rw = NULL;
  tre = bt->list;
  while (tre != NULL) {
    tmp = tre->next;
    free(tre);
    tre = tmp;
  }

  /* $B%(%s%H%j!<%]%$%s%H$r=i4|@_Dj(B */
  bt->list = NULL;
}  


/* $B%H%l%j%9>pJs$r3JG<$9$k(B */
/* $BBh(B1$B%Q%9Cf$O%j%9%H7A<0$G3JG<$9$k(B */
void
bt_store(BACKTRELLIS *bt, TRELLIS_ATOM *tatom)
{
#ifdef WORD_GRAPH
  tatom->within_wordgraph = FALSE;
#endif
  tatom->next = bt->list;
  bt->list = tatom;
  /*j_printf("  %3d: %15s %f < %3d %15s\n", tatom->endtime,
	 winfo.woutput[tatom->wid], tatom->backscore,
	 tatom->begintime, winfo.woutput[tatom->last_tre->wid]);*/
}

/* $BBh(B1$B%Q%9=*N;8e!$%U%l!<%`C10L$N%"%/%;%9$KE,$7$?9=B$$XJQ7A$9$k(B */
void
bt_relocate_rw(BACKTRELLIS *bt)
{
  TRELLIS_ATOM *tre;
  int t;

  /* (framelen)$BD9$NNN0h$r3NJ](B */
  bt->num = (int *)mymalloc(sizeof(int) * bt->framelen);
  bt->rw  = (TRELLIS_ATOM ***)mymalloc(sizeof(TRELLIS_ATOM **) * bt->framelen);

  /* $B3F%U%l!<%`$K$*$1$k(Brw$B$N?t$r%+%&%s%H(B */
  for (t=0;t<bt->framelen;t++) bt->num[t] = 0;
  for (tre=bt->list;tre;tre=tre->next) {
    bt->num[tre->endtime]++;
  }
  /* $B3F%U%l!<%`$4$H$N3JG<NN0h$r3NJ](B */
  for (t=0;t<bt->framelen;t++) {
    if (bt->num[t] > 0) {
      bt->rw[t] = (TRELLIS_ATOM **)mymalloc(sizeof(TRELLIS_ATOM *) * bt->num[t]);
    }
  }
  /* $B3F%U%l!<%`Kh$K(B TRELLIS_ATOM $B$X$N%]%$%s%?$r3JG<(B */
  for (t=0;t<bt->framelen;t++) bt->num[t] = 0;
  for (tre=bt->list;tre;tre=tre->next) {
    t = tre->endtime;
    bt->rw[t][bt->num[t]] = tre;
    bt->num[t]++;
  }
}


/* $B0J9_$N4X?t$O%U%l!<%`C10L$G%"%/%;%9$9$k$?$a!$(B
   bt_relocate_rw $B8e$G$"$k$3$H$,A0Ds$H$J$k(B */

/* $BBh(B2$B%Q%9$N$?$a$K!$:G8e$N2;6A%9%3%"$r3d$j0z$$$F$*$/(B(t$B$r4^$^$J$$(B) */
void
bt_discount_pescore(BACKTRELLIS *bt, HTK_Param *param)
{
  int t,i;
  TRELLIS_ATOM *tre;
  
  for (t=0; t<bt->framelen; t++) {
    for (i=0; i<bt->num[t]; i++) {
      tre = bt->rw[t][i];
      tre->backscore -=
#ifdef PASS1_IWCD
	outprob_style(wchmm.wordend[tre->wid], tre->last_tre->wid, t, param);
#else
        outprob(t,wchmm.state[wchmm.wordend[tre->wid]].out, param);
#endif
    }
  }

}

/* $B%U%l!<%`$4$H$N%H%l%j%9>pJs$rC18l(BID$B$G%=!<%H$9$k(B */
static int
compare_wid(TRELLIS_ATOM **a, TRELLIS_ATOM **b)
{
  if ((*a)->wid > (*b)->wid) return 1;
  if ((*a)->wid < (*b)->wid) return -1;
  return 0;
}
void
bt_sort_rw(BACKTRELLIS *bt)
{
  int t;

  for (t=0;t<bt->framelen;t++) {
    qsort(bt->rw[t], bt->num[t], sizeof(TRELLIS_ATOM *),
	  (int (*)(const void *,const void *))compare_wid);
  }
}


/* $B0J9_$N4X?t$O!$(Bbt_sort_rw $B:Q$_$G$"$k$3$H$,A0Ds(B */

/* $B%U%l!<%`$HC18l(BID -> $B%H%l%j%9>e$NC18l>pJs(B $B$N8!:w(B (binary search)*/
TRELLIS_ATOM *
bt_binsearch_atom(BACKTRELLIS *bt, int t, WORD_ID wkey)
{
  /* do binary search */
  /* assume rw are ordered by wid */
  int left, right, mid;
  TRELLIS_ATOM *tmp;
#ifdef WPAIR
  int i;
  LOGPROB maxscore;
  TRELLIS_ATOM *maxtre;
#endif
  
  if (bt->num[t] == 0) return(NULL);
  
  left = 0;
  right = bt->num[t] - 1;
  while (left < right) {
    mid = (left + right) / 2;
    if ((bt->rw[t][mid])->wid < wkey) {
      left = mid + 1;
    } else {
      right = mid;
    }
  }
  tmp = bt->rw[t][left];
  if (tmp->wid == wkey) {
#ifdef WPAIR
    /* $BF1$8(BID$B$NCf$G(Bbackscore$B$,:G$b9b$$$b$N$rJV$9(B */
    /* $B8+$D$+$C$?%-!<$NA08e$r8!:w(B */
    maxscore = LOG_ZERO;
    i = left;
    while (i >= 0 && (tmp = bt->rw[t][i])->wid == wkey) {
      if (maxscore < tmp->backscore) {
	maxscore = tmp->backscore;
	maxtre = tmp;
      }
      i--;
    }
    i = left;
    while (i < bt->num[t] && (tmp = bt->rw[t][i])->wid == wkey) {
      if (maxscore < tmp->backscore) {
	maxscore = tmp->backscore;
	maxtre = tmp;
      }
      i++;
    }

    tmp = maxtre;
#endif /* WORD_GRAPH */

    return(tmp);
  } else {
    return(NULL);
  }
}
