/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* frame-synchronous beam search on 1st pass */
/* use (1-best|word-pair) with static tree structure */

#undef DEBUG

#include <julius.h>

static boolean idc_on;
#ifdef FANCY_INDICATOR
static int idc_interval;
static char *idc_s = "|/-\\|/-\\";
static char *idc_p;
static int idc_count;
#endif


#ifdef WORD_GRAPH
/* -------------------------------------------------------------------- */
/* $B7k2L$+$iC18l%0%i%U$r@8@.(B */

static int glevel;		/* generate_lattice$B$N8F$S=P$7$N?<$5(B */
static int gnodes;		/* node$B$N?t(B */
static int garcs;		/* arc(=$BC18l(B)$B$N?t(B */

static void
generate_lattice(
     int endtime,		/* $BC18l$N=*C<$N%U%l!<%`(B */
     WORD_ID wid,		/* $BC18l(BID */
     BACKTRELLIS *bt)
{
  TRELLIS_ATOM *ta, *ta_last;
  int i,j;
  boolean new_node = FALSE;

  glevel++;

  /* $B%N!<%I$+$i1d$S$kA+0\$r:n$k(B(recursive call) */
  /* $BF1;~4V$r=*C<$H$9$kF1$8C18l$,%"!<%/(B */
  if (endtime >= 0) {
    for (i=0;i<bt->num[endtime];i++) {
      ta = bt->rw[endtime][i];
      if (ta->wid == wid) {
	if (ta->within_wordgraph) {
	  /* $B4{$K%^!<%/$,$D$$$F$$$l$P!$$J$K$b$7$J$$(B */
	} else {
	  /* $B%^!<%/$r$D$1$k(B  */
	  ta->within_wordgraph = TRUE;
	  garcs++;
	  new_node = TRUE;	/* new node mark */
	  ta_last = ta->last_tre;
	  if (debug2_flag) {
	    for(j=0;j<glevel;j++) j_printf(" ");
	    j_printf("%s: %d->", winfo.wname[ta->wid], ta->endtime);
	    if (ta_last->wid == WORD_INVALID) {
	      j_printf("%d(WORD_INVALID)\n", ta_last->endtime);
	    } else {
	      j_printf("%d(%s)\n", ta_last->endtime, winfo.wname[ta_last->wid]);
	    }
	  }
	  /* $B$=$N@h$r:F5"8F$S=P$7(B*/
	  generate_lattice(ta_last->endtime, ta_last->wid, bt);
	}
      }
    }
  }
  /* $B0l$D$G$b?7$?$JA+0\$,:n$l$?$i!$$3$l$O?7%N!<%I(B */
  if (new_node) {
    gnodes++;
  }
  glevel--;
}
#endif

/* ------------------------------------------------------------ */
/* $B7k2L=PNO4X?t(B */

/* $BG'<1C18lNs$rC`<!I=<(!J(B1$B9T$r1[$($k=PNO$KBP1~!K(B */
static int wst;		/* $BA0$N9T$N:G8e$NC18l$N0LCV(B */
static void
put_seq_repeat_init()
{
  wst = 0;
}
static void
put_seq_repeat(WORD_ID *seq, int num)
{
  int i,bgn;
  int len;
  
  j_printf("\r");

  len = 0;
  if (wst == 0) {		/* $B:G=i$N9T$J$i%X%C%@$rI=<((B */
    len += 11;
    j_printf("pass1_best:");
  }
  bgn = wst;
  for (i=bgn;i<num;i++) {
    len += strlen(winfo.woutput[seq[num-1-i]]) + 1;
    if (len > FILLWIDTH) {
      wst = i;
      j_printf("\n");
      len = 0;
    }
    j_printf(" %s",winfo.woutput[seq[num-1-i]]);
  }
  j_flushprint();
}
static void
put_seq(WORD_ID *seq, int num)
{
  int i;

  j_printf("pass1_best:");
  for (i=num-1;i>=0;i--) {
    j_printf(" %s",winfo.woutput[seq[i]]);
  }
  j_printf("\n");
}
static void
put_seq_others(WORD_ID *seq, int num, LOGPROB score, LOGPROB langscore)
{
  int i,j;
  char buf[20];

  j_printf("pass1_best_wordseq:");
  for (i=num-1;i>=0;i--) {
    j_printf(" %s",winfo.wname[seq[i]]);
  }
  j_printf("\n");
  j_printf("pass1_best_phonemeseq:");
  for (i=num-1;i>=0;i--) {
    for (j=0;j<winfo.wlen[seq[i]];j++) {
      center_name(winfo.wseq[seq[i]][j]->name, buf);
      j_printf(" %s", buf);
    }
    if (i > 0) j_printf(" |");
  }
  if (debug2_flag) {
    j_printf("pass1_best_HMMseq_logical:");
    for (i=num-1;i>=0;i--) {
      for (j=0;j<winfo.wlen[seq[i]];j++) {
	j_printf(" %s", winfo.wseq[seq[i]][j]->name);
      }
      if (i > 0) j_printf(" |");
    }
    j_printf("pass1_best_HMMseq_physical:");
    for (i=num-1;i>=0;i--) {
      for (j=0;j<winfo.wlen[seq[i]];j++) {
	j_printf(" %s", winfo.wseq[seq[i]][j]->def->name);
      }
      if (i > 0) j_printf(" |");
    }
  }
  j_printf("\npass1_best_score: %f", score);
  if (separate_score_flag) {
    j_printf(" (AM: %f  LM: %f)", score-langscore, langscore);
  }
  j_printf("\n\n");
}
static void
put_atom(TRELLIS_ATOM *atom)
{
  int i;
  j_printf("%3d,%3d %f %16s (id=%5d)", atom->begintime, atom->endtime,
	 atom->backscore, winfo.wname[atom->wid], atom->wid);
  for (i=0;i<winfo.wlen[atom->wid]; i++) {
    j_printf(" %s",winfo.wseq[atom->wid][i]->name);
  }
  j_printf("\n");
}

/* ------------------------------------------------------------ */
/* backptr$B$r%H%l!<%9$7$F%Q%9>e$NC18l8uJd$r=PNO(B */

static LOGPROB			/* return value (total language score) */
trace_backptr(WORD_ID wordseq[MAXSEQNUM], /* return (reverse order)*/
	      int *rt_wordlen,
	      TRELLIS_ATOM *atom, /* start BACKTRELLIS atom */
	      BACKTRELLIS *backtrellis)	/* backward trellis (return value) */
{
  int wordlen = 0;		/* word length of best sentence hypothesis */
  TRELLIS_ATOM *tretmp;
  LOGPROB langscore = 0.0;
  
  /* wordseq, resultseq $B$N=i4|2=(B */
  wordseq[0] = atom->wid;	/* atom$B$+$i$?$I$k(B */
  wordlen = 1;
  tretmp = atom;
  langscore += tretmp->lscore;
  if (debug2_flag) {
    put_atom(tretmp);
  }
  /* backtrellis$B$r$?$I$C$F$$$/(B */
  while (tretmp->begintime > 0) {
    tretmp = tretmp->last_tre;
/*    t = tretmp->boundtime - 1;
    tretmp = bt_binsearch_atom(backtrellis, tretmp->boundtime-1, tretmp->last_wid);*/
    if (tretmp == NULL) {	/* should not happen */
      j_printerr("ERROR: BackTrellis Pass missing??\n");
      exit(1);
    }
    langscore += tretmp->lscore;
    wordseq[wordlen] = tretmp->wid;
    wordlen++;
    if (debug2_flag) {
      put_atom(tretmp);
    }
    if (wordlen >= MAXSEQNUM) {
      j_printerr("sentence length exceeded ( > %d)\n",MAXSEQNUM);
      exit(1);
    }
  }
  /* $BF@$i$l$?:GL`C18lNs$rI=<($9$k(B */
  *rt_wordlen = wordlen;
  return(langscore);
}

/* $BBh(B1$B%Q%9$N7k2L$rI=<((B */
/* $BJV$jCM(B: backmax*/
static LOGPROB
print_1pass_result(
     BACKTRELLIS *backtrellis,	/* backward trellis (return value) */
     int framelen		/* length of above */
		   )
{
  WORD_ID wordseq[MAXSEQNUM];
  int wordlen;
  int i;
  TRELLIS_ATOM *best;
  int last_time;
  LOGPROB total_lscore;

  for (last_time = framelen - 1; last_time >= 0; last_time--) {
    /* $B:G=*C18l$O(B winfo.tail_silwid $B$K8GDj(B */
    best = bt_binsearch_atom(backtrellis, last_time, winfo.tail_silwid);
    if (best != NULL) break;
  }
  if (last_time < 0) {
    j_printerr("[no sentence-end word survived on last beam]\n");
    /* print null result */
    put_seq(NULL, 0);
    if (debug_flag) put_seq_others(NULL, 0, LOG_ZERO, LOG_ZERO);
    return(LOG_ZERO);
  }
  
  /* $B%Y%9%H%Q%9$r%H%l!<%9(B */
#ifdef WPAIR
  /* bt_binsearch_atom$B$OF1$8C18l(BID$B$J$i:G$b%9%3%"$NNI$$$b$N$rJV$9(B */
#endif
  total_lscore = trace_backptr(wordseq, &wordlen, best, backtrellis);

  /* $B=PNO(B */
  if (progout_flag) {		/* flush last output and end */
    put_seq_repeat(wordseq, wordlen);
    j_printf("\n");
  }
  if (verbose_flag || !progout_flag) {
    put_seq(wordseq, wordlen);	/* output 1st pass result */
				/* for progout, This is re-output */
  }
  if (debug_flag) put_seq_others(wordseq, wordlen, best->backscore, total_lscore);
  
  /* $B7k2L$r(Bglobal$B$K3JG<(B($B5U8~$-(B) */
  for(i=0;i<wordlen;i++) pass1_wseq[i] = wordseq[i];
  pass1_wnum = wordlen;
  pass1_score = best->backscore;

#ifdef WORD_GRAPH
  /* $BC18l%H%l%j%9$+$i!$%i%F%#%9$r@8@.$9$k(B */
  /* $B$3$3$G$O<B:]$K9=C[$9$k$,!$1?MQ>e$O(B within_wordgraph $B$r(B on $B$K$9$k$@$1$G$b(B
   $BNI$$!%(B(bt_store()$B$G(BFALSE$B$K=i4|2=$5$l$F$$$k(B)*/
  glevel = 0;
  gnodes = 1;			/* frame = 0 $B$NJ,(B */
  garcs = 0;
  generate_lattice(last_time, winfo.tail_silwid, backtrellis);
  if (debug_flag) j_printf("word graph generated (nodes=%d,arcs=%d)\n",gnodes, garcs);
#endif

#ifdef TRELLIS_VISUALIZE
  if (trevis_flag) {
    TrevisPutAll(backtrellis);
    TrevisPutPathTrail(best, last_time, backtrellis, 1, 3);
    if (trellis_check_flag) {
      check_trellis(backtrellis, framelen);
    }
  }
#endif

  /* return backmax */
  return(best->backscore);
}

/* $B;~4V(B t $B$G$N%Y%9%H%P%9$rI=<((B */
static void
bt_current_max(BACKTRELLIS *bt, int t)
{
  static WORD_ID wordseq[MAXSEQNUM];
  int wordlen;
  TRELLIS_ATOM *tre;
  TRELLIS_ATOM *tremax;
  LOGPROB maxscore;

  /* bt->list $B$O;~4V=g$K3JG<$5$l$F$$$k(B */
  maxscore = LOG_ZERO;
  tremax = NULL;
  tre = bt->list;
  while (tre != NULL && tre->endtime == t) {
    if (maxscore < tre->backscore) {
      maxscore = tre->backscore;
      tremax = tre;
    }
    tre = tre->next;
  }

  if (maxscore != LOG_ZERO) {
    trace_backptr(wordseq, &wordlen, tremax, bt);
  } else {
    wordlen = 0;
  }
  put_seq_repeat(wordseq, wordlen);

}

static void
bt_current_max_word(BACKTRELLIS *bt, int t)
{
  TRELLIS_ATOM *tre;
  TRELLIS_ATOM *tremax;
  LOGPROB maxscore;
  WORD_ID w;

  /* bt->list $B$O;~4V=g$K3JG<$5$l$F$$$k(B */
  maxscore = LOG_ZERO;
  tremax = NULL;
  tre = bt->list;
  while (tre != NULL && tre->endtime == t) {
    if (maxscore < tre->backscore) {
      maxscore = tre->backscore;
      tremax = tre;
    }
    tre = tre->next;
  }

  if (maxscore != LOG_ZERO) {
    j_printf("%3d: ",t);
    w = tremax->wid;
    j_printf("\"%s[%s]\"(%d)",
	   winfo.wname[w], winfo.woutput[w], w);
    j_printf(" [%d] <- ", tremax->begintime);
    w = tremax->last_tre->wid;
    if (w != WORD_INVALID) {
      j_printf("\"%s[%s]\"(%d) %f\n",
	     winfo.wname[w], winfo.woutput[w],
	     w, tremax->backscore);
    } else {
      j_printf("bgn %f\n", tremax->backscore);
    }
  }
}



/* -------------------------------------------------------------------- */
/* $B%S!<%`C5:wCf$N%Q%9>pJs$r07$&4X?t(B */

/* tlist[] $B$K(Btoken$BK\BN$,G[Ns$H$7$F3NJ]$7$F$"$k(B */
/*         $BI,MW$K1~$8$F(B1$B$D$E$D;H$&!%G[NsHV9f$G%"%/%;%9(B */
/* token[] $B$O%N!<%I$KBP$9$k(Btlist$B>e$N(Btoken$B$X$N%]%$%s%?(B($BG[NsHV9f(B) */

static TOKEN2 *tlist[2];	/* sequencial list of all tokens */
static TOKENID *tindex[2];	/* index for sort */
static int tnum[2];		/* number of tokens used in tlist */
static int maxtnum;		/* maximum number of tokens */
static TOKENID *token;		/* sequencial list per tree node */
static int totalnodenum;	/* allocated node num of above */
static int tl,tn;		/* node id flip/flop */
static int n_start, n_end;

static TRELLIS_ATOM bos;	/* trellis temporary buffer */

static void
malloc_nodes()
{
  totalnodenum = wchmm.n;
  token        = mymalloc(sizeof(TOKENID)*totalnodenum);
  maxtnum = wchmm.n;
  tlist[0]     = mymalloc(sizeof(TOKEN2)*maxtnum);
  tlist[1]     = mymalloc(sizeof(TOKEN2)*maxtnum);
  tindex[0]     = mymalloc(sizeof(TOKENID)*maxtnum);
  tindex[1]     = mymalloc(sizeof(TOKENID)*maxtnum);
  tnum[0] = tnum[1] = 0;
}
static void
expand_tlist()
{
  maxtnum += wchmm.n;
  tlist[0]     = myrealloc(tlist[0],sizeof(TOKEN2)*maxtnum);
  tlist[1]     = myrealloc(tlist[1],sizeof(TOKEN2)*maxtnum);
  tindex[0]     = myrealloc(tindex[0],sizeof(TOKENID)*maxtnum);
  tindex[1]     = myrealloc(tindex[1],sizeof(TOKENID)*maxtnum);
  j_printerr("warn: token space expanded\n");
}
static void
free_nodes()
{
  free(token);
  free(tlist[0]);
  free(tlist[1]);
  free(tindex[0]);
  free(tindex[1]);
}
static void
clear_tlist(int tt)		/* reset token buffer */
{
  tnum[tt] = 0;
}
static void
clear_tokens(int tt)		/* clear node-token for viterbi */
{
  int j;
  for (j=0; j<tnum[tt]; j++) {
    token[tlist[tt][j].node] = TOKENID_UNDEFINED;
  }
}
static TOKENID
create_token()
{
  TOKENID newid;
  newid = tnum[tn];
  tindex[tn][newid] = newid;
  tnum[tn]++;
  if (tnum[tn]>=maxtnum) expand_tlist();
#ifdef WPAIR
  tlist[tn][newid].next = TOKENID_UNDEFINED;
#endif
  return(newid);
}
static void
node_assign_token(int node, TOKENID tkid)
{
#ifdef WPAIR
  /* $B%j%9%H$KDI2C(B */
  tlist[tn][tkid].next = token[node];
#endif
  token[node] = tkid;
  tlist[tn][tkid].node = node;
}
static TOKENID
node_exist_token(int tt, int node, WORD_ID wid)
{
#ifdef WPAIR
  /* $B%j%9%HCf$+$iF1$8D>A0C18l$r;}$D$b$N$r8!:w(B */
  /* $BD>A0C18l$,0[$J$k%Q%9$OJL!9$KJ]B8$5$l$k(B */
#ifdef WPAIR_KEEP_NLIMIT
  /* 1$B%N!<%I$4$H$KJ];}$9$k(Btoken$B?t$N>e8B$r@_Dj(B */
  /* token$B$,L5$$$,>e8B$KC#$7$F$$$k$H$-$O0lHV%9%3%"$NDc$$(Btoken$B$r>e=q$-$9$k(B */
  int i = 0;
  TOKENID lowest_token = TOKENID_UNDEFINED;
#endif
  TOKENID tmp;
  for(tmp=token[node]; tmp != TOKENID_UNDEFINED; tmp=tlist[tt][tmp].next) {
    if (tlist[tt][tmp].last_tre->wid == wid) {
      return(tmp);
    }
#ifdef WPAIR_KEEP_NLIMIT
    if (lowest_token == TOKENID_UNDEFINED ||
	tlist[tt][lowest_token].score < tlist[tt][tmp].score)
      lowest_token = tmp;
    if (++i >= wpair_keep_nlimit) break;
#endif
  }
#ifdef WPAIR_KEEP_NLIMIT
  if (i >= wpair_keep_nlimit) { /* overflow, overwrite lowest score */
    return(lowest_token);
  } else {
    return(TOKENID_UNDEFINED);
  }
#else 
  return(TOKENID_UNDEFINED);
#endif
  
#else  /* not WPAIR */
  /* 1$B$D$@$1J];}(B,$B$3$l$r>o$K>e=q$-(B */
  return(token[node]);
#endif
}

#ifdef DEBUG
/* tlist $B$H(B token $B$NBP1~$r%A%'%C%/$9$k(B(debug) */
static void
node_check_token(int tt)
{
  int i;
  for(i=0;i<tnum[tt];i++) {
    if (node_exist_token(tt, tlist[tt][i].node, tlist[tt][i].last_tre->wid) != i) {
      j_printerr("token %d not found on node %d\n", i, tlist[tt][i].node);
    }
  }
}
#endif


/* ------------------------------------------------------------ */
/* $B;^4"$jMQ$K(Btlist$B$r%=!<%H$9$k4X?t(B */
/* $B>e0LFb$N=g0L$OLd$o$J$$!$(B"$B>e0L(Bn$B8D(B"$B$N6h@Z$j$,J,$+$l$P$h$$(B */
/* $B%R!<%W%=!<%H$r9T$&(B */

#define SD(A) tindex[tn][A-1]
#define SCOPY(D,S) D = S
#define SVAL(A) (tlist[tn][tindex[tn][A-1]].score)
#define STVAL (tlist[tn][s].score)

/* $B>e8~$-(B */
static void
sort_token_upward(int neednum, int totalnum)
{
  int n,root,child,parent;
  TOKENID s;
  for (root = totalnum/2; root >= 1; root--) {
    SCOPY(s, SD(root));
    parent = root;
    while ((child = parent * 2) <= totalnum) {
      if (child < totalnum && SVAL(child) < SVAL(child+1)) {
	child++;
      }
      if (STVAL >= SVAL(child)) {
	break;
      }
      SCOPY(SD(parent), SD(child));
      parent = child;
    }
    SCOPY(SD(parent), s);
  }
  n = totalnum;
  while ( n > totalnum - neednum) {
    SCOPY(s, SD(n));
    SCOPY(SD(n), SD(1));
    n--;
    parent = 1;
    while ((child = parent * 2) <= n) {
      if (child < n && SVAL(child) < SVAL(child+1)) {
	child++;
      }
      if (STVAL >= SVAL(child)) {
	break;
      }
      SCOPY(SD(parent), SD(child));
      parent = child;
    }
    SCOPY(SD(parent), s);
  }
}

/* $B2<8~$-(B */
static void
sort_token_downward(int neednum, int totalnum)
{
  int n,root,child,parent;
  TOKENID s;
  for (root = totalnum/2; root >= 1; root--) {
    SCOPY(s, SD(root));
    parent = root;
    while ((child = parent * 2) <= totalnum) {
      if (child < totalnum && SVAL(child) > SVAL(child+1)) {
	child++;
      }
      if (STVAL <= SVAL(child)) {
	break;
      }
      SCOPY(SD(parent), SD(child));
      parent = child;
    }
    SCOPY(SD(parent), s);
  }
  n = totalnum;
  while ( n > totalnum - neednum) {
    SCOPY(s, SD(n));
    SCOPY(SD(n), SD(1));
    n--;
    parent = 1;
    while ((child = parent * 2) <= n) {
      if (child < n && SVAL(child) > SVAL(child+1)) {
	child++;
      }
      if (STVAL <= SVAL(child)) {
	break;
      }
      SCOPY(SD(parent), SD(child));
      parent = child;
    }
    SCOPY(SD(parent), s);
  }
}

/* $B%=!<%H$9$k!J>e8~$-!&2<8~$-$GAa$$J}$r;H$&!K(B */
static void
sort_token_no_order(int neednum, int *start, int *end)
{
  int totalnum = tnum[tn];
  int restnum;

  restnum = totalnum - neednum;

  if (neednum >= totalnum) {
    /* no need to sort */
    *start = 0;
    *end = totalnum - 1;
  } else if (neednum < restnum)  {
    sort_token_upward(neednum,totalnum);
    *start = totalnum - neednum;
    *end = totalnum - 1;
  } else {
    sort_token_downward(restnum,totalnum);
    *start = 0;
    *end = neednum - 1;
  }
}
    

/* ====================================================================== */
/* $B%S!<%`C5:w%a%$%s4X?t72(B */

/* $B=i4|%9%3%"$r@_Dj$9$k(B */
static void
init_nodescore(HTK_Param *param)
{
  TOKENID newid;
  TOKEN2 *new;
  int node;

  /* $B=i4|2=(B */
  bos.wid = WORD_INVALID;
  bos.begintime = bos.endtime = -1;
  for(node=0;node<totalnodenum;node++) {
    token[node] = TOKENID_UNDEFINED;
  }
  tnum[0] = tnum[1]  = 0;
#ifdef PASS1_IWCD
  outprob_style_cache_init();
#endif
  if (hmminfo.is_tied_mixture) {
    calc_all_mixtures_and_cache(0, param, &hmminfo);
  }
  /* $B3+;OE@!aC18l(BID winfo.head_silwid $B$NC18l$N;OC<(B */
  newid = create_token();
  new = &(tlist[tn][newid]);
  node = wchmm.offset[winfo.head_silwid][0];
  if (wchmm.state[node].sc != NULL
#ifdef UNIGRAM_FACTORING
      || wchmm.state[node].fscore != LOG_ZERO
#endif
      ) {
    new->last_lscore = max_successor_prob(WORD_INVALID, node);
  } else {
    new->last_lscore = 0.0;
  }
#ifdef FIX_PENALTY
  new->last_lscore = new->last_lscore * lm_weight;
#else
  new->last_lscore = new->last_lscore * lm_weight + lm_penalty;
#endif
  new->last_tre = &bos;
#ifdef PASS1_IWCD
  new->score = outprob_style(node, WORD_INVALID, 0, param) + new->last_lscore;
#else  /* ~PASS1_IWCD */
  new->score = outprob(0, wchmm.state[node].out, param) + new->last_lscore;
#endif /* PASS1_IWCD */
  node_assign_token(node, newid);
}

/* $B:G=i$N%U%l!<%`(B */
void
get_back_trellis_init(HTK_Param *param,	/* parameter vectors of speech data (only t=0 used) */
		      BACKTRELLIS *backtrellis	/* backward trellis (return value) */
		      )
{
  
  /* tl,tn $B$N=i4|CM@_Dj(B */
  tn = 0;
  tl = 1;

  bt_prepare(backtrellis);

  /* $B%N!<%I$rMQ0U(B */
  malloc_nodes();
  
  /* $B=i4|%9%3%"$r(B nodescore[tn] $B$K%;%C%H(B */
  init_nodescore(param);

  /* $B0l2s%=!<%H$r9T$C$F$*$/(B */
  sort_token_no_order(trellis_beam_width, &n_start, &n_end);

  /* $B=PNO>uBV$r=i4|>uBV$K(B */
  if (progout_flag) {
    put_seq_repeat_init();
  }

  if (!realtime_flag && verbose_flag && (!progout_flag) && isatty(1)) {
    idc_on = TRUE;
#ifdef FANCY_INDICATOR
  /* $B%$%s%8%1!<%?@_Dj(B */
  /* $BI8=`=PNO$,(BTTY$B$J$i(BON */
    /* $B0lIC$N%U%l!<%`?t(B */
    idc_interval = OUTPUT_INTERVAL;
    /* $B%]%$%s%?$H%+%&%s%H$r=i4|2=(B */
    idc_p = idc_s;
    idc_count = 0;
#endif
  } else { 
    idc_on = FALSE;
  }
  
}

#ifdef FANCY_INDICATOR
static void
indicator()
{
  if (++idc_count <= idc_interval) return;
  idc_count = 0;
  fprintf(stderr, "\r%c", *idc_p);
  idc_p ++; if (*idc_p == '\0') idc_p = idc_s;
}
#endif /* FANCY_INDICATOR */


static LOGPROB maxscore, minscore;
static TRELLIS_ATOM *tre;
static int node, stid;
static A_CELL *ac;
static LOGPROB *iwparray;

/* $B%S!<%`$r(B1$B%U%l!<%`?J$a$k!J%a%$%s!K(B */
boolean
get_back_trellis_proceed(
     int t,			/* $B;~4V(B t  */
     HTK_Param *param,		/* parameter vector of speech data (use only t) */
     BACKTRELLIS *backtrellis	/* backward trellis (return value) */
     )
{
  LOGPROB tmpsum;
  int j, next_node, sword;
  TOKEN2  *tk, *tknext;
  TOKENID  tknextid;
  LOGPROB ngram_score_cache;	/* N-gram$B%G!<%?$N%-%c%C%7%e$5$l$?%G!<%?(B */

  /********************************/
  /* backtrellis proceed one step */
  /********************************/

  /* tl ... 0 1 0 1 */
  /* tn ... 1 0 1 0 */
  tn = t % 2;
  tl = (t+1) % 2 ;
  if (idc_on) {
#ifdef FANCY_INDICATOR
    indicator();
#else  /* ~FANCY_INDICATOR */
    j_printerr(".");
#endif /* ~FANCY_INDICATOR */
  }
  maxscore = LOG_ZERO;
  minscore = 0;

#ifdef DEBUG
  /* debug */
  /* node_check_token(tl); */
#endif

  /* clear node token pointers (clear only last used token)*/
  clear_tokens(tl);

/* 
 *   {
 *     int i;
 *     for(i=0;i<totalnodenum;i++) {
 *	 if (token[i] != TOKENID_UNDEFINED) {
 *	   j_printerr("Sdasadasdlsaskd\n");
 *	   exit(1);
 *	 }
 *     }
 *   }
 */

  /* $B%=!<%H:Q$_$N(BTOKEN$B=89g$+$i(Bviterbi$B7PO)$r7W;;$9$k(B */
  for (j=n_start;j<=n_end;j++) {
    tk = &(tlist[tl][tindex[tl][j]]);
    node = tk->node;
    if (tk->score <= LOG_ZERO) continue;

    /**************/
    /* $BC18lFbA+0\(B */
    /**************/
    for (ac = wchmm.state[node].ac; ac; ac = ac->next) {
      next_node = ac->arc;
      tmpsum = tk->score + ac->a;
      ngram_score_cache = LOG_ZERO;

      /* factoring:
         arc$B$,<+8JA+0\$G$J$$C18lFb$NA+0\$G!$$+$DA+0\@h$K(Bsuccessor$B%j%9%H(B
	 $B$,$"$l$P!$(Blexicon tree $B$NJ,4tItJ,$NA+0\$G$"$k(B */
      if (next_node != node) {
	if (wchmm.state[next_node].sc != NULL
#ifdef UNIGRAM_FACTORING
	    || wchmm.state[next_node].fscore != LOG_ZERO
#endif
	    ){
	  /* $B$3$N(Btmpsum$B$KEAHB$7$F$$$k8@8l%9%3%"$r!$A+0\@h$N(Bsuccessor$B%j%9%HCf$N(B
	     $BC18l$H%Q%9>e$N:G8e$NC18l$H$N(B bi-gram$B@\B33NN($N:GBgCM$K99?7(B */
#ifdef FIX_PENALTY
	  if (tk->last_tre->wid == WORD_INVALID) {
	    ngram_score_cache = max_successor_prob(tk->last_tre->wid, next_node) * lm_weight;
	  } else {
	    ngram_score_cache = max_successor_prob(tk->last_tre->wid, next_node) * lm_weight + lm_penalty;
	  }
#else
	  ngram_score_cache = max_successor_prob(tk->last_tre->wid, next_node) * lm_weight + lm_penalty;
#endif
	  /* $BC18lFb$N99?7(B:
	     $B:G8e$KB-$7$?CM$r0z$$$F$+$iB-$7D>$9!%(B */
	  tmpsum -= tk->last_lscore;
	  tmpsum += ngram_score_cache;
	}
      }

      if ((tknextid = node_exist_token(tn, next_node, tk->last_tre->wid)) != TOKENID_UNDEFINED) {
	tknext = &(tlist[tn][tknextid]);
	if (tknext->score < tmpsum) {
	  /* overwrite preserving link */
	  tknext->last_tre = tk->last_tre;
	  tknext->last_lscore = (ngram_score_cache != LOG_ZERO) ? ngram_score_cache : tk->last_lscore;
	  tknext->score = tmpsum;
	}
      } else {			/* assign new */
	if (tmpsum > LOG_ZERO) { /* valid token */
	  tknextid = create_token();
	  tknext = &(tlist[tn][tknextid]);
	  tk = &(tlist[tl][tindex[tl][j]]);
	  tknext->last_tre = tk->last_tre;
	  tknext->last_lscore = (ngram_score_cache != LOG_ZERO) ? ngram_score_cache : tk->last_lscore;
	  tknext->score = tmpsum;
	  node_assign_token(next_node, tknextid);
	}
      }
    }
    
    
    if (wchmm.stend[node] != WORD_INVALID) {
      sword = wchmm.stend[node];
      /********************************/
      /* $BC18l$N=*C<%N!<%I!DC18l4VA+0\(B */
      /********************************/

      /* $BC18l$N=*C<$K$"$?$k(Bnode -> backtrellis $B$KJ]B8(B */
      tre = (TRELLIS_ATOM *)mymalloc(sizeof(TRELLIS_ATOM));
      tre->wid = sword;
      tre->backscore = tk->score;
      tre->begintime = tk->last_tre->endtime + 1;
      tre->endtime   = t-1;
      tre->last_tre  = tk->last_tre;
      tre->lscore    = tk->last_lscore;
      bt_store(backtrellis, tre);
      /* $B$3$N(Btre$B$rA+0\@h$ND>A0(Btrellis$BC18l$H$7$FM?$($k(B */

      /* $BKvHxC18l$N=*C<$J$i!$$I$3$X$bA+0\$5$;$J$$(B */
      if (sword == winfo.tail_silwid) continue;
      /* $BC18l4V(Bfactoring$BCM$r$"$i$+$8$aF@$k(B */
      iwparray = max_successor_prob_iw(sword);
      
      for (stid = wchmm.startnum - 1; stid >= 0; stid--) {
	next_node = wchmm.startnode[stid]; /* $BC18l$N;OC<%N!<%I(B */

/*#ifdef USE_NGRAM*/
	  /* $B@hF,C18l$N;OC<$X$OA+0\$5$;$J$$(B -> wchmm.c $B$K0\F0(B*/
/*	  if(wchmm.ststart[next_node] == winfo.head_silwid) continue;*/
/*#endif*/
	tmpsum = tk->score + wchmm.wordend_a[sword];
	ngram_score_cache = iwparray[stid] * lm_weight + lm_penalty;
	tmpsum += ngram_score_cache;
	if ((tknextid = node_exist_token(tn, next_node, tre->wid)) != TOKENID_UNDEFINED) {
	  tknext = &(tlist[tn][tknextid]);
	  if (tknext->score < tmpsum) {
	    /* overwrite preserving link */
	    tknext->last_lscore = ngram_score_cache;
	    tknext->score = tmpsum;
	    /* $BA*$P$l$?A+0\@h$K!$(Bi $B$KBP1~$9$kC18l$r?7$?$K!VD>A0$NC18l!W$H$7$F(B
	       $BEAHB(B */
	    tknext->last_tre = tre;
	  }
	} else {
	  if (tmpsum > LOG_ZERO) { /* valid token */
	    tknextid = create_token();
	    tknext = &(tlist[tn][tknextid]);
	    tk = &(tlist[tl][tindex[tl][j]]);
	    tknext->last_lscore = ngram_score_cache;
	    tknext->score = tmpsum;
	    tknext->last_tre = tre;
	    node_assign_token(next_node, tknextid);
	  }
	}
      }
    }
    
  }

  /* $B<!CJ$NM-8z%N!<%IFb$N=PNO3NN($r7W;;$7$F%9%3%"$K2C$($k(B */
  if (hmminfo.is_tied_mixture) {
    calc_all_mixtures_and_cache(t, param, &hmminfo);
  }
  
  for (j=0;j<tnum[tn];j++) {
    tk = &(tlist[tn][tindex[tn][j]]);
#ifdef PASS1_IWCD
    tk->score += outprob_style(tk->node, tk->last_tre->wid, t, param);
#else  /* ~PASS1_IWCD */
    tk->score += outprob(t, wchmm.state[tk->node].out, param);
#endif /* PASS1_IWCD */
  }

  /* tlist[tl]$B$r<!CJ$N$?$a$K%j%;%C%H(B */
  clear_tlist(tl);

  /* $B%R!<%W%=!<%H$rMQ$$$F$3$NCJ$N%N!<%I=89g$+$i>e0L(B(bwidth)$B8D$rF@$F$*$/(B */
  /* ($B>e0LFb$N=gNs$OI,MW$J$$(B) */
  sort_token_no_order(trellis_beam_width, &n_start, &n_end);

  /* check for sort result */
/* 
 *     j_printf("%d: vwidth=%d / %d\n",t, vwidth[tn], wchmm.n);
 */
    /* make sure LOG_ZERO not exist width vwidth[tn] */
/* 
 *     for (j=0;j<vwidth[tn];j++) {
 *	 if (nodescore[tn][hsindex[tn][j+1]] <= LOG_ZERO) {
 *	   j_printf("dsadsadas %d %d\n",hsindex[tn][j+1], nodescore[tn][hsindex[tn][j+1]]);
 *	   exit(1);
 *	 }
 *     }
 */
  
  if (progout_flag) {
    /* $BA2<!=PNO(B */
    if ((t % OUTPUT_INTERVAL) == 0) {
      bt_current_max(backtrellis, t-1);
    }
  }
  /* j_printf("%d: %d\n",t,tnum[tn]); */

  /* for debug */
  if (debug2_flag) {
    bt_current_max_word(backtrellis, t-1);
  }

  /* $B%S!<%`Fb%N!<%I?t$,(B 0 $B$K$J$C$F$7$^$C$?$i!$6/@)=*N;(B */
  if (tnum[tn] == 0) {
    j_printerr("Error: %dth frame: no nodes left in beam! model mismatch or wrong input?\n", t);
    return(FALSE);
  }

  return(TRUE);
    
}

/* $B:G8e$N%U%l!<%`=*N;8e$N8e=hM}(B */
void
get_back_trellis_end(
     HTK_Param *param,		/* parameters */
     BACKTRELLIS *backtrellis,	/* backward trellis (return value) */
     LOGPROB *backmax		/* backmax (retuen value) */
     )
{
  int t, node, j;
  TOKEN2 *tk;

  /* $B:G8e$K%S!<%`Fb$K;D$C$?(B node $B$N=hM}(B */
  t = param->samplenum;
  tn = t % 2;
  tl = (t+1) % 2 ;

  for (j=n_start; j<=n_end; j++) {
    tk = &(tlist[tl][tindex[tl][j]]);
    node = tk->node;

    /* $BC18l$N=*C<$K$"$?$k(Bnode -> backtrellis $B$KJ]B8(B */
    if (wchmm.stend[node] != WORD_INVALID) {
      tre = (TRELLIS_ATOM *)mymalloc(sizeof(TRELLIS_ATOM));
      tre->wid = wchmm.stend[node];
      tre->backscore = tk->score;
      tre->begintime = tk->last_tre->endtime + 1;
      tre->endtime   = t-1;
      tre->last_tre  = tk->last_tre;
      tre->lscore    = tk->last_lscore;
      bt_store(backtrellis, tre);
    }
  }
  
  /* framelen $B$r%;%C%H(B */
  backtrellis->framelen = param->samplenum;

  /* $B%H%l%j%9C18l$N:FG[CV$H%=!<%H(B */
  bt_relocate_rw(backtrellis);
  bt_sort_rw(backtrellis);
  
  /* $B7k2L$rI=<($9$k(B (best pass $B$rI=<((B)*/
  if (verbose_flag && (!progout_flag)) j_printerr("\n");
  *backmax = print_1pass_result(backtrellis, param->samplenum);

  /* succesor cache $B$r3+J|(B */
  /* max_successor_cache_free(); *//* wchmm.winfo$B$H$b$KL5JQ99$N>l9g$O(B */
                                 /* free$B$7$J$/$F$b$$$$(B */
  
  free_nodes();
}
  


/* -------------------------------------------------------------------- */
void
get_back_trellis(
     HTK_Param *param,		/* parameter */
     BACKTRELLIS *backtrellis,	/* backward trellis (return value) */
     LOGPROB *backmax		/* backmax (retuen value) */
     )
{
  int t;

  /* $B=i4|2=5Z$S(B t=0 $B$r7W;;(B */
  get_back_trellis_init(param, backtrellis);

  /* $B%k!<%W(B */
  for (t = 1; t < param->samplenum; t++) {
    if (get_back_trellis_proceed(t,param,backtrellis) == FALSE
	|| catch_intr_flag) {	/* terminate by user */
      /* search failed */
      *backmax = LOG_ZERO;
      free_nodes();
      return;
    }
  }
  /* $B=*N;=hM}(B */
  get_back_trellis_end(param,backtrellis,backmax);
}

/* end of file */
