/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* top-level external functions */
/* should be included after all include files */

/* backtrellis.c */
void bt_init(BACKTRELLIS *);
void bt_prepare(BACKTRELLIS *);
void bt_store(BACKTRELLIS *, TRELLIS_ATOM *);
void bt_relocate_rw(BACKTRELLIS *);
void bt_discount_pescore(BACKTRELLIS *, HTK_Param *);
void bt_sort_rw(BACKTRELLIS *);
TRELLIS_ATOM *bt_binsearch_atom(BACKTRELLIS *, int, WORD_ID);

/* backtrellis_check.c */
#ifdef TRELLIS_VISUALIZE
void check_trellis(BACKTRELLIS *backtrellis, int framelen);
#endif

/* factoring_sub.c */
void make_successor_list();
void max_successor_cache_init(void);
void max_successor_cache_free(void);
LOGPROB max_successor_prob(WORD_ID, int);
LOGPROB *max_successor_prob_iw(WORD_ID);
void  calc_all_unigram_factoring_values();

/* beam.c */
void get_back_trellis_init(HTK_Param *param, BACKTRELLIS *backtrellis);
boolean get_back_trellis_proceed(int t, HTK_Param *param, BACKTRELLIS *backtrellis);
void get_back_trellis_end(HTK_Param *param, BACKTRELLIS *backtrellis, LOGPROB *backmax);
void get_back_trellis(HTK_Param *param, BACKTRELLIS *backtrellis, LOGPROB *backmax);

/* outprob_style.c */
#ifdef PASS1_IWCD
void outprob_style_cache_init();
LOGPROB outprob_style(int node, int last_wid, int t, HTK_Param *param);
#endif

/* ngram_decode.c */
#include "search.h"
int ngram_firstwords(NEXTWORD **nw, int peseqlen);
int ngram_nextwords(NODE *hypo, NEXTWORD **nw);
boolean ngram_acceptable(NODE *hypo);
LOGPROB ngram_eosscore(NODE *hypo);

/* search_bestfirst_main.c */
#ifdef SCAN_BEAM
#ifdef SCAN_BEAM_1
void envl_init(int framenum);
#endif
#ifdef SCAN_BEAM_2
void envl_premake(BACKTRELLIS *bt, LOGPROB backmax);
#endif
#endif /* SCAN_BEAM */
void putsentence(NODE *hypo, int num);
void wchmm_fbs(HTK_Param *param, BACKTRELLIS *backtrellis, LOGPROB backmax, int stacksize, int ncan, int maxhypo, int (*gm_get_first_words)(NEXTWORD **, int), int (*gm_get_next_words)(NODE *, NEXTWORD **), boolean (*gm_acceptable)(NODE *), LOGPROB (*gm_get_eos_score)(NODE *));

/* search_bestfirst_v?.c */
void free_node(NODE *node);
NODE *cpy_node(NODE *dst, NODE *src);
NODE *newnode();
void malloc_wordtrellis();
void free_wordtrellis();
void scan_word(NODE *now, HTK_Param *param);
void next_word(NODE *now, NODE *new, NEXTWORD *nword, HTK_Param *param, BACKTRELLIS *backtrellis);
void start_word(NODE *new, NEXTWORD *nword, HTK_Param *param, BACKTRELLIS *backtrellis);
void last_next_word(NODE *now, NODE *new, HTK_Param *param);

/* wav2mfcc.c */
HTK_Param *new_wav2mfcc(SP16 speech[], int speechlen);

#ifdef TRELLIS_VISUALIZE
/* trevis_x11.c */
void TrevisInit();
void TrevisPrepare();
void TrevisPutAll(BACKTRELLIS *bt);
void TrevisPutWord(TRELLIS_ATOM *atom, int t, int cidx, BACKTRELLIS *bt);
void TrevisPutPathTrail(TRELLIS_ATOM *atom, int t, BACKTRELLIS *bt, int cidx, int width);
#endif /* TRELLIS_VISUALIZE */

/* version.c */
void put_header(FILE *stream);
void put_compile_defs(FILE *stream);

/* wchmm.c */
void print_wchmm_info();
void build_wchmm();
void build_wchmm2();

/* wchmm_check.c */
void wchmm_check_interactive();
void check_wchmm();

/* realtime.c --- callback for adin_cut() */
void RealTimeInit();
void RealTimePipeLinePrepare();
boolean RealTimePipeLine(SP16 *Speech, int len);
HTK_Param *RealTimeParam(LOGPROB *backmax);

/* word_align.c */
void word_align(WORD_ID *words, short wnum, HTK_Param *param);
void phoneme_align(WORD_ID *words, short wnum, HTK_Param *param);
void word_rev_align(WORD_ID *revwords, short wnum, HTK_Param *param);
void phoneme_rev_align(WORD_ID *revwords, short wnum, HTK_Param *param);

/* main.c */
char *get_line(char *prompt);
#ifdef WORD_GRAPH
void verify_WORD_GRAPH_define();
#else
void verify_WORD_GRAPH_undefine();
#endif
#ifdef WPAIR
void verify_WPAIR_define();
#else
void verify_WPAIR_undefine();
#endif
