/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* factoring_sub.c --- subroutines for factoring */

/* $Id: factoring_sub.c,v 1.1.1.1 1998/07/16 07:40:23 ri Exp $ */

#include <julius.h>


/*----------------------------------------------------------------------*/
/* $B>uBV$4$H$K!$$=$N(B successor word $B$N%j%9%H$rM?$($k(B */
/* sc->word $B$O(B N-gram$BC18lHV9f(B */
/* $B>:=g$KJ]B8$9$k(B */
static void
add_successor(int node, WORD_ID w)
{
  S_CELL *sctmp, *sc;
  sctmp=(S_CELL *) mymalloc(sizeof(S_CELL));
  if (sctmp == NULL) {
    j_printerr("malloc fault at add_succesor(%d,%d)\n",node,w);
    exit(1);
  }
  sctmp->word = winfo.wton[w];
  
  sc = wchmm.state[node].sc;
  if (sc == NULL || sctmp->word < sc->word) {
    sctmp->next = sc;
    wchmm.state[node].sc = sctmp;
  } else {
    for(;sc;sc=sc->next) {
      if (sc->next == NULL || sctmp->word < (sc->next)->word) {
	if (sctmp->word == sc->word) break; /* 2$B=EEPO?2sHr(B */
	sctmp->next = sc->next;
	sc->next = sctmp;
	break;
      }
    }
  }
}

static boolean
match_successor(int node1, int node2)
{
  S_CELL *sc1,*sc2;

  /* successor$B$O>:=g$K3JG<$5$l$F$k$O$:(B */
  sc1 = wchmm.state[node1].sc;
  sc2 = wchmm.state[node2].sc;
  for (;;) {
    if (sc1 == NULL || sc2 == NULL) {
      if (sc1 == NULL && sc2 == NULL) {
	return TRUE;
      } else {
	return FALSE;
      }
    } else if (sc1->word != sc2->word) {
      return FALSE;
    }
    sc1 = sc1->next;
    sc2 = sc2->next;
  }
}

static void
free_successor(int node)
{
  S_CELL *sc;
  S_CELL *sctmp;

  sc = wchmm.state[node].sc;
  while (sc != NULL) {
    sctmp = sc;
    sc = sc->next;
    free(sctmp);
  }
  wchmm.state[node].sc = NULL;
}

void
make_successor_list()
{
  int node;
  WORD_ID w;
  int i;
  int freed_num = 0;
  boolean *freemark;

  VERMES("  make successor lists for factoring...");

  /* $B=i4|2=(B */
  freemark = (boolean *)mymalloc(sizeof(boolean) * wchmm.n);
  for (node=0;node<wchmm.n;node++) {
    wchmm.state[node].sc = NULL;
    freemark[node] = FALSE;
  }

  /* successor list $B$r9=C[(B */
  for (w=0;w<winfo.num;w++) {
    /* $B3F2;AG$N:G=i$N(Bnode$B$K$D$$$F(B */
    for (i=0;i<winfo.wlen[w];i++) {
/*      j_printf("node=%d ststart=%d\n", node, wchmm.ststart[node]);*/
      add_successor(wchmm.offset[w][i], w);
    }
    /* $BC18l$N=*C<>uBV$KBP$7$F(B */
    add_successor(wchmm.wordend[w], w);
/*    j_printf("%s[id=%d] endnode=%d ststart=%d\n", winfo.wname[w], w, node, wchmm.ststart[node]);*/
  }

  /* $BJ,4tE@(B(=successor$B$NJQ$o$jL\(B)$B$N$_;D$9(B */
  /* $B%"%k%4%j%:%`(B: $BLZ$NA4$F$N%j!<%U$+$i$?$I$j!$(B
   *   1)$B?F$H<+J,$N(Bsuccessor_list$B$rHf3S$7!$(B
   *          $BF1$8$J$i!$<+J,$N$r>C$9%^!<%/$r$D$1$k!%(B
   *          $B0c$($P!$$=$N$^$^!%(B
   */
  for (w=0;w<winfo.num;w++) {
    node = wchmm.wordend[w];	/* $BC18l$N=*C<>uBV$+$i%9%?!<%H(B */
    i = winfo.wlen[w]-1;
    while (i >= 0) {
      if (node == wchmm.offset[w][i]) {	/* $B>uBV?t(B=1$B$N2;AG$O!$(B
					   wchmm.wordend[word] ==
					   wchmm.offset[word][$B:G8e$N2;AG(B]
					   $B$J$N$G!$%Q%9(B */
	i--;
	continue;
      }
      if (match_successor(node, wchmm.offset[w][i])) {
	freemark[node] = TRUE;
      }
/* 
 *	 if (freemark[wchmm.offset[w][i]] != FALSE) {
 *	   break;
 *	 }
 */
      node = wchmm.offset[w][i];
      i--;
    }
  }
  for (node=0;node<wchmm.n;node++) {
    if (freemark[node] == TRUE) {
      freed_num++;
      free_successor(node);
    }
  }

  if (debug2_flag) {
    j_printerr("%d freed...", freed_num);
  }

  free(freemark);

  wchmm.scnum = 0;
  for (node=0;node<wchmm.n;node++) {
    if (wchmm.state[node].sc != NULL) wchmm.scnum++;
  }

  VERMES("done\n");
}



/* -------------------------------------------------------------------- */
/* factoring $B$N<B9T(B */


/* cache$B$OD>A0$@$13P$($k!$(Bnode$BC10L$G4IM}(B */
static LOGPROB *probcache;
static WORD_ID *lastwcache;
/* $BC18l4V$G$N(Bfactoring$B$OJL%k!<%A%s(B */
/* $BC18lC10L$G!$$=$l$>$lA4C18l@hF,%N!<%I$X$NCM$r(Bcache */
/* $B0lEY7W;;$7$?$i:G8e$^$G3P$($F$*$/(B */
static LOGPROB **iw_sc_cache;
#ifdef HASH_CACHE_IW
static WORD_ID *iw_lw_cache;
static int iw_cache_limit;
#endif

/* $B%-%c%C%7%eNN0h$N3NJ](B */
void
max_successor_cache_init()
{
  int i;

  probcache = (LOGPROB *) mymalloc(sizeof(LOGPROB) * wchmm.n);
  lastwcache = (WORD_ID *) mymalloc(sizeof(WORD_ID) * wchmm.n);
  for (i=0;i<wchmm.n;i++) {
    lastwcache[i] = WORD_INVALID;
  }
#ifdef HASH_CACHE_IW
  iw_cache_limit = ngram.max_word_num * iw_cache_rate / 100;
  if (iw_cache_limit < 10) iw_cache_limit = 10;
  iw_lw_cache = (WORD_ID *)mymalloc(sizeof(WORD_ID) * iw_cache_limit);
  iw_sc_cache = (LOGPROB **)mymalloc(sizeof(LOGPROB *) * iw_cache_limit);
  for (i=0;i<iw_cache_limit;i++) {
    iw_lw_cache[i] = WORD_INVALID;
    iw_sc_cache[i] = NULL;
  }
#else  /* FULL CACHE */
  iw_sc_cache = (LOGPROB **)mymalloc(sizeof(LOGPROB *) * ngram.max_word_num);
  for (i=0;i<ngram.max_word_num;i++) {
    iw_sc_cache[i] = NULL;
  }
#endif
}

/* $B%-%c%C%7%eNN0h$N2rJ|(B */
void
max_successor_cache_free()
{
  int i;
  free(probcache);
  free(lastwcache);
#ifdef HASH_CACHE_IW
  for (i=0;i<iw_cache_limit;i++) {
    if (iw_sc_cache[i] != NULL) free(iw_sc_cache[i]);
  }
  free(iw_sc_cache);
  free(iw_lw_cache);
#else
  for (i=0;i<ngram.max_word_num;i++) {
    if (iw_sc_cache[i] != NULL) free(iw_sc_cache[i]);
  }
  free(iw_sc_cache);
#endif
}

#ifdef UNIGRAM_FACTORING
/* 1-gram factoring$BMQ(B: */
/* $B3F%N!<%I$N(B 1-gram factoring$BCM!J8GDj!K$r7W;;$7$F3JG<$9$k(B */
/* $BC5:wA0$K0lEYFI$s$G$*$1$P$h$$(B */
/* $B$^$?!$KvC<0J30$N(Bsc$B$OI,MW$J$$$N$G%U%j!<$9$k(B */
void
calc_all_unigram_factoring_values()
{
  S_CELL *sc, *sctmp;
  LOGPROB tmpprob, maxprob;
  int n;

  for (n=0;n<wchmm.n;n++) {
    wchmm.state[n].fscore = LOG_ZERO;
    sc = wchmm.state[n].sc;
    if (sc != NULL) {
      if (sc->next == NULL) {
	/* $BKvC<(B ... $B$N$3$9(B */
      } else {
	/* 1-gram$B$N(Bmax$BCM$r7W;;(B */
	/* fscore$B$K3JG<8e!$(Bsc$B$r2rJ|(B */
	maxprob = LOG_ZERO;
	for (sctmp = sc; sctmp; sctmp = sctmp->next) {
	  tmpprob = uni_prob(&ngram, sctmp->word);
	  if (maxprob < tmpprob) maxprob = tmpprob;
	}
	wchmm.state[n].fscore = maxprob;
	free_successor(n);
      }
    }
  }
}
/* fscore\SC   != NULL   == NULL */
/* !=LOG_ZERO   1-gram    1-gram */
/* ==LOG_ZERO   2-gram      ---  */

#else  /* ~UNIGRAM_FACTORING */

/* $B%N!<%I$KBP1~$9$k(B 2-gram factoring$BCM$rJV$9(B */
static LOGPROB
calc_successor_prob(WORD_ID last_nword, int node)
{
  S_CELL *sc;
  LOGPROB tmpprob, maxprob;

  maxprob = LOG_ZERO;
  for (sc = wchmm.state[node].sc; sc; sc = sc->next) {
    tmpprob = bi_prob_lr(&ngram,last_nword, sc->word);
    if (maxprob < tmpprob) maxprob = tmpprob;
  }
  return(maxprob);
}

#endif  /* ~UNIGRAM_FACTORING */

/* $B%N!<%I$KBP1~$9$k(Bfactoring$BCM$rJV$9(B($BC18lFb(B) */
LOGPROB
max_successor_prob(WORD_ID lastword, int node)
{
  LOGPROB maxprob;
  WORD_ID last_nword;

  if (lastword != WORD_INVALID) { /* return nothing */
    last_nword = winfo.wton[lastword];
#ifdef UNIGRAM_FACTORING
    if (wchmm.state[node].fscore != LOG_ZERO) {
      /* return 1-gram factoring value already calced */
      return(wchmm.state[node].fscore);
    } else {
      /* return precise 2-gram score */
      if (last_nword != lastwcache[node]) {
	/* calc and cache */
	maxprob = bi_prob_lr(&ngram, last_nword, (wchmm.state[node].sc)->word);
	lastwcache[node] = last_nword;
	probcache[node] = maxprob;
	return(maxprob);
      } else {
	/* return cached */
	return (probcache[node]);
      }
    }
#else  /* UNIGRAM_FACTORING */
    /* 2-gram */
    if (last_nword != lastwcache[node]) {
      maxprob = calc_successor_prob(last_nword, node);
      /* store to cache */
      lastwcache[node] = last_nword;
      probcache[node] = maxprob;
      return(maxprob);
    } else {
      return (probcache[node]);
    }
#endif /* UNIGRAM_FACTORING */
  } else {
    return(0.0);
#if 0
    maxprob = LOG_ZERO;
    for (sc=wchmm.state[node].sc;sc;sc=sc->next) {
      tmpprob = uni_prob(&ngram, sc->word);
      if (maxprob < tmpprob) maxprob = tmpprob;
    }
    return(maxprob);
#endif
  }

}

/* $B%N!<%I$KBP1~$9$k(Bfactoring$BCM$rJV$9(B($BC18l4V(B) */
/* $B@h$K7W;;$7$FG[Ns$rJV$9(B */
#ifdef HASH_CACHE_IW
#define hashid(A) A % iw_cache_limit
#endif
LOGPROB *
max_successor_prob_iw(WORD_ID lastword)
{
  int i, x, node;
  int last_nword;

  /* lastword$B$O%"%k%4%j%:%`>e(BWORD_INVALID$B$K$O$J$jF@$J$$(B */
  last_nword = winfo.wton[lastword];
#ifdef HASH_CACHE_IW
  x = hashid(last_nword);
  if (iw_lw_cache[x] == last_nword) { /* cache hit */
    return(iw_sc_cache[x]);
  }
#else  /* full cache */
  if (iw_sc_cache[last_nword] != NULL) { /* cache hit */
    return(iw_sc_cache[last_nword]);
  }
  x = last_nword;
  /* cache mis-hit, calc probs and cache new */
#endif

  /* allocate cache memory */
  if (iw_sc_cache[x] == NULL) {
    iw_sc_cache[x] = (LOGPROB *)mymalloc(sizeof(LOGPROB)*wchmm.startnum);
  }

  /* calc prob for all startid */
  for (i=0;i<wchmm.startnum;i++) {
    node = wchmm.startnode[i];
#ifdef UNIGRAM_FACTORING
    if (wchmm.state[node].fscore != LOG_ZERO) {
      iw_sc_cache[x][i] = wchmm.state[node].fscore;
    } else {
      iw_sc_cache[x][i] = bi_prob_lr(&ngram, last_nword, (wchmm.state[node].sc)->word);
    }
#else  /* UNIGRAM_FACTORING */
    iw_sc_cache[x][i] = calc_successor_prob(last_nword, node);
#endif
  }
#ifdef HASH_CACHE_IW
  iw_lw_cache[x] = last_nword;
#endif

  return(iw_sc_cache[x]);
}




