/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* global definitions */

/* $Id: global.h,v 1.4 1998/09/02 15:47:01 ri Exp $ */

#ifndef __SENT_EXTERNAL_DEFINITION__
#define __SENT_EXTERNAL_DEFINITION__

#include <sent/stddefs.h>
#include <sent/hmm.h>
#include <sent/vocabulary.h>
#include <sent/ngram2.h>
#include "wchmm.h"

/* If GLOBAL_VARIABLE_DEFINE is defined, global variables are actually made */
/* Else, these are external definition */
#ifdef GLOBAL_VARIABLE_DEFINE
#define GLOBAL /*  */
#define GLOBAL_VAL(v) = (v)
#else
#define GLOBAL extern
#define GLOBAL_VAL(v) /*  */
#endif


/* exec option flags */
GLOBAL boolean verbose_flag GLOBAL_VAL(TRUE); /* verbose output */
GLOBAL boolean debug_flag GLOBAL_VAL(TRUE); /* debug output */
GLOBAL boolean debug2_flag GLOBAL_VAL(FALSE); /* debug2 output (numerous)*/
GLOBAL int speech_input GLOBAL_VAL(SP_MFCFILE); /* speech input */
GLOBAL int trellis_beam_width GLOBAL_VAL(-1); /* trellis beam width */
GLOBAL LOGPROB lm_weight; /* language model weight */
GLOBAL LOGPROB lm_penalty; /* language model logscore penalty */
GLOBAL LOGPROB lm_weight2; /* language model weight */
GLOBAL LOGPROB lm_penalty2; /* language model logscore penalty */
GLOBAL boolean separate_score_flag GLOBAL_VAL(FALSE); /* output score in AM/LM */
GLOBAL boolean wchmm_check_flag GLOBAL_VAL(FALSE); /* wchmm interactive check */
GLOBAL boolean trellis_check_flag GLOBAL_VAL(FALSE); /* trellis interactive check */
GLOBAL boolean result_reorder_flag GLOBAL_VAL(TRUE); /* re-order result by score */
GLOBAL int output_hypo_maxnum GLOBAL_VAL(1); /* max num of hypo to output */
#ifdef TRELLIS_VISUALIZE
GLOBAL boolean trevis_flag GLOBAL_VAL(FALSE); /* graphical output of word trellis index */
#endif
GLOBAL boolean realtime_flag GLOBAL_VAL(FALSE);
GLOBAL boolean ccd_flag GLOBAL_VAL(TRUE);
GLOBAL boolean ccd_flag_force GLOBAL_VAL(FALSE);
GLOBAL boolean progout_flag GLOBAL_VAL(FALSE);
GLOBAL boolean align_result_word_flag GLOBAL_VAL(FALSE);
GLOBAL boolean align_result_phoneme_flag GLOBAL_VAL(FALSE);
GLOBAL boolean catch_intr_flag GLOBAL_VAL(FALSE); /* TRUE when catch SIGINT */
#ifdef WPAIR
# ifdef WPAIR_KEEP_NLIMIT
GLOBAL int wpair_keep_nlimit GLOBAL_VAL(3); /* keeps only N token on wpair */
# endif
#endif
#ifdef WORST_BEAM
GLOBAL int worst_beam_width GLOBAL_VAL(30); /* -1 means no beam-lize */
#endif
#ifdef LOOKAROUND
GLOBAL int lookup_range GLOBAL_VAL(5); /* lookup range of word expansion in 2nd pass */
#endif
#ifdef SCAN_BEAM
GLOBAL LOGPROB scan_beam_thres GLOBAL_VAL(80.0); /* beam threshold rate in 2nd pass */
GLOBAL LOGPROB *framemaxscore;/* $B3F%U%l!<%`$G$N:GL`%9%3%"(B */
#endif
#ifdef DISCOUNT_BACKSCORE
GLOBAL LOGPROB discount_backscore_rate GLOBAL_VAL(1.0); /* discount rate of 1st pass score as heuristics at 2nd pass */
#endif
#ifdef HASH_CACHE_IW
GLOBAL int iw_cache_rate GLOBAL_VAL(10);	/* inter-word LM cache size rate for vocabulary size */
#endif
#ifdef SEPARATE_BY_UNIGRAM
GLOBAL int separate_wnum GLOBAL_VAL(DEFAULT_SEPARATE_WNUM); /* word num from best freq. to separate from tree */
#endif

/* function entry point */
GLOBAL LOGPROB(*outprob)(int t, HTK_HMM_State *hmm, HTK_Param *p);
GLOBAL void (*mixture_cache_init)(HTK_HMM_INFO *hmminfo, int max_n);
GLOBAL void (*mixture_cache_prepare)(int framenum);
GLOBAL void (*calc_all_mixtures_and_cache)(int t, HTK_Param *param, HTK_HMM_INFO *hmminfo);

GLOBAL int adinnet_sd GLOBAL_VAL(-1); /* socket for adinserv */
GLOBAL int adinnet_asd GLOBAL_VAL(-1); /* socket for adinserv */



/* for sentence recognition */

/* phoneme HMM */
GLOBAL HTK_HMM_INFO hmminfo;

/* LM model */
GLOBAL WORD_INFO winfo;		/* name, syllables of vocaburary */
GLOBAL NGRAM_INFO ngram;	/* N-gram language model */


/* word-conjunction HMM */
GLOBAL WCHMM_INFO wchmm;	/* word-conjunction HMM */

GLOBAL BACKTRELLIS backtrellis; /* heuristic infos */
GLOBAL LOGPROB backmax;

/* 1st pass result (reverse order) */
GLOBAL WORD_ID pass1_wseq[MAXSEQNUM]; /* word seq. of best hypo on 1st pass */
GLOBAL int pass1_wnum; /* word num of above */
GLOBAL LOGPROB pass1_score;	/* score of above */

/* for internal notification */
GLOBAL boolean beam_token_is_expanded GLOBAL_VAL(FALSE);

/* abbrev for 2nd pass routines */
GLOBAL int peseqlen;

/* global macros */

#endif /* __SENT_EXTERNAL_DEFINITION__ */
