/* Copyright(c)1998-2000 Information-technology Promotion Agency, Japan (IPA)*/
/*   All rights reserved   */

/*   The code is developed by Doshita-lab Speech Group, Kyoto University  */

/* main.c --- main routine */

#define GLOBAL_VARIABLE_DEFINE	/* actually make global vars in global.h */
#include <julius.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#ifdef HAVE_LIBREADLINE
#include <readline/readline.h>
#include <readline/history.h>
#endif


/* -------- run-time option variables ------------------------------------- */
static char *hmmfilename = NULL; /* hmm file */
static char *mapfilename = NULL; /* HMM logical mapping file */
static char *dictfilename = NULL;	/* dictionary file */
static char *ngram_filename = NULL; /* n-gram file */
static char *ngram_filename_lr_arpa = NULL; /* n-gram file in ARPA format(LR)*/
static char *ngram_filename_rl_arpa = NULL; /* n-gram file in ARPA format(LR)*/
static boolean lmp_specified = FALSE; /* true if -lmp explicitly specified */
static boolean lmp2_specified = FALSE; /* true if -lmp2 explicitly specified */
static LOGPROB def_lm_weight_mono; /* default LM weights and penalties */
static LOGPROB def_lm_penalty_mono;
static LOGPROB def_lm_weight2_mono;
static LOGPROB def_lm_penalty2_mono;
static LOGPROB def_lm_weight_tri;
static LOGPROB def_lm_penalty_tri;
static LOGPROB def_lm_weight2_tri;
static LOGPROB def_lm_penalty2_tri;
#ifdef USE_NETAUDIO
static char *netaudio_devname = "localhost:0";
#endif
#ifdef USE_MIC
static int level_thres = 3000;
#endif
static int mixnum_thres = 2;
#ifdef GPRUNE_DEFAULT_SAFE
static int gprune_method = GPRUNE_SEL_SAFE;
#elif GPRUNE_DEFAULT_HEURISTIC
static int gprune_method = GPRUNE_SEL_HEURISTIC;
#elif GPRUNE_DEFAULT_BEAM
static int gprune_method = GPRUNE_SEL_BEAM;
#endif

/* search parameters */
char *head_silname = BEGIN_WORD_DEFAULT; /* sil model name to begin search */
char *tail_silname = END_WORD_DEFAULT; /* sil model name to end search */
#define DEFAULT_STACK_SIZE 500	/* default stack size for search */
static int stack_size = DEFAULT_STACK_SIZE;
static int hypo_overflow = 2000;
static boolean compute_only_1pass = FALSE; /* if TRUE, compute only 1pass */
static int nbest;		/* search till candidate reaches this number */

/* input data variables */
static SP16 speech[MAXSPEECHLEN]; /* speech data */
static int speechlen;		/* speech length */
static char *speechfilename;	/* filename of speech data (or param file) */
HTK_Param *param;		/* parameter vector */
static boolean paramtype_check_flag = TRUE; /* with "-notypecheck", don't check param type */
/*char *param_kind = NULL;*/	/* specify recognition-time parameter def */
static boolean force_realtime_flag = FALSE; /* whether to force */
static boolean forced_realtime = FALSE;	/* forced value */
static boolean strip_zero_sample = TRUE; /* strip off zero samples */
#ifdef HAVE_PTHREAD
/* adin will be threaded */
#include <pthread.h>
static boolean transfer_online = FALSE;	/* control flag for adin-cut thread */
static pthread_t adin_thread;	/* thread ID */
static ADIN_THREAD_P adin_thread_arg; /* argument to pass adin-cut-thread */
#endif

void check_specs();	/* check needed parameter */
void print_setting();		/* output current setting */
void print_info();		/* output current gammar info */
static void opt_parse(int argc, char *argv[], char *cwd); /* option parsing */


/* ---------- utility functions -----------------------------------------*/
/* int signal handler */
static RETSIGTYPE
my_sig_int()
{
  catch_intr_flag = TRUE;
  j_printerr("** catch sigint **\n");
}
/* start trap */
static RETSIGTYPE
sigint_trap()
{
  catch_intr_flag = FALSE;
  if (signal(SIGINT, my_sig_int) == SIG_ERR) {
    j_printerr("Warning: cannot set signal handler\n");
  }
}
/* end trap */
static RETSIGTYPE
sigint_untrap()
{
  if (signal(SIGINT, SIG_DFL) == SIG_ERR) {
    j_printerr("Warning: cannot set signal handler\n");
  }
}

/* get 1 line input */
/* return value: newly allocated buffer */
/* repeat if no input, and */
/* returns NULL on EOF */
char *
get_line(char *prompt)
{
  char *buf = NULL;
  char *p;

#ifdef HAVE_LIBREADLINE
  /* get a line */
  do {
    if (buf) {
      free(buf);
      buf = NULL;
    }
    buf = readline(prompt);
    if (buf && *buf) {		/* any input */
      add_history(buf);		/* save it on the history */
    }
    if (buf == NULL) return(NULL); /* return NULL on EOF */
  } while (!buf[0]);	/* loop on newline */
#else
  buf = (char *)mymalloc(500);
  do {
    j_printerr("%s",prompt);
    if (fgets(buf, 500, stdin) == NULL) {
      free(buf);
      return(NULL);
    }
  } while (!buf[0]);		/* loop till some input */
  /* chop last newline */
  p = buf + strlen(buf);
  while (*(--p) == '\n') *p = '\0';
#endif
  /* chop last space */
  p = buf + strlen(buf);
  while (*(--p) == ' ') *p = '\0';

  return(buf);
}

	  


/* --------------------- usage and help --------------------------------*/
static char *execname;		/* exec name of this process */
/* terminate on option error */
static void
opt_terminate()
{
  j_printerr("Try `-help' for more information.\n", execname);
  exit(1);
}

/* when no argument specified */
static void
usage()
{
  put_header(stderr);
  opt_terminate();
}

/* when more arguments required */
static char *
args_needed_exit(char *opt)
{
  j_printerr("%s: option requires argument -- %s\n", PRODUCT, opt);
  opt_terminate();
  return(NULL);
}

/* detailed info */
static char *
detailed_help()
{
  put_header(stdout);
  put_compile_defs(stdout);
  j_printf("\nOptions:\n", execname);
  
  j_printf("  Language model (\"-d\" or \"-nlr -nrl\" needed):\n");
  j_printf("    -d file.bingram     n-gram file name in Julius binary format\n");
  j_printf("    -nlr file.arpa      2-gram file name in ARPA format\n");
  j_printf("    -nrl file.arpa      reverse 3-gram file name in ARPA format\n");
  j_printf("    [-lmp float float]  LM weight and penalty (tri: %.1f %.1f mono: %.1f %1.f)\n", def_lm_weight_tri, def_lm_penalty_tri, def_lm_weight_mono, def_lm_penalty_mono);
  j_printf("    [-lmp2 float float]          for 2nd pass (tri: %.1f %.1f mono: %.1f %1.f)\n", def_lm_weight2_tri, def_lm_penalty2_tri, def_lm_weight2_mono, def_lm_penalty2_mono);

  j_printf("  Acoustic model:\n");
  j_printf("    -h hmmdefsfile      HMM definition file name\n");
  j_printf("    [-hlist HMMlistfile] HMMlist filename (must for triphone model)\n");
  j_printf("    [-tmix gaussnum]    (TM/PTM) calculated Gaussian num per codebook (2)\n");
  j_printf("    [-gprune methodname](TM/PTM) Gaussian pruning method        (safe)\n");
  j_printf("             safe          safe pruning\n");
  j_printf("             heuristic     heuristic pruning\n");
  j_printf("             beam          dimension-independent beam pruning\n");
  j_printf("    [-notypecheck]      don't check input parameter type\n");
  j_printf("    [-no_ccd]           don't handle IWCD\n");
  j_printf("    [-force_ccd]        force to handle IWCD\n");

  j_printf("  Dictionary:\n");
  j_printf("    -v dictfile         dictionary file name\n");
  j_printf("    [-silhead wordname] specify the beginning-of-sentence word    (%s)\n", head_silname);
  j_printf("    [-siltail wordname] specify the end-of-sentence word          (%s)\n", tail_silname);

  j_printf("  Search parameters -- 1st pass:\n");
  j_printf("    [-b beamwidth]      beam width (by state num)             (guessed)\n");
#ifdef WPAIR
# ifdef WPAIR_KEEP_NLIMIT
  j_printf("    [-nlimit N]         keeps only N tokens on each state     (%d)\n", wpair_keep_nlimit);
# endif
#endif
#ifdef SEPARATE_BY_UNIGRAM
  j_printf("    [-sepnum wordnum]   # of hi-freq word isolated from tree  (%d)\n", separate_wnum);
#endif
#ifdef HASH_CACHE_IW
  j_printf("    [-iwcache percent]  amount of inter-word LM cache         (%3d)\n", iw_cache_rate);
#endif
  j_printf("    [-1pass]            do 1st pass only, omit 2nd pass\n");
  j_printf("    [-force_realtime on] force on-the-fly decoding even for a file\n");
  j_printf("    [-force_realtime off] don't decode on-the-fly even for mic\n");
  
  j_printf("  Search parameters -- 2nd pass:\n");
#ifdef WORST_BEAM
  j_printf("    [-b2 hyponum]       word envelope beam width (by hypo num) (%d)\n",worst_beam_width);
#endif
  j_printf("    [-n N]              # of sentence to find                 (%d)\n", nbest);
  j_printf("    [-output N]         # of sentence to output               (%d)\n",output_hypo_maxnum);
#ifdef SCAN_BEAM
  j_printf("    [-sb score]         score beam threshold (by score)       (%.1f)\n", scan_beam_thres);
#endif
  j_printf("    [-s hyponum]        global stack size of hypotheses       (%d)\n", stack_size);
  j_printf("    [-m hyponum]        hypotheses overflow threshold num     (%d)\n", hypo_overflow);

#ifdef LOOKAROUND 
  j_printf("    [-lookuprange N]    frame lookup range in word expansion  (%d)\n", lookup_range);
#endif
#ifdef DISCOUNT_BACKSCORE
  j_printf("    [-discount float]   heuristic discount rate               (%f)\n", discount_backscore_rate);
#endif

  j_printf("  Speech Input:\n");
  j_printf("    [-input devname]    speech input from:                    (mfcfile)\n");
  j_printf("            mfcfile       HTK parameter file\n");
  j_printf("      (below are available only for MFCC_E_D(_N)_Z models)\n");
  j_printf("            rawfile       wave file (16kHz,16bit,mono,linear)\n");
  j_printf("                          in .WAV or RAW (noheader) format\n");
#ifdef USE_MIC
  j_printf("            mic           microphone device\n");
#endif
#ifdef USE_NETAUDIO
  j_printf("            netaudio      DatLink/NetAudio server\n");
#endif
  j_printf("            adinserv      adin network client (experimental)\n");
#ifdef USE_NETAUDIO
  j_printf("    [-NA host:unit]     netaudio server address\n");
#endif
#if defined(USE_MIC) || defined(USE_NETAUDIO)
  j_printf("    [-lv unsignedshort] input level threshold for mic         (%d)\n", level_thres);
#endif
  j_printf("    [-nostrip]          not strip off zero samples\n");
  j_printf("  Message level:\n");
  j_printf("    [-separatescore]    output LM and AM score independently\n");
  j_printf("    [-quiet]            reduce output to only word string\n");
  j_printf("    [-progout]          progressive output in 1st pass\n");
  j_printf("    [-demo]             equal to \"-quiet -progout\"\n");
  j_printf("    [-walign]           optionally output word alignments\n");
  j_printf("    [-palign]           optionally output phoneme alignments\n");
  j_printf("  Misc:\n");
  j_printf("    [-C jconffile]      load options from jconf file\n");
  j_printf("    [-debug]            (for debug) dump numerous log\n");
  j_printf("    [-check (wchmm|trellis)] (for debug) check internal structure\n");
  j_printf("    [-version]          print version info and exit\n");
  j_printf("    [-help]             print this message and exit\n");
#ifdef TRELLIS_VISUALIZE
  j_printf("    [-V]                visualize trellis in X11 -- not yet\n");
#endif

  exit(1);
}
/* --------------------- read jconf file (recursively) ------------------ */
/* generate file path */
static char *
filepath(char *filename, char *dirname)
{
  char *p;
  if (dirname != NULL && filename[0] != '/') {
    p = (char *)mymalloc(strlen(filename) + strlen(dirname) + 1);
    strcpy(p, dirname);
    strcat(p, filename);
  } else {
    p = filename;
  }
  return p;
}
/* file access check */
static void
checkpath(char *filename)
{
  if (access(filename, R_OK) == -1) {
    perror("Error");
    j_printerr("%s: cannot access %s\n", PRODUCT, filename);
    exit(1);
  }
}
/* get directory name of a file path, with trailing slash */
static void
get_dirname(char *path)
{
  char *p;
  /* /path/file -> /path/ */
  /* path/file  -> path/  */
  /* /file      -> / */
  /* file       ->  */
  /* ../file    -> ../ */
  p = path + strlen(path) - 1;
  while (*p != '/' && p != path) p--;
  if (p == path && *p != '/') *p = '\0';
  else *(p+1) = '\0';
}
/* parse config file and process options */
#define ISTOKEN(A) (A == ' ' || A == '\t' || A == '\n')
static void
config_file_parse(char *conffile)
{
  int c_argc;
  char **c_argv;
  FILE *fp;
  int maxnum, step;
  char buf[512], *cpy;
  char *p, *dst, *dst_from;
  char *cdir;

  j_printerr("include config: %s\n", conffile);
  
  /* file -> option string */
  if ((fp = fopen_readfile(conffile)) == NULL) {
    j_printerr("%s: failed to open jconf file: %s\n",PRODUCT, conffile);
    exit(1);
  }
  step = 5;
  maxnum = step;
  c_argv = (char **)mymalloc(sizeof(char *) * maxnum);
  c_argv[0] = strcpy((char *)mymalloc(strlen(conffile)+1), conffile);
  c_argc = 1;
  while (getl(buf, 512, fp) != NULL) {
    cpy = (char *)mymalloc(strlen(buf)+1);
    p = buf; dst = cpy;
    while (1) {
      while (*p != '\0' && ISTOKEN(*p)) p++;
      if (*p == '\0') break;
      
      dst_from = dst;
      
      while (*p != '\0' && (!ISTOKEN(*p))) {
	if (*p == 92) {	/* escape by '\' */
	  if (*(++p) == '\0') break;
	  *(dst++) = *(p++);
	} else {
	  if (*p == 34) { /* quote by "" */
	    p++;
	    while (*p != '\0' && *p != 34) *(dst++) = *(p++);
	    if (*p == '\0') break;
	    p++;
	  } else if (*p == 39) { /* quote by '' */
	    p++;
	    while (*p != '\0' && *p != 39) *(dst++) = *(p++);
	    if (*p == '\0') break;
	    p++;
	  } else if (*p == 35) { /* comment out by '#' */
	    *p = '\0';
	    break;
	  } else {		/* other */
	    *(dst++) = *(p++);
	  }
	}
      }
      if (dst != dst_from) {
	*dst = '\0'; dst++;
	if (c_argc >= maxnum) {
	  maxnum += step;
	  c_argv = (char **)myrealloc(c_argv, sizeof(char *) * maxnum);
	}
	c_argv[c_argc++] = dst_from;
      }
    }
  }
  if (fclose_readfile(fp) == -1) {
    j_printerr("%s: jconf file cannot close\n", PRODUCT);
    exit(1);
  }
  if (debug2_flag) {
    int i;
    printf("\t");
    for (i=1;i<c_argc;i++) printf(" %s",c_argv[i]);
    printf("\n");
  }

  /* parse string as argument, changing dirname */
  cdir = strcpy((char *)mymalloc(strlen(conffile)+1), conffile);
  get_dirname(cdir);
  opt_parse(c_argc, c_argv, (cdir[0] == '\0') ? NULL : cdir);
  free(cdir);

  /* should not free arg area */
}

/* ------------ argument parsing ----------------------------------------- */
static void
opt_parse(int argc, char *argv[], char *cwd)
{
  char *tmparg;
  int i;

#define NEXTARG (++i >= argc) ? (char *)args_needed_exit(argv[i-1]) : argv[i]

  if (argc == 1) {		/* no argument */
    usage();
  }
  
  for (i=1;i<argc;i++) {
    if (strmatch(argv[i],"-C")) { /* config file  */
      config_file_parse(filepath(NEXTARG, cwd));
      continue;
    } else if (strmatch(argv[i],"-input")) { /* speech input */
      tmparg = NEXTARG;
      if (strmatch(tmparg,"rawfile")) {
	speech_input = SP_RAWFILE;
	realtime_flag = FALSE;
      } else if (strmatch(tmparg,"mfcfile")) {
	speech_input = SP_MFCFILE;
	realtime_flag = FALSE;
      } else if (strmatch(tmparg,"adinserv")) {
	speech_input = SP_ADINSERV;
	realtime_flag = TRUE;
#ifdef USE_NETAUDIO
      } else if (strmatch(tmparg,"netaudio")) {
	speech_input = SP_NETAUDIO;
	realtime_flag = TRUE;
#endif
#ifdef USE_MIC
      } else if (strmatch(tmparg,"mic")) {
	speech_input = SP_MIC;
	realtime_flag = TRUE;
#endif
      } else if (strmatch(tmparg,"file")) { /* for 1.1 compat */
	speech_input = SP_RAWFILE;
	realtime_flag = FALSE;
      } else if (strmatch(tmparg,"mfc")) { /* for 1.1 compat */
	speech_input = SP_MFCFILE;
	realtime_flag = FALSE;
      } else {
	j_printerr("%s: no such speech input source \"%s\" supported\n", argv[0], tmparg);
	opt_terminate();
      }
      continue;
    } else if (strmatch(argv[i],"-force_realtime")) {
      tmparg = NEXTARG;
      if (strmatch(tmparg, "on")) {
	forced_realtime = TRUE;
      } else if (strmatch(tmparg, "off")) {
	forced_realtime = FALSE;
      } else {
	j_printerr("%s: \"-force_realtime\" allows \"on\" or \"off\"\n", PRODUCT);
	exit(1);
      }
      force_realtime_flag = TRUE;
      continue;
    } else if (strmatch(argv[i],"-check")) {
      tmparg = NEXTARG;
      if (strmatch(tmparg, "wchmm")) {
	wchmm_check_flag = TRUE;
      } else if (strmatch(tmparg, "trellis")) {
	trellis_check_flag = TRUE;
#ifdef TRELLIS_VISUALIZE
	trevis_flag = TRUE;
#endif
      } else {
	j_printerr("%s: no such check style: %s\n", argv[0], tmparg);
	opt_terminate();
      }
      continue;
    } else if (strmatch(argv[i],"-notypecheck")) {
      paramtype_check_flag = FALSE;
      continue;
    } else if (strmatch(argv[i],"-separatescore")) {
      separate_score_flag = TRUE;
      continue;
    } else if (strmatch(argv[i],"-nlimit")) {
#ifdef WPAIR_KEEP_NLIMIT
      wpair_keep_nlimit = atoi(NEXTARG);
#else
      j_printerr("Warning: option \"-nlimit\" ignored\n");
#endif
      continue;
    } else if (strmatch(argv[i],"-lookuprange")) {
#ifdef LOOKAROUND
      lookup_range = atoi(NEXTARG);
#else
      j_printerr("Warning: option \"-lookuprange\" ignored\n");
#endif
      continue;
    } else if (strmatch(argv[i],"-sb")) {
#ifdef SCAN_BEAM
      scan_beam_thres = atof(NEXTARG);
#else
      j_printerr("Warning: option \"-sb\" ignored\n");
#endif
      continue;
    } else if (strmatch(argv[i],"-discount")) {
#ifdef DISCOUNT_BACKSCORE
      discount_backscore_rate = atof(NEXTARG);
#else
      j_printerr("Warning: option \"-discount\" ignored\n");
#endif
      continue;
#if defined(USE_MIC) || defined(USE_NETAUDIO)
    } else if (strmatch(argv[i],"-lv")) {
      level_thres = atoi(NEXTARG);
      continue;
#endif
    } else if (strmatch(argv[i],"-version")) {
      put_header(stderr);
      put_compile_defs(stderr);
      exit(0);
    } else if (strmatch(argv[i],"--version")) {
      put_header(stderr);
      put_compile_defs(stderr);
      exit(0);
    } else if (strmatch(argv[i],"-quiet")) {
      debug2_flag = debug_flag = verbose_flag = FALSE;
      continue;
    } else if (strmatch(argv[i],"-debug")) {
      debug2_flag = debug_flag = verbose_flag = TRUE;
      continue;
    } else if (strmatch(argv[i],"-progout")) {
      progout_flag = TRUE;
      continue;
    } else if (strmatch(argv[i],"-demo")) {
      /* quiet + progout */
      debug2_flag = debug_flag = verbose_flag = FALSE;
      progout_flag = TRUE;
      continue;
    } else if (strmatch(argv[i],"-walign")) {
      align_result_word_flag = TRUE;
      continue;
    } else if (strmatch(argv[i],"-palign")) {
      align_result_phoneme_flag = TRUE;
      continue;
    } else if (strmatch(argv[i],"-output")) {
      output_hypo_maxnum = atoi(NEXTARG);
      continue;
    } else if (strmatch(argv[i],"-1pass")) {
      compute_only_1pass = TRUE;
      continue;
    } else if (strmatch(argv[i],"-hlist")) {
      mapfilename = filepath(NEXTARG, cwd);
      continue;
    } else if (strmatch(argv[i],"-nlr")) {
      ngram_filename_lr_arpa = filepath(NEXTARG, cwd);
      ngram_filename = NULL;
      continue;
    } else if (strmatch(argv[i],"-nrl")) {
      ngram_filename_rl_arpa = filepath(NEXTARG, cwd);
      ngram_filename = NULL;
      continue;
    } else if (strmatch(argv[i],"-lmp")) {
      lm_weight = (LOGPROB)atof(NEXTARG);
      lm_penalty = (LOGPROB)atof(NEXTARG);
      lmp_specified = TRUE;
      continue;
    } else if (strmatch(argv[i],"-lmp2")) {
      lm_weight2 = (LOGPROB)atof(NEXTARG);
      lm_penalty2 = (LOGPROB)atof(NEXTARG);
      lmp2_specified = TRUE;
      continue;
    } else if (strmatch(argv[i],"-silhead")) {
      head_silname = NEXTARG;
      continue;
    } else if (strmatch(argv[i],"-siltail")) {
      tail_silname = NEXTARG;
      continue;
    } else if (strmatch(argv[i],"-iwcache")) {
#ifdef HASH_CACHE_IW
      iw_cache_rate = atof(NEXTARG);
      if (iw_cache_rate > 100) iw_cache_rate = 100;
      if (iw_cache_rate < 1) iw_cache_rate = 1;
#else
      j_printerr("Warning: option \"-iwcache\" ignored\n");
#endif
      continue;
    } else if (strmatch(argv[i],"-sepnum")) {
#ifdef SEPARATE_BY_UNIGRAM
      separate_wnum = atoi(NEXTARG);
#else
      j_printerr("Warning: option \"-sepnum\" ignored\n");
#endif
      continue;
#ifdef USE_NETAUDIO
    } else if (strmatch(argv[i],"-NA")) {
      netaudio_devname = NEXTARG;
      continue;
#endif
    } else if (strmatch(argv[i],"-nostrip")) {
      strip_zero_sample = FALSE;
      continue;
    } else if (strmatch(argv[i],"-gprune")) {
      tmparg = NEXTARG;
      if (strmatch(tmparg,"safe")) {
	gprune_method = GPRUNE_SEL_SAFE;
      } else if (strmatch(tmparg,"heuristic")) {
	gprune_method = GPRUNE_SEL_HEURISTIC;
      } else if (strmatch(tmparg,"beam")) {
	gprune_method = GPRUNE_SEL_BEAM;
      } else {
	j_printerr("%s: no such pruning method \"%s\"\n", argv[0], tmparg);
	opt_terminate();
      }
      continue;
/* 
 *     } else if (strmatch(argv[i],"-reorder")) {
 *	 result_reorder_flag = TRUE;
 *	 continue;
 */
    } else if (strmatch(argv[i],"-no_ccd")) {
      ccd_flag = FALSE;
      ccd_flag_force = TRUE;
      continue;
    } else if (strmatch(argv[i],"-force_ccd")) {
      ccd_flag = TRUE;
      ccd_flag_force = TRUE;
      continue;
    } else if (strmatch(argv[i],"-tmix")) {
      if (i + 1 < argc && isdigit(argv[i+1][0])) {
	mixnum_thres = atoi(argv[++i]);
      }
      continue;
    } else if (strmatch(argv[i],"-b2") || strmatch(argv[i],"-bw") || strmatch(argv[i],"-wb")) {
#ifdef WORST_BEAM
      worst_beam_width = atoi(NEXTARG);
#else
      j_printerr("Warning: option \"-b2\" ignored\n");
#endif
      continue;
    } else if (strmatch(argv[i],"-help")) {
      detailed_help();
    } else if (strmatch(argv[i],"--help")) {
      detailed_help();
    }
    if (argv[i][0] == '-' && strlen(argv[i]) == 2) {
      /* 1-letter options */
      switch(argv[i][1]) {
      case 'h':
	hmmfilename = filepath(NEXTARG, cwd);
	break;
      case 'v':
	dictfilename = filepath(NEXTARG, cwd);
	break;
      case 'd':
	ngram_filename = filepath(NEXTARG, cwd);
	ngram_filename_lr_arpa = NULL;
	ngram_filename_rl_arpa = NULL;
	break;
      case 'b':
	trellis_beam_width = atoi(NEXTARG);
	break;
      case 's':
	stack_size = atoi(NEXTARG);
	break;
      case 'n':
	nbest = atoi(NEXTARG);
	break;
      case 'm':
	hypo_overflow = atoi(NEXTARG);
	break;
#ifdef TRELLIS_VISUALIZE
      case 'V':
	trevis_flag = TRUE;
	break;
#endif
      default:
	j_printerr("%s: wrong argument: %s\n", argv[0], argv[i]);
	opt_terminate();
      }
    } else {			/* error */
      j_printerr("%s: wrong argument: %s\n", argv[0], argv[i]);
      opt_terminate();
    }
  }
}

/* ------------- argument requirement check ----------------------- */
void
check_specs()  /* set parameters which was not explicitly specified */
{
  boolean ok_p;

  /* check files */
  ok_p = TRUE;
  if (hmmfilename == NULL) {
    j_printerr("Error: needs HMM definition file (-h hmmdef_file)\n");
    ok_p = FALSE;
  }
  if (dictfilename == NULL) {
    j_printerr("Error: needs dictionary file (-v dict_file)\n");
    ok_p = FALSE;
  }
  /* only LR 2-gram specified .... calculate only 1-pass */
  /* only RL 3-gram specified .... ERROR */
  if (ngram_filename == NULL) {
    if (ngram_filename_lr_arpa == NULL && ngram_filename_rl_arpa == NULL) {
      j_printerr("Error: needs word n-gram file (-d bingram | -nlr ARPA_2gram -nrl ARPA_r3gram)\n");
      ok_p = FALSE;
    } else if (ngram_filename_lr_arpa == NULL) {
      j_printerr("Error: also needs ARPA 2-gram file (-nlr ARPA_2gram_file)\n");
      ok_p = FALSE;
    } else if (ngram_filename_rl_arpa == NULL) {
      compute_only_1pass = TRUE;
    }
  }

  /* path check */
  if (hmmfilename != NULL) checkpath(hmmfilename);
  if (mapfilename != NULL) checkpath(mapfilename);
  if (dictfilename != NULL) checkpath(dictfilename);
  if (ngram_filename != NULL) checkpath(ngram_filename);
  if (ngram_filename_lr_arpa != NULL) checkpath(ngram_filename_lr_arpa);
  if (ngram_filename_rl_arpa != NULL) checkpath(ngram_filename_rl_arpa);

  /* check input mode */
  if (force_realtime_flag) {
    realtime_flag = forced_realtime;
  }

  if (!ok_p) exit(1);
}

/* return default beam width */
/* these are values tuned for speed on IPA models */
static int
default_width()
{
  if (strmatch(SETUP, "fast")) { /* for fast setup */
    if (hmminfo.is_triphone) {
      if (hmminfo.is_tied_mixture) {
	/* tied-mixture triphones (PTM etc.) */
	return(600);
      } else {
	/* shared-state triphone */
#ifdef PASS1_IWCD
	return(800);
#else
	/* v2.1 compliant (no IWCD on 1st pass) */
	return(1000);		
#endif
      }
    } else {
      /* monophone */
      return(400);
    }
  } else {			/* for standard / v2.1 setup */
    if (hmminfo.is_triphone) {
      if (hmminfo.is_tied_mixture) {
	/* tied-mixture triphones (PTM etc.) */
	return(800);
      } else {
	/* shared-state triphone */
#ifdef PASS1_IWCD
	return(1500);
#else
	return(1500);		/* v2.1 compliant (no IWCD on 1st pass) */
#endif
      }
    } else {
      /* monophone */
      return(700);
    }
  }
}

/* set beam width automatically */
static void
set_beam_width()
{
  int standard_width;
  
  if (trellis_beam_width < 0) {
    /* if not specified, try to guess from vocabulary size */
    standard_width = default_width();
    trellis_beam_width = sqrt(winfo.num) * 15;
    if (trellis_beam_width > standard_width) trellis_beam_width = standard_width;
  } else if (trellis_beam_width == 0) {
    VERMES("doing full search\n");
    VERMES("Warning: no routine for full search. use beam routine with full beam width\n");
    VERMES("Warning: this will be extremely slow\n");
    trellis_beam_width = wchmm.n;
  }
}

/* ------------- output system status ---------------------------------- */
void
print_setting()
{
#ifdef USE_NETAUDIO
  char *p;
#endif
  
  j_printf("hmmfilename=%s\n",hmmfilename);
  if (mapfilename != NULL) {
    j_printf("hmmmapfilename=%s\n",mapfilename);
  }
  j_printf("vocabulary filename=%s\n",dictfilename);
  if (ngram_filename != NULL) {
    j_printf("   n-gram  filename=%s (binary format)\n",ngram_filename);
  } else {
    j_printf(" LR 2-gram filename=%s\n",ngram_filename_lr_arpa);
    if (ngram_filename_rl_arpa != NULL) {
      j_printf(" RL 3-gram filename=%s\n",ngram_filename_rl_arpa);
    }
  }
  j_printf("speech input source: ");
  if (speech_input == SP_RAWFILE) {
    j_printf("raw file\n");
  } else if (speech_input == SP_MFCFILE) {
    j_printf("MFCC parameter file (HTK format)\n");
  } else if (speech_input == SP_ADINSERV) {
    j_printf("adin server mode\n");
#ifdef USE_NETAUDIO
  } else if (speech_input == SP_NETAUDIO) {
    j_printf("NetAudio server on ");
    if (netaudio_devname != NULL) {
      j_printf("%s\n", netaudio_devname);
    } else if ((p = getenv("AUDIO_DEVICE")) != NULL) {
      j_printf("%s\n", p);
    } else {
      j_printf("local port\n");
    }
#endif
  } else if (speech_input == SP_MIC) {
    j_printf("microphone at local host\n");
  }
#ifdef HAVE_PTHREAD
  j_printf("Using PThread for ad-in\n");
#endif
  if (strip_zero_sample) {
    j_printf("Zero frames stripping: on\n");
  } else {
    j_printf("Zero frames stripping: off\n");
  }
  
  j_printf("\n");
  if (progout_flag) j_printf("output result in progress on 1st pass.\n");
  /* if (param_kind != NULL) {
    j_printf("Selectively use input parameter vector as: %s\n", param_kind);
  } */
  if (compute_only_1pass) {
    j_printf("Compute only 1-pass\n");
  }
}

void
print_info()			/* print grammar information */
{
  j_printf("------------- System Info begin -------------\n");
  put_header(stdout);
  if (verbose_flag) {
    put_compile_defs(stdout);
    j_printf("Large Vocabulary Continuous Speech Recognition Based on N-gram\n");
  }
  
  /* print current setting to stdout */
  print_setting();
  j_printf("\n");

  print_hmmdef_info(&hmminfo); j_printf("\n");
  print_voca_info(&winfo); j_printf("\n");
  print_wchmm_info(&wchmm); j_printf("\n");
  print_ngram_info(&ngram);

  j_printf("\nN-gram cache parameters: \n");
#ifdef HASH_CACHE_IW
  j_printf("\t inter-word LM cache size = %dMB max (-iwcache %3d)\n", ((iw_cache_rate * ngram.max_word_num) / 100 ) * (wchmm.startnum + 2) / 1000 * sizeof(LOGPROB) / 1000, iw_cache_rate);
#else  /* full cache */
  j_printf("\t                     size = %dMB max\n", ngram.max_word_num * (wchmm.startnum + 2) / 1000 * sizeof(LOGPROB) / 1000);
#endif /* HASH_CACHE_IW */

  j_printf("\nSearch parameters: \n");
  j_printf("\t1st pass process method = ");
  if (force_realtime_flag) j_printf("(forced) ");
  if (realtime_flag) {
    j_printf("on-the-fly decoding with realtime CMN\n");
  } else {
    j_printf("batch decoding\n");
  }
  j_printf("\t       outprob function = ");
  if (hmminfo.is_tied_mixture) {
    j_printf("mixture-driven\n");
    j_printf("\t(-gprune)Gauss. pruning = ");
    switch(gprune_method){
    case GPRUNE_SEL_BEAM: j_printf("beam\n"); break;
    case GPRUNE_SEL_HEURISTIC: j_printf("heuristic\n"); break;
    case GPRUNE_SEL_SAFE: j_printf("safe\n"); break;
    }
  } else {
    j_printf("state-driven\n");
  }
  if (hmminfo.is_tied_mixture) {
    j_printf("\t(-tmix)   mixture thres = %d / %d\n", mixnum_thres, hmminfo.maxcodebooksize);
    
  }
  j_printf("\t(-lmp) LM weight param1 = * %2.1f %+2.1f\n", lm_weight, lm_penalty);
  j_printf("\t(-lmp2)LM weight param2 = * %2.1f %+2.1f\n", lm_weight2, lm_penalty2);
  j_printf("\t(-b) trellis beam width = %d\n", trellis_beam_width);
  j_printf("\t(-silhead)head sil word = ");
  put_voca(&winfo, winfo.head_silwid);
  j_printf("\t(-siltail)tail sil word = ");
  put_voca(&winfo, winfo.tail_silwid);
  j_printf("\t        1st pass method = ");
#ifdef WPAIR
# ifdef WPAIR_KEEP_NLIMIT
  j_printf("word-pair approx., keeping only N tokens ");
# else
  j_printf("word-pair approx. ");
# endif
#else
  j_printf("1-best approx. ");
#endif
#ifdef WORD_GRAPH
  j_printf("generating word_graph\n");
#else
  j_printf("generating indexed trellis\n");
#endif
  
  j_printf("\t(-n)search candidate num= %d\n", nbest);
  j_printf("\t(-s)  search stack size = %d\n", stack_size);
  j_printf("\t(-m)    search overflow = after %d hypothesis poped\n", hypo_overflow);
#ifdef WORST_BEAM
  if (worst_beam_width >= 0) {
    j_printf("\t(-b2)  worst beam width = %d\n", worst_beam_width);
  }
#endif
#ifdef LOOKAROUND
  j_printf("\t(-lookuprange)lookup range= %d  (tm-%d <= t <tm+%d)\n",lookup_range,lookup_range,lookup_range);
#endif
#ifdef SCAN_BEAM
  j_printf("\t(-sb)2nd scan beamthres = %.1f (in logscore)\n",scan_beam_thres);
#endif
#ifdef DISCOUNT_BACKSCORE
  j_printf("\t(-discount) h(n) weight = * %.3f from 1st pass score\n", discount_backscore_rate);
#endif
  if (ccd_flag) {
    j_printf("\t IWCD handling:\n");
#ifdef PASS1_IWCD
    j_printf("\t   1st pass: approximation (get max. prob. of same LC) \n");
#else
    j_printf("\t   1st pass: ignored\n");
#endif
#ifdef PASS2_STRICT_IWCD
    j_printf("\t   2nd pass: strict (apply when expanding hypo. )\n");
#else
    j_printf("\t   2nd pass: loose (apply when hypo. is popped and scanned)\n");
#endif
  }
  
  j_printf("\t factoring score: ");
#ifdef UNIGRAM_FACTORING
  j_printf("1-gram prob. (statically assigned beforehand)\n");
#else
  j_printf("2-gram prob. (dynamically computed while search)\n");
#endif
  if (result_reorder_flag) {
    if (nbest > output_hypo_maxnum) {
      j_printf("\t will get %d candidates, sort them and output %d best\n",
	       nbest, output_hypo_maxnum);
    } else {
      if (nbest == 1) {
	j_printf("\t will get %d candidate\n", nbest);
      } else {
	j_printf("\t will get %d candidates and sort them\n", nbest);
      }
    }	  
  }
  if (align_result_word_flag) {
    j_printf("\t output word alignments\n");
  }
  if (align_result_phoneme_flag) {
    j_printf("\t output phoneme alignments\n");
  }
  j_printf("------------- System Info end -------------\n");
}


/* ------------- adin callback only to store to speech[] ---------------- */
#ifndef HAVE_PTHREAD		/* not needed for thread version */
static int dotcount = 0;
static boolean
adin_cut_callback_store_buffer(SP16 *now, int len)
{
  if (speechlen == 0) dotcount = 0;
  if (speechlen + len > MAXSPEECHLEN) {
    j_printerr("Error: too long input (> %d samples)\n", MAXSPEECHLEN);
    return(FALSE);
  }
  memcpy(&(speech[speechlen]), now, len * sizeof(SP16));
  speechlen += len;
  /* output progress bar in dots */
  if ((++dotcount) % 3 == 1) j_printerr(".");
  return(TRUE);
}
#endif /* HAVE_PTHREAD */
				       
				       
/* ------------- The Main Routine --------------------------------------- */
int
main(int argc, char *argv[])
{
  HTK_Param *selected_param;
  boolean force_cancel_2nd_pass;

  /******************/
  /* Initialization */
  /******************/
  /* set block malloc parameter (page size etc.) */
  /*mybmalloc_init(100000);*/	/* 100KB */

#ifdef HAVE_LIBREADLINE
  /* force tilde expansion */
  {
    char buf[100];
    strcpy(buf, "set expand-tilde on");
    rl_parse_and_bind(buf);
  }
#endif

  /******************/
  /* set default parameters according to engine configuration */
  /* LM practically best values as default */
  /* monophone: independent on config */
  def_lm_weight_mono = 5.0;
  def_lm_penalty_mono = -1.0;
  def_lm_weight2_mono = 6.0;
  def_lm_penalty2_mono = 0.0;
  /* triphone */
#ifdef PASS1_IWCD
  def_lm_weight_tri = 8.0;
  def_lm_penalty_tri = -2.0;
  def_lm_weight2_tri = 8.0;
  def_lm_penalty2_tri = -2.0;
#else  /* no IWCD on 1st pass */
  def_lm_weight_tri = 9.0;
  def_lm_penalty_tri = 8.0;
  def_lm_weight2_tri = 11.0;
  def_lm_penalty2_tri = -2.0;
#endif /* PASS1_IWCD */
  if (strmatch(SETUP, "fast")) {
    nbest = 1;
    worst_beam_width = 30;
  } else {
    nbest = 10;
    worst_beam_width = 100;
  }
    
  /******************/
  /* parse options and set variables */
  execname = argv[0];
  opt_parse(argc,argv,NULL);
  /* check option specs */
  check_specs();
    
  /******************/
  /* input device initialization */
#ifdef USE_MIC
  if (speech_input == SP_MIC) {
    /* set microphne functions as adin_cut callback */
    adin_setup_adfunc(adin_mic_start, adin_mic_stop, adin_mic_read, level_thres);
    /* initialize mic port */
    if (adin_mic_standby() == FALSE) {
      j_printerr("%s: mic device init failed\n", PRODUCT);
      exit(1);
    }
  }
#endif
#ifdef USE_NETAUDIO
  if (speech_input == SP_NETAUDIO) {
    adin_setup_adfunc(adin_netaudio_start, adin_netaudio_stop, adin_netaudio_read, level_thres);
    /* connect to NetAudio server */
    if (adin_netaudio_standby(netaudio_devname) == FALSE) {
      j_printerr("Error: NetAudio init failed\n");
      exit(1);
    }
  }
#endif

#ifdef TRELLIS_VISUALIZE
  /* initialization for trellis visualization */
  if (trevis_flag) {
    TrevisInit();
  }
#endif

  /******************/
  /* reading in models */
  VERMES("###### initializations\n");
  /* set mybmalloc (block memory allocation) parameters first (optional) */
  mybmalloc_set_param();
  
  /******************/
  /* Acoustic HMM   */
  init_hmminfo(&hmminfo, hmmfilename, mapfilename);
  /* for direct speech input, only MFCC_E_D_Z(_N) 25vec is supported */
  if (speech_input != SP_MFCFILE) {
    if ((hmminfo.opt.param_type != param_str2code("MFCC_E_D_Z") || hmminfo.opt.vec_size != 26) &&
	(hmminfo.opt.param_type != param_str2code("MFCC_E_D_N_Z") || hmminfo.opt.vec_size != 25)) {
      j_printerr("Error: for direct speech input, only 25(26)-dim. MFCC_E_D(_N)_Z is supported\n");
      j_printerr("Error: with this HMM, input must be given in HTK parameter file via \"-input mfcfile\"\n");
      exit(1);
    }
  }
  if (hmminfo.is_tied_mixture && hmminfo.codebooknum <= 0) {
    j_printerr("%s: this tied-mixture model has no codebook!?\n", PRODUCT);
    exit(1);
  }
#ifdef PASS1_IWCD
  /* make left-context-dependent state set for inter-word triphone approx. */
  if (hmminfo.is_triphone) {
    j_printerr("Making left context dependent state clusters for IW-triphone handling...");
    if (make_lcdset(&hmminfo) == FALSE) {
      j_printerr("\nError: failed to make left-context-dependent state set\n");
      exit(1);
    }
    j_printerr("done\n");
    /*put_all_lcdinfo(&hmminfo);*/
  }
#endif /* PASS1_IWCD */
  if (!ccd_flag_force) {
    if (hmminfo.is_triphone) {
      ccd_flag = TRUE;
    } else {
      ccd_flag = FALSE;
    }
  }

  /******************/
  /* Word dictionary */
#ifdef MONOTREE
  /* leave winfo monophone for 1st pass lexicon tree */
  init_voca(&winfo, dictfilename, &hmminfo, TRUE);
#else 
  init_voca(&winfo, dictfilename, &hmminfo, FALSE);
#endif
  
  /******************/
  /* word N-gram */
  if (ngram_filename != NULL) {	/* binary format */
    init_ngram_bin(&ngram, ngram_filename);
  } else {			/* ARPA format */
    init_ngram_arpa(&ngram, ngram_filename_lr_arpa, ngram_filename_rl_arpa);
  }
  /* map dict item to N-gram entry */
  make_voca_ref(&ngram, &winfo);
  /* set {head,tail}_silwid */
  winfo.head_silwid = voca_lookup_wid(head_silname, &winfo);
  if (winfo.head_silwid == WORD_INVALID) { /* not exist */
    j_printerr("ERROR: head sil word \"%s\" not exist in voca", head_silname);
    exit(1);
  }
  winfo.tail_silwid = voca_lookup_wid(tail_silname, &winfo);
  if (winfo.tail_silwid == WORD_INVALID) { /* not exist */
    j_printerr("ERROR: tail sil word \"%s\" not exist in voca", tail_silname);
    exit(1);
  }
  /* set default lm parameter */
  if (!lmp_specified) {
    if (hmminfo.is_triphone) {
      lm_weight = def_lm_weight_tri;
      lm_penalty = def_lm_penalty_tri;
    } else {
      lm_weight = def_lm_weight_mono;
      lm_penalty = def_lm_penalty_mono;
    }
  }
  if (!lmp2_specified) {
    if (hmminfo.is_triphone) {
      lm_weight2 = def_lm_weight2_tri;
      lm_penalty2 = def_lm_penalty2_tri;
    } else {
      lm_weight2 = def_lm_weight2_mono;
      lm_penalty2 = def_lm_penalty2_mono;
    }
  }
  if (lmp_specified != lmp2_specified) {
    j_printerr("Warning: only -lmp or -lmp2 specified, LM weights may be unbalanced\n");
  }

  /* finished reading models */

  /* build tree lexicon (word-conjunction HMM) for first pass */
  build_wchmm2();
  
  /* initialize cache for factoring */
  max_successor_cache_init();
  /* guess beam width from models, when not specified */
  set_beam_width();

#ifdef MONOTREE
  /* after building tree lexocon, */
  /* convert monophone to triphone in winfo for 2nd pass */
  if (hmminfo->is_triphone) {
    j_printerr("convert monophone dictionary to word-internal triphone...");
    if (voca_mono2tri(&winfo, &hmminfo) == FALSE) {
      j_printerr("failed\n");
      exit(1);
    }
    j_printerr("done\n");
  }
#endif
  
  /* backtrellis initialization */
  bt_init(&backtrellis);
  /* select outprob function */
  outprob_init(&hmminfo);
  if (hmminfo.is_tied_mixture) {
    outprob = outprob_tm;
    switch(gprune_method){
    case GPRUNE_SEL_BEAM:
      mixture_cache_init = mixture_cache_init_beam;
      mixture_cache_prepare = mixture_cache_prepare_beam;
      calc_all_mixtures_and_cache = calc_all_mixtures_and_cache_beam;
      break;
    case GPRUNE_SEL_HEURISTIC:
      mixture_cache_init = mixture_cache_init_heuristic;
      mixture_cache_prepare = mixture_cache_prepare_heuristic;
      calc_all_mixtures_and_cache = calc_all_mixtures_and_cache_heuristic;
      break;
    case GPRUNE_SEL_SAFE:
    default:
      mixture_cache_init = mixture_cache_init_safe;
      mixture_cache_prepare = mixture_cache_prepare_safe;
      calc_all_mixtures_and_cache = calc_all_mixtures_and_cache_safe;
      break;
    }
    mixture_cache_init(&hmminfo, mixnum_thres);
  } else {
    outprob = outprob_ss;
  }

  if (realtime_flag) {
    /* prepare for 1st pass pipeline processing */
    RealTimeInit();
  }

  VERMES("All init successfully done\n");

  /* print out all setting informations */
  print_info();

#ifdef USE_MIC
  if (speech_input == SP_MIC) {
    j_printf("\n");
    j_printf("\t******************************************************************\n");
    j_printf("\t* NOTICE: Please speak \"mic test\" (or any short term) at first,*\n");
    j_printf("\t*         and don't care about the result.                       *\n");
    j_printf("\t* see README.MIC for details.                                    *\n");
    j_printf("\t******************************************************************\n");
  }
#endif  
  if (speech_input == SP_ADINSERV) {
    if ((adinnet_sd = ready_as_server(ADINNET_PORT)) < 0) {
      j_printf("adin server setup error\n");
      exit(1);
    }
    if (adinnet_asd >= 0) close(adinnet_asd);
    j_printerr("waiting...\n");
    adinnet_asd = accept_from(adinnet_sd);
  }

#ifdef HAVE_PTHREAD
#if defined(USE_MIC) || defined(USE_NETAUDIO)
  if (speech_input == SP_MIC || speech_input == SP_NETAUDIO) {
    /* create thread for adin-cut */
    adin_thread_arg.speech = speech;
    adin_thread_arg.len = &speechlen;
    adin_thread_arg.maxlen = MAXSPEECHLEN;
    adin_thread_arg.stflag = strip_zero_sample;
    adin_thread_arg.tflag = &transfer_online;
    pthread_create(&adin_thread, NULL, (void *)adin_cut_thread, (void *)&adin_thread_arg);
    pthread_detach(adin_thread); /* not join, run forever */
    j_printf("AD-in thread created\n");
  }
#endif
#endif /* HAVE_PTHREAD */
  
  /***************************/
  /***************************/
  /** Main Recognition Loop **/
  /***************************/
  /***************************/

  sigint_trap();		/* trap SIGINT till next prompt */
  for (;;) {
    
    j_printf("\n");
    if (verbose_flag) j_printf("------\n");
    j_flushprint();

    force_cancel_2nd_pass = FALSE;
    
    /**************************************************/
    /* Read speech input and speech signal processing */
    /**************************************************/
    switch(speech_input) {

    case SP_RAWFILE:
    case SP_MIC:
    case SP_ADINSERV:
#ifdef USE_NETAUDIO
    case SP_NETAUDIO:
#endif

      /*******************************************/
      /* read waveform input from various source */
      /*******************************************/
      VERMES("### read waveform input\n");
      
      switch(speech_input) {
      case SP_RAWFILE:		/* speech waveform file */

	speechfilename = get_line("enter filename->");
	if (speechfilename == NULL) { /* end of file */
	  /* end loop */
	  return(0);
	}
	if (debug_flag) j_printf("\ninput speechfile: %s\n",speechfilename);
	if (rddata(speechfilename,speech,&speechlen,MAXSPEECHLEN) != 0) {
	  j_printerr("Error in reading speech data: \"%s\"\n",speechfilename);
	  continue;
	}

	if (strip_zero_sample) {
	  int newlen;
	  newlen = strip_zero(speech, speechlen);
	  if (newlen != speechlen) speechlen = newlen;
	}

	free(speechfilename);
	if (realtime_flag) {
	  /* use pipeline-1st pass function */
	  RealTimePipeLinePrepare();
	  RealTimePipeLine(speech, speechlen);
	}
	break;

      case SP_MIC:		/* direct audio input from mic */
      case SP_NETAUDIO:		/* waveform via TCP/IP (netaudio server) */
	
#if defined(USE_MIC) || defined(USE_NETAUDIO)
	sigint_untrap();	/* sigint will terminate process while speech input */
	j_printerr("<<< please speak >>>\n");
#ifdef HAVE_PTHREAD
	/* get source on another thread */
	/* the incoming samples are dynamically stored to speech[],
	   and the current length is stored on speechlen.  */
	/* When adin thread found long pause, the transfer_online is
	   set to FALSE */
	if (realtime_flag) {
	  int oldlen, tmplen;
	  RealTimePipeLinePrepare();
	  speechlen = oldlen = 0;
	  transfer_online = TRUE; /* tell adin-mic thread to start recording */
	  /*printf("main: transfer turned on\n");*/
	  while(1) {
	    tmplen = speechlen;
	    if (oldlen < tmplen) {
	      /*printf("main: read %d-%d\n", oldlen, tmplen);*/
	      RealTimePipeLine(&(speech[oldlen]), tmplen - oldlen);
	      oldlen = tmplen;
	    } else if (transfer_online == FALSE) {
	      /*printf("main: transfer off detected\n");*/
	      break;
	    }
	  }
	} else {
	  int oldlen, tmplen;
	  speechlen = 0;
	  transfer_online = TRUE; /* tell adin-mic thread to start recording */
	  /* only wait and let adin-thread store sample to speech */
	  while((tmplen = speechlen) != oldlen || transfer_online == TRUE) {
	    if (oldlen != tmplen) { /* thread is recording now */
	      oldlen = tmplen;
	      j_printerr(".");
	    }
	    usleep(100000);	/* wait = 0.1sec*/
	  }
	}
#else  /* ~HAVE_PTHREAD */
	if (realtime_flag) {
	  RealTimePipeLinePrepare();
	  adin_cut(RealTimePipeLine, strip_zero_sample);
	} else {
	  speechlen = 0;
	  adin_cut(adin_cut_callback_store_buffer, strip_zero_sample);
	}
	sigint_trap();
#endif /* HAVE_PTHREAD */
#else
	j_printerr("sorry, mic suppor is not compiled in.\n");
	exit(1);
#endif
	break;
	
      case SP_ADINSERV:		/* waveform via TCP/IP (adinserv) */
	{
	  int step;

	  j_printerr("listening:\n");
	  sigint_untrap();
	  speechlen = 0;
	  RealTimePipeLinePrepare();
	  for(;;) {
	    rd(adinnet_asd, (char *)&(speech[speechlen]),&step,(MAXSPEECHLEN-speechlen)*sizeof(SP16));
	    if (step == 0) {	/* break-of-speech ack */
	      break;		/* end 1st pass and proceed to 2nd pass */
	    } else if (step == -1) { /* cancel input ack */
	      force_cancel_2nd_pass = TRUE; /* terminate 1st pass and return to prompt */
	      break;
	    }
	    RealTimePipeLine(&(speech[speechlen]), step / sizeof(SP16));
	    speechlen += step / sizeof(SP16);
	  }
	  sigint_trap();
	}
	break;
      }

	
      if (realtime_flag) {
	/* end procedure of 1st-pass pipeline */
	param = RealTimeParam(&backmax);
	/* 1st pass ended, jump to 2nd pass */
	goto end_1pass;
      }
      j_printf("%d samples (%.2f sec.)\n", speechlen, (float)speechlen / (float)16000.0);
      
      /******************************************/
      /* acoustic analysis and encoding process */
      /******************************************/
      VERMES("### speech analysis (waveform -> MFCC)\n");
      /* subtract zero mean to remove DC offset */
      /* sub_zmean(speech, speechlen); */
      /* wav2mfcc() has no ZMEANSOURCE */
      param = new_wav2mfcc(speech, speechlen);

      break;
      
    case SP_MFCFILE:		/* read HTK Parameter file */
      
      /*******************************************/
      /* read pre-analyzed parameter file (MFCC) */
      /*******************************************/
      VERMES("### read analyzed parameter\n");
      
      /* from MFCC parameter file (in HTK format) */
      speechfilename = get_line("enter MFCC filename->");
      if (speechfilename == NULL) {
	/* end */
	return(0);
      }
      if (debug_flag) j_printf("\ninput MFCC file: %s\n",speechfilename);
      /* read parameter file */
      param = new_param();
      if (rdparam(speechfilename, param) == FALSE) {
	j_printerr("error in reading MFCC file: %s\n",speechfilename);
	free(speechfilename);
	free_param(param);
	continue;
      }
      /* check and strip invalid frames */
      if (strip_zero_sample) {
	param_strip_zero(param);
      }
      
      free(speechfilename);
      break;
    }

    /* adjust parameters to HMM if necessary */
    if (verbose_flag) {
      put_param_info(param);
    }
    /* parameter type check */
    if (paramtype_check_flag) {
      /* return param itself or new malloced param */
      selected_param = new_param_check_and_adjust(&hmminfo, param, verbose_flag);
      if (selected_param == NULL) { /* failed */
	free_param(param);
	continue;
      }
      param = selected_param;
    }

    
    /****************************************************/
    /* 1-pass --- backward search to compute heuristics */
    /****************************************************/
    sigint_trap();		/* trap SIGINT till next prompt */
    VERMES("### Recognition: 1st pass (LR beam with 2-gram)\n");
/* 
 * #ifdef WPAIR
 *     VERMES("with word-pair approx. ");
 * #else
 *     VERMES("with 1-best approx. ");
 * #endif
 *     VERMES("generating ");
 * #ifdef WORD_GRAPH
 *     VERMES("word graph\n");
 * #else
 *     VERMES("back trellis\n");
 * #endif
 */

    if (!realtime_flag) {
      /* free old and allocate outprob cache for needed length */
      if (hmminfo.is_tied_mixture) {
	mixture_cache_prepare(param->samplenum);
      }
      outprob_cache_prepare(param->samplenum);
      outprob_cache_clear(param->samplenum);
    }

#ifdef TRELLIS_VISUALIZE
    /* prepare for drawing trellis */
    if (trevis_flag) TrevisPrepare();
#endif 

    /* compute left-to-right backtrellis */
    get_back_trellis(param,&backtrellis,&backmax);

  end_1pass:

    /* if [-1pass] is specified or input canceled by adin source, 
       terminate search here */
    if (compute_only_1pass) {
      free_param(param);
      continue;
    }
    if (force_cancel_2nd_pass) {
      j_printerr("input canceled by adinserv\n");
      free_param(param);
      continue;
    }
    /* if backtrellis end with bad status, terminate */
    if (backmax == LOG_ZERO) {
      /* j_printerr("Terminate 2nd pass.\n"); */
      putsentence((NODE *)NULL, 1);/* print NULL result */
      free_param(param);
      continue;
    }

    /* 2-pass preparation */
#ifdef SCAN_BEAM
    /* prepare score envelope for 2nd pass */
    framemaxscore = (LOGPROB *)mymalloc(sizeof(LOGPROB)*backtrellis.framelen);
    envl_init(backtrellis.framelen);
#ifdef SCAN_BEAM_2
    /* make score envelope from 1st pass result */
    envl_premake(&backtrellis, backmax);
#endif /* SCAN_BEAM_2 */
#endif /* SCAN_BEAM */
    
    /* discount outprob of heuristic score for each frame */

#ifndef PASS2_STRICT_IWCD
    bt_discount_pescore(&backtrellis, param);
#endif
    

    /*********************************************/
    /* 2-pass --- forward search with heuristics */
    /*********************************************/
    VERMES("### Recognition: 2nd pass (RL heuristic best-first with 3-gram)\n");

    /* call stack-decoding module with N-gram decoding functions */
    wchmm_fbs(param, &backtrellis, backmax, stack_size,
	      nbest, hypo_overflow,				
              ngram_firstwords,
              ngram_nextwords,
              ngram_acceptable,
              ngram_eosscore
	      );

    free_param(param);
#ifdef SCAN_BEAM
    free(framemaxscore);
#endif
  }

}

/* for object link error by WORD_GRAPH defined */
#ifdef WORD_GRAPH
void verify_WORD_GRAPH_define(){}
#else
void verify_WORD_GRAPH_undefine(){}
#endif
#ifdef WPAIR
void verify_WPAIR_define(){}
#else
void verify_WPAIR_undefine(){}
#endif
