/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* ngram_decode.c --- N-gram decoder for search */

/* $Id: ngram_decode.c,v 1.1.1.1 1998/07/16 07:40:24 ri Exp $ */

/* ---------------------------------------------------------------------- */
/* $BC5:wMQ8@8l@)Ls%k!<%A%s72(B for N-gram */
/* wchmm_fbs$B$X$N0z?t$H$7$F4X?t$X$N%]%$%s%?$rEO$9(B*/
/*	      ngram_firstwords() $B=i4|2>@b$rJV$9(B */
/*	      ngram_nextwords()	 $B2>@b$KBP$9$k<!C18l$rJV$9(B */
/*	      ngram_acceptable() $B2>@b$,<uM}2DG=$+H]$+$rJV$9(B */
/*	      ngram_eosscore()	 $B2>@b$KBP$9$k:G8e$N%9%3%"D4@0CM$rJV$9(B */

/*
 * $B8@8l@)Ls%k!<%A%s72$N8@8l;EMM(B:
 * int gm_get_first_words(NEXTWORD **nw, int peseqlen);
 * int gm_get_next_words(NODE *hypo, NEXTWORD **nw);
 * boolean gm_acceptable(NODE *hypo);
 * LOGPROB gm_get_eos_score(NODE *hypo);
 */


#include <julius.h>


/* quick hack */
#define ISKUTEN(A)	((A) < winfo.tail_silwid)
#define ISTOUTEN(A)	((A) > winfo.tail_silwid && (A) < winfo.head_silwid) 



#ifdef LOOKAROUND
static int
compare_nw(NEXTWORD **a, NEXTWORD **b)
{
  if ((*a)->id > (*b)->id) return 1;
  if ((*a)->id < (*b)->id) return -1;
  return 0;
}
static NEXTWORD *
search_nw(NEXTWORD **nw, WORD_ID w, int num)
{
  int left,right,mid;
  NEXTWORD *tmp;

  if (num == 0) return NULL;
  left = 0;
  right = num - 1;
  while (left < right) {
    mid = (left + right) / 2;
    if ((nw[mid])->id < w) {
      left = mid + 1;
    } else {
      right = mid;
    }
  }
  tmp = nw[left];
  if (tmp->id == w) {
    return tmp;
  } else {
    return NULL;
  }
}
#endif /* LOOKAROUND */

static int
pick_backtrellis_words(NEXTWORD **nw, int oldnum, NODE *hypo, short t)
{
  int i;
  WORD_ID w;
  int nword1=0, nword2=0;
  LOGPROB rawscore;
#ifdef WPAIR
  int w_old = WORD_INVALID;
#endif
  int num;

  num = oldnum;
  /* $BC18l%;%C%H$r(B backtrellis $B$+$iF@$k(B */
  if (hypo->seqnum >= 1) {
    nword1 = winfo.wton[hypo->seq[hypo->seqnum-1]];
  }
  if (hypo->seqnum >= 2) {
    nword2 = winfo.wton[hypo->seq[hypo->seqnum-2]];
  }
  for (i=0;i<backtrellis.num[t];i++) {
    w = (backtrellis.rw[t][i])->wid;
#ifdef WORD_GRAPH
    /* $BC18l%0%i%U$N0lIt$G$J$1$l$PE83+$7$J$$(B */
    if (!(backtrellis.rw[t][i])->within_wordgraph) continue;
    /* $B2>@b$N:G8e$N(BTRELLIS_ATOM$B$N?F(BTRELLIS_ATOM$B$HF1$8(BID$B$@$1E83+(B */
    if (w != ((hypo->tre)->last_tre)->wid) continue;
#else /* not WORD_GRAPH */
#ifdef WPAIR
    /* $B%H%l%j%9$+$DC18lBP$N;~$O!$F1$8C18l$OE83+$7$J$$(B */
    /* $B%(%s%H%j$OC18l(BID$B$G%=!<%H:Q$_(B->$BO"B3$7$F=P8=(B */
    /* $B8e$G(Bnext_word()$B$G:GL`$N$b$N$H@\B3$5$l$k(B */
    if (w == w_old) continue;
    else w_old = w;
#endif /* WPAIR */
#endif /* not WORD_GRAPH */
#ifdef LOOKAROUND
    if (search_nw(nw, w, oldnum) != NULL) continue;
#endif
    switch(hypo->seqnum) {
    case 0:			/* unigram */
      rawscore = uni_prob(&ngram, winfo.wton[w]);
      break;
    case 1:			/* bigram */
      rawscore = bi_prob_rl(&ngram, winfo.wton[w], nword1);
      break;
    default:			/* trigram */
      rawscore = tri_prob_rl(&ngram, winfo.wton[w], nword1, nword2);
      break;
    }
    nw[num]->tre   = backtrellis.rw[t][i];
    nw[num]->id    = w;
    nw[num]->lscore = lm_weight2 * rawscore + lm_penalty2;
    /* j_printf("%d: %s added\n", num, winfo.wname[nw[num]->id]); */
    num++;
  }
  return num;
}

#ifdef WORD_GRAPH
  /* $BC18l%0%i%U$H$7$F;D$C$?C18l%H%l%j%9(B(within_wordgraph$B%^!<%/IU$-(B)
   * $B$N$_E83+$9$k!%A08e(B 10$B%U%l!<%`$r8+$F!$$=$NCf$K$"$kC18l$rE83+8uJd$H$9$k(B */
#endif
/* $BJV$jCM(B:$BC18l?t(B*/
int
get_backtrellis_words(
     NEXTWORD **nw,		/* $B<!C18l9=B$BN(B($BJV$jCM(B) */
     NODE *hypo,		/* $BD>A0$N2>@b(B */
     short tm, short t_end)			/* $B;~4V(B */
{
  int num = 0;
#ifdef LOOKAROUND
  int t;
  int oldnum=0;
#endif

  if (tm < 0) return(0);

#ifdef LOOKAROUND
  /* $BA0J}8~$X(B */
  for(t = tm; t >= tm - lookup_range; t--) {
    if (t < 0) break;
    num = pick_backtrellis_words(nw, oldnum, hypo, t);
    if (num > oldnum) {
      qsort(nw, num, sizeof(NEXTWORD *),
	    (int (*)(const void *,const void *))compare_nw);
      oldnum = num;
    }
  }
  /* $B8e$mJ}8~$X(B */
  for(t = tm + 1; t < tm + lookup_range; t++) {
    if (t > backtrellis.framelen - 1) break;
    if (t >= t_end) break;
    num = pick_backtrellis_words(nw, oldnum, hypo, t);
    if (num > oldnum) {
      qsort(nw, num, sizeof(NEXTWORD *),
	    (int (*)(const void *,const void *))compare_nw);
      oldnum = num;
    }
  }
#else  /* no LOOKAROUND */
  num = pick_backtrellis_words(nw, 0, hypo, tm);
#endif

  return num;
}


/* $BE83+C18l$K@)8B$r@_$1$k(B */
/* $BJV$jCM(B:$B;D$j$NC18l?t(B*/
int
limit_nw(NEXTWORD **nw, NODE *hypo, int num)
{
  int src,dst;
  int newnum;

  /* <s>$B$+$i$O2?$bE83+$7$J$$(B */
  if (hypo->seq[hypo->seqnum-1] == winfo.head_silwid) {
    return(0);
  }

  dst = 0;
  for (src=0; src<num; src++) {
    if (nw[src]->id == winfo.tail_silwid) {
      /* </s> $B$OE83+$7$J$$(B */
      continue;
#if 0
    } else if (ISKUTEN(nw[src]->id)) {
      /* $B6gE@7O(BNS$BC18l$NE83+$O:G=i$N$_(B */
      if (hypo->seqnum > 1) continue;
    } else if (ISTOUTEN(nw[src]->id)) {
      /* $BFIE@7O(BNS$BC18l$NE83+$OESCf$N$_(B */
      if (hypo->seqnum <= 1) continue;
    } else if (ISTOUTEN(hypo->seq[hypo->seqnum-1])) {
      /* <s> TOUTEN $B$H$J$k2>@b$OE83+$7$J$$(B */
      if (nw[src]->id == winfo.head_silwid) continue;
#endif
    }
    if (src != dst) memcpy(nw[dst], nw[src], sizeof(NEXTWORD));
    dst++;
  }
  newnum = dst;

  return newnum;
}
	


/* $B:G=i$NC18l$rEO$9!%(B */
/* $BJV$jCM(B: $BC18l?t(B*/
int
ngram_firstwords(NEXTWORD **nw, int peseqlen)
{
#ifdef WORD_GRAPH
  int last_time;
#endif

  /* $B=i4|2>@b$O(B $BC18l$NKvHx$NL52;C18l(B(=winfo.tail_silwid) */
  /* score$B$K$O(Bunigram$B$rF~$l$k(B */
  nw[0]->id = winfo.tail_silwid;
  /*nw[0]->lscore =  lm_weight2 * uni_prob(&ngram, winfo.wton[nw[0]->id]) + lm_penalty2;*/
#ifdef FIX_PENALTY
  nw[0]->lscore = 0.0;
#else
  nw[0]->lscore = lm_penalty2;
#endif
#ifdef WORD_GRAPH
  /* $B=i4|2>@b$N(BTRELLIS_ATOM$B$r$3$3$G7hDj$9$k(B */
  for (last_time = peseqlen - 1; last_time >= 0; last_time--) {
    nw[0]->tre = bt_binsearch_atom(&backtrellis, last_time, winfo.tail_silwid);
    if (nw[0]->tre != NULL) break;
  }
  if (nw[0]->tre == NULL) {		/* </s> not found */
    j_printerr("no wordend left in beam!\n");
    j_printerr("beam width too small or input speech too short\n");
    return 0;
  }
#endif

  return 1;
}

/* $B<!$K@\B3$7F@$kC18l72$rJV$9(B */
/* $BJV$jCM(B:$BC18l?t(B */
int
ngram_nextwords(NODE *hypo, NEXTWORD **nw)
{
  int num, num2;

  if (hypo->seqnum == 0) {
    j_printerr("gs_get_next_words: hypo contains no word\n");
    exit(1);
  }

  /* $B2>@b$N?dDj=*C<;~9o$K$*$$$F(B backtrellis$BFb$K;D$C$F$$$kC18l$rF@$k(B */ 
  num = get_backtrellis_words(nw, hypo, hypo->estimated_next_t, hypo->bestt);

  if (debug2_flag) j_printf("%d",num);

  /* $BE83+C18l$K@)8B$r2C$($k(B */
  num2 = limit_nw(nw, hypo, num);

  if (debug2_flag) j_printf("-%d=%d unfolded\n",num-num2,num2);

  return(num2);
}

/* $B2>@b$,J8$H$7$F<uM}2DG=$G$"$k$+$I$&$+$rJV$9(B */
boolean
ngram_acceptable(NODE *hypo)
{
  /* $B:G8e$,(B sil $B$@$C$?$i(B */
  if (hypo->seq[hypo->seqnum-1] == winfo.head_silwid) {
    return TRUE;
  } else {
    return FALSE;
  }
}

/* $B2>@b$HJ8$N:G8e$H$N@\B33NN($rJV$9(B (bogus) */
LOGPROB
ngram_eosscore(NODE *hypo)
{
  LOGPROB p;

  p =  uni_prob(&ngram, winfo.wton[hypo->seq[hypo->seqnum-1]]) * lm_weight2 + lm_penalty2;

  return p;

}

