/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* switch wchmm outprob calculation according to its style */

#include <julius.h>

#ifdef PASS1_IWCD

void
outprob_style_cache_init()
{
  int n;
  for(n=0;n<wchmm.n;n++) {
    if (wchmm.state[n].outstyle == AS_RSET) {
      (wchmm.state[n].out.rset)->cache = NULL;
    } else if (wchmm.state[n].outstyle == AS_LRSET) {
      (wchmm.state[n].out.lrset)->cache.state = NULL;
    }
  }
}

/* lcdset$B$O=PNO3NN($N:GBg$r5a$a$k(B */
static LOGPROB
outprob_lcd_max(LC_State_Set *lset, int t, HTK_Param *param)
{
  LOGPROB maxprob, prob;
  int i;
  maxprob = LOG_ZERO;
  for(i=0;i<lset->num;i++) {
    prob = outprob(t, lset->s[i], param);
    if (maxprob < prob) maxprob = prob;
  }
  return(maxprob);
/* 
 *   LOGPROB sum, p;
 *   int i,j;
 *   sum = 0.0;
 *   j = 0;
 *   for(i=0;i<lset->num;i++) {
 *     p = outprob(t, lset->s[i], param);
 *     if (p > LOG_ZERO) {
 *	 sum += p;
 *	 j++;
 *     }
 *   }
 *   return(sum/(float)j);
 */
}


/* outstyle $B$4$H$K0[$J$k7W;;J}K!$r;H$&(B */
static char rbuf[40];
LOGPROB
outprob_style(int node, int last_wid, int t, HTK_Param *param)
{
  HMM_Logical *ohmm, *rhmm;
  RC_INFO *rset;
  LRC_INFO *lrset;
  LC_Set *lcd;
  
  switch(wchmm.state[node].outstyle) {
  case AS_STATE:		/* normal state */
    return(outprob(t, wchmm.state[node].out.state, param));
  case AS_LSET:			/* end of word */
    return(outprob_lcd_max(wchmm.state[node].out.lset, t, param));
  case AS_RSET:			/* beginning of word */
    rset = wchmm.state[node].out.rset;
    if (rset->lastwid_cache != last_wid || rset->cache == NULL) {
      /* cache miss...calculate */
      rhmm = rset->hmm;
      if (last_wid != WORD_INVALID) {
	/* lookup triphone with left-context (= last phoneme) */
	strcpy(rbuf, rhmm->name);
	add_left_context(rbuf, (winfo.wseq[last_wid][winfo.wlen[last_wid]-1])->name);
	if ((ohmm = htk_hmmdata_lookup_logical(&hmminfo, rbuf)) != NULL) {
	  rhmm = ohmm;
	}
      }
      /* store to cache */
      rset->cache         = rhmm->def->s[rset->state_loc];
      rset->lastwid_cache = last_wid;
    }
    return(outprob(t, rset->cache, param));
  case AS_LRSET:		/* 1 phoneme --- beggining and end */
    lrset = wchmm.state[node].out.lrset;
    if (lrset->lastwid_cache != last_wid || lrset->cache.state == NULL) {
      /* cache miss...calculate */
      rhmm = lrset->hmm;
      strcpy(rbuf, rhmm->name);
      if (last_wid != WORD_INVALID) {
	add_left_context(rbuf, (winfo.wseq[last_wid][winfo.wlen[last_wid]-1])->name);
      }
      /* lookup lcdset for given left-context (= last phoneme) */
      if ((lcd = lcdset_lookup_by_hmmname(&hmminfo, rbuf)) != NULL) {
	/* found, store lcdset to cache and return */
	lrset->last_is_lset  = TRUE;
	lrset->cache.lset    = &(lcd->stateset[lrset->state_loc]);
	lrset->lastwid_cache = last_wid;
	/*printf("[%s->%s+*]\n", lrset->hmm->name, lcd->name);*/
	return(outprob_lcd_max(lrset->cache.lset, t, param));
      } else if ((lcd = lcdset_lookup_by_hmmname(&hmminfo, rhmm->name)) != NULL) {
	/* lookup lcdset ignoring left-context */
	lrset->last_is_lset  = TRUE;
	lrset->cache.lset    = &(lcd->stateset[lrset->state_loc]);
	lrset->lastwid_cache = last_wid;
	/*printf("[%s->%s+*]\n", lrset->hmm->name, lcd->name);*/
	return(outprob_lcd_max(lrset->cache.lset, t, param));
      } else {
	/* no relating lcdset found, falling to normal state */
	lrset->last_is_lset  = FALSE;
	lrset->cache.state   = rhmm->def->s[lrset->state_loc];
	lrset->lastwid_cache = last_wid;
	/*printf("[%s->%s]\n", lrset->hmm->name, rhmm->name);*/
	return(outprob(t, lrset->cache.state, param));
      }
    } else {			/* cache hit */
      if (lrset->last_is_lset) {
	return(outprob_lcd_max(lrset->cache.lset, t, param));
      } else {
	return(outprob(t, lrset->cache.state, param));
      }
    }
  default:
    /* should not happen */
    j_printerr("no outprob style??\n");
    return(LOG_ZERO);
  }
}

#endif /* PASS1_IWCD */

