/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* search_bestfirst_main.c --- main stack decoding loop for 2nd pass */

/* nw_* ... nextword related functions */
/* gm_* ... grammar related functions */
/* gm_* $B$O0J2<$N;EMM$rK~$?$9$Y$-(B:
 * int gm_get_first_words(NEXTWORD **nw, int peseqlen);
 * int gm_get_next_words(NODE *hypo, NEXTWORD **nw);
 * boolean gm_acceptable(NODE *hypo);
 * LOGPROB gm_get_eos_score(NODE *hypo);
 */
/* wchmm_fbs$B$N0z?t$H$7$F0J>e$N(B4$B$D$N4X?t$X$N%]%$%s%?$rM?$($k(B */

#include <julius.h>

static NODE *get_best_from_stack(NODE **start, int *stacknum);
static int put_to_stack(NODE *new, NODE **start, NODE **bottom, int *stacknum, int stacksize);
static void put_all_in_stack(NODE **start, int *stacknum);
static void free_all_nodes(NODE *node);
static void put_hypo_woutput(NODE *hypo);
static void put_hypo_wname(NODE *hypo);
static void put_hypo_phoneme(NODE *hypo);


/**********************************************************************/
/* $B<!C18l3JG<NN0h$N3d$jEv$F(B */
/**********************************************************************/

static NEXTWORD *nwtmp;
static NEXTWORD **
nw_malloc()
{
  NEXTWORD **nw;
  int i;

  /* $BO"B3NN0h$rG[Ns$K3d$jEv$F$k(B */
  nw = (NEXTWORD **)mymalloc(winfo.num * sizeof(NEXTWORD *)); /* $B<!C18l$N?t$O8lWC?t$r1[$($J$$(B */
  nwtmp = (NEXTWORD *)mymalloc(winfo.num * sizeof(NEXTWORD));
  nw[0] = nwtmp;
  for (i=1;i<winfo.num; i++) {
    nw[i] = &(nwtmp[i]);
  }
  return nw;
}

/* $BM=B,<!C18l3JG<NN0h$N2rJ|(B */
static void
nw_free(NEXTWORD **nw)
{
  free(nwtmp);
  free(nw);
}


/**********************************************************************/
/* $BJ82>@b%9%?%C%/A`:n(B */
/**********************************************************************/

/* $B%9%?%C%/%H%C%W$r(BPOP$B$9$k(B */
static NODE *
get_best_from_stack(NODE **start, int *stacknum)
{
  NODE *tmp;

  tmp=(*start);
  if ((*start)!=NULL) {
    (*start)=(*start)->next;
    if ((*start)!=NULL) (*start)->prev=NULL;
    (*stacknum)--;
    return(tmp);
  }
  else {
    return(NULL);
  }
}

/* $B%9%?%C%/$K%G!<%?$rF~$l$k(B(PUSH$B$K$"$i$:(B) */
static int
put_to_stack(
     NODE *new,
     NODE **start,
     NODE **bottom,
     int *stacknum,
     int stacksize)
{
  NODE *tmp;

  (*stacknum)++;

  if ((*stacknum)>=stacksize) {	/* overflow */
    (*stacknum)--;
    if ((*bottom)->score < new->score) { /* free bottom */
      tmp=(*bottom);
      (*bottom)->prev->next=NULL;
      (*bottom)=(*bottom)->prev;
      free_node(tmp);
    }
    else {
      free_node(new);		/* free newnode */
      return(-1);
    }
  }
    
  if ((*start)==NULL) {		/* no data */
    (*start)=new;
    (*bottom)=new;
    new->next=NULL;
    new->prev=NULL;
    return(0);
  }
  if ((*start)->score <= new->score) { /* on the top */
    new->next = (*start);
    new->next->prev = new;
    (*start)=new;
    new->prev=NULL;
    return(0);
  }
  if ((*bottom)->score >= new->score) {	/* on the bottom */
    new->prev = (*bottom);
    new->prev->next = new;
    (*bottom)=new;
    new->next=NULL;
    return(0);
  }
    
  /* now
     (*start)->score < tmp->score < (*bottom)->score
     */
  if (((*start)->score + (*bottom)->score) / 2 > new->score) {
    /* search from bottom */
    tmp=(*bottom);
    while(tmp->score < new->score) tmp=tmp->prev;
    new->prev=tmp;
    new->next=tmp->next;
    tmp->next->prev=new;
    tmp->next=new;
  } else {
    /* search from start */
    tmp=(*start);
    while(tmp->score > new->score) tmp=tmp->next;
    new->next=tmp;
    new->prev=tmp->prev;
    tmp->prev->next=new;
    tmp->prev=new;
  }
  return(0);
}

/* $B%9%?%C%/Fb$N2>@b$rA4$F=PNO(B */
/* $B%9%?%C%/$NCf?H$O<:$o$l$k(B */
static void
put_all_in_stack(NODE **start, int *stacknum)
{
  NODE *ntmp;
  
  j_printf("stack remains::\n");
  while ((ntmp = get_best_from_stack(start, stacknum)) != NULL) {
    j_printf("%3d: s=%f ",*stacknum, ntmp->score);
    put_hypo_woutput(ntmp);
    free_node(ntmp);
  }
}

/* $B%9%?%C%/Fb$NA42>@b$r2rJ|(B */
static void
free_all_nodes(NODE *node)
{
  NODE *tmp;
  NODE *next;

  tmp=node;
  while(tmp) {
    next=tmp->next;
    free_node(tmp);
    tmp=next;
  }
}


/**********************************************************************/
/* enveloped best-first */
/**********************************************************************/

#ifdef WORST_BEAM
/* $BF1$8D9$5$N2>@b$r!"0lDj?t0J>e$OJ]B8$7$J$$$h$&$K$9$k(B */
/* $B:G0-$G$b%S!<%`C5:w$K$J$k(B */
static int hypo_len_count[MAXSEQNUM+1];

static void
wb_init()
{
  int i;
  for(i=0;i<=MAXSEQNUM;i++) hypo_len_count[i] = 0;
}

static boolean
wb_ok(NODE *now)
{
  if (hypo_len_count[now->seqnum] >= worst_beam_width) {
    if (debug2_flag) {
      j_printf("xxx");
      put_hypo_woutput(now);
      j_printf("xxx pruned xxx\n");
    }
    return FALSE;
  } else {
    hypo_len_count[now->seqnum]++;
    return TRUE;
  }
}
#endif /* WORST_BEAM */


/**********************************************************************/
/* score-based envelope */
/**********************************************************************/

#ifdef SCAN_BEAM
#ifdef SCAN_BEAM_1
/* $BJ}K!(B1: $BDL>o$NJ}K!(B */
void
envl_init(int framenum)
{
  int i;
  for(i=0;i<framenum;i++) framemaxscore[i] = LOG_ZERO;
}
/* $B2>@b(BN$B$NA08~$-%9%3%"$+$i(Benvelope$B$r99?7$9$k(B */
static void
envl_update(NODE *n, int framenum)
{
  int t;
  for(t=framenum-1;t>=0;t--) {
    if (framemaxscore[t] < n->g[t]) framemaxscore[t] = n->g[t];
  }
}
#endif /* SCAN_BEAM_1 */
  
#ifdef SCAN_BEAM_2
/* $BJ}K!(B2: envelope$B$rBh(B1$B%Q%9$N%9%3%"$+$i$"$i$+$8$a;;=P$9$k(B */
void
envl_premake(BACKTRELLIS *bt, LOGPROB backmax)
{
  int t,i;

  /* $B%U%l!<%`$4$H$N%9%3%"$N:GBgCM$r5a$a$k(B */
  for(t=0;t<bt->framelen;t++) {
    for (i=0; i<bt->num[t]; i++) {
      if (framemaxscore[t] < (bt->rw[t][i])->backscore) {
	framemaxscore[t] = (bt->rw[t][i])->backscore;
      }
    }
  }
  /* $B:GL`2>@b%9%3%"$+$i$N:9J,$r$H$C$F3JG<(B */
  j_printf("-------- framemaxscore -------\n");
  for(t=0;t<bt->framelen;t++) {
    if (framemaxscore[t] <= LOG_ZERO) {
      framemaxscore[t] = backmax;
    } else {
      framemaxscore[t] = backmax - framemaxscore[t];
    }
    j_printf("%d %f\n",t,framemaxscore[t]);
  }
  j_printerr("envelope made\n");
  
}
#endif /* SCAN_BEAM_2 */

#endif /* SCAN_BEAM */




/**********************************************************************/
/* $BJ82>@b$N=PNO4X?t(B */
/**********************************************************************/

static HTK_Param *tparam;	/* for "-walign" and "-palign" option */

/* $B2>@b$N=PNO%7%s%\%k$r=PNO(B */
static void
put_hypo_woutput(NODE *hypo)
{
  int i,w;

  if (hypo != NULL) {
    for (i=hypo->seqnum-1;i>=0;i--) {
      w = hypo->seq[i];
      j_printf(" %s",winfo.woutput[w]);
    }
    /*    if (speech_input == SP_ADINSERV) {
      for (i=hypo->seqnum-1;i>=0;i--) {
	w = hypo->seq[i];
	if (strlen(winfo.woutput[w]) > 0)
	  wt(adinnet_asd, winfo.woutput[w], strlen(winfo.woutput[w]));
      }
      wt(adinnet_asd, (char *)&i, 0);
      }*/
  }
  j_printf("\n");  
}
/* $B2>@b$N8@8l%(%s%H%jC18l$rI8=`=PNO$K=PNO$9$k(B */
static void
put_hypo_wname(NODE *hypo)
{
  int i,w;

  if (hypo != NULL) {
    for (i=hypo->seqnum-1;i>=0;i--) {
      w = hypo->seq[i];
      j_printf(" %s",winfo.wname[w]);
    }
  }
  j_printf("\n");  
}
/* $B2>@b$N2;AGNs$r=PNO(B */
static void
put_hypo_phoneme(NODE *hypo)
{
  int i,j,w;
  char buf[20];

  if (hypo != NULL) {
    for (i=hypo->seqnum-1;i>=0;i--) {
      w = hypo->seq[i];
      for (j=0;j<winfo.wlen[w];j++) {
	center_name(winfo.wseq[w][j]->name, buf);
	j_printf(" %s", buf);
      }
      if (i > 0) j_printf(" |");
    }
  }
  j_printf("\n");  
}

/* $BF@$i$l$?J82>@b$rI=<((B */
void
putsentence(NODE *hypo, int num)
{
  char ec[5] = {0x1b, '[', '1', 'm', 0};
		
  if (debug2_flag) {
    j_printf("\n%s",ec);		/* newline & bold on */
  }
  j_printf("sentence%d:",num);
  put_hypo_woutput(hypo);
  if (debug_flag) {
    j_printf("wseq%d:",num);
    put_hypo_wname(hypo);
    j_printf("phseq%d:", num);
    put_hypo_phoneme(hypo);
  }
  if (debug2_flag) {
    ec[2] = '0';
    j_printf("%s\n",ec);		/* bold off & newline */
  }
  if (debug_flag) {
    j_printf("score%d: %f",num, (hypo != NULL) ? hypo->score : LOG_ZERO);
    if (separate_score_flag) {
      if (hypo == NULL) {
	j_printf(" (AM: %f  LM: %f)", LOG_ZERO, LOG_ZERO);
      } else {
	j_printf(" (AM: %f  LM: %f)", hypo->score - hypo->totallscore, hypo->totallscore);
      }
    }
    j_printf("\n");
  }
  j_flushprint();
}

/* $BJ88uJdMQ%9%?%C%/Fb$+$i>e0L$N7k2L$r0l5$$K=PNO$9$k(B */
static void
result_reorder_and_output(r_start, r_bottom, r_stacknum, ncan)
     NODE **r_start;
     NODE **r_bottom;
     int *r_stacknum;
     int ncan;
{
  NODE *now;
  int num;

  num = 0;
  while ((now = get_best_from_stack(r_start,r_stacknum)) != NULL && num < ncan) {
    /* $B=PNO(B */
    num++;
    putsentence(now, num);
    if (align_result_word_flag) word_rev_align(now->seq, now->seqnum, tparam);
    if (align_result_phoneme_flag) phoneme_rev_align(now->seq, now->seqnum, tparam);
    /* $B<N$F$k(B */
    free_node(now);
  }
  /* $B;D$j$bA4$F(B free $B$9$k(B */
  if (now != NULL) free_node(now);
  free_all_nodes(*r_start);
}  



/**********************************************************************/
/* $BC5:w<B9T%a%$%s4X?t(B */
/**********************************************************************/

void
wchmm_fbs(
     HTK_Param *param,		/* parameter vectors */
     BACKTRELLIS *backtrellis,	/* backward trellis */
     LOGPROB backmax,		/* backmax (retuen value) */
     int stacksize,		/* stack size of sentence hypothesis */
     int ncan,			/* num of hypothesis to get */
     int maxhypo,		/* maximum number of poped hypothesis */
     int (*gm_get_first_words)(NEXTWORD **, int),
     int (*gm_get_next_words)(NODE *, NEXTWORD **),
     boolean (*gm_acceptable)(NODE *),
     LOGPROB (*gm_get_eos_score)(NODE *))
{
  /* $BJ82>@b%9%?%C%/(B */
  int stacknum;			/* $B%9%?%C%/$NBg$-$5(B ( < stacksize ) */
  NODE *start = NULL;
  NODE *bottom = NULL;

  /* $BF@$i$l$?J88uJd%9%?%C%/(B for re-ordering */
  int r_stacksize = ncan + 1;
  int r_stacknum;
  NODE *r_start = NULL;
  NODE *r_bottom = NULL;

  /* $B%+%&%s%?N`(B */
  int popctr = 0;		/* $B%9%?%C%/$+$i<h$j=P$7$?J82>@b?t(B */
  int genectr = 0;		/* $B@8@.2>@b?t(B */
  int pushctr = 0;		/* $B%9%?%C%/$KF~$C$?J82>@b?t(B */
  int finishnum = 0;		/* $BC5:w$N7k2LF@$i$l$?J82>@b?t(B */

  /* $B%o!<%/%(%j%"(B */
  NEXTWORD **nextword;		/* $BM=B,<!C18l$N3JG<NN0h(B */
  int nwnum;
  NODE *now, *new;
  int w,i,j;
  LOGPROB last_score = LOG_ZERO;

  if (align_result_word_flag || align_result_phoneme_flag) tparam = param; /* store temporary */
  
  /* 
   * $B=i4|2=(B
   */
  peseqlen = param->samplenum;	/* for quick access */
  nextword = nw_malloc();	/* $BM=B,C18l3JG<NN0h$r3NJ](B */
  malloc_wordtrellis();		/* scan_word$BMQNN0h(B */
  start = bottom = NULL;	/* $BJ82>@b%9%?%C%/=i4|2=(B */
  stacknum = 0;
  if (result_reorder_flag) {
    r_start = r_bottom = NULL;
    r_stacknum = 0;
  }
#ifdef WORST_BEAM
  if (worst_beam_width >= 0) wb_init();
#endif
  
  if (debug_flag) j_printf("samplenum=%d\n", param->samplenum);
  if (result_reorder_flag) {
    if (debug2_flag) VERMES("getting %d candidates...",ncan);
  }

  /* 
   * $B=i4|$N(B1$BC18l$+$i$J$k2>@b$r:n$j!$J82>@b%9%?%C%/$K$$$l$k(B
   */
  nwnum = (*gm_get_first_words)(nextword,param->samplenum); /* nextwords$B$K3JG<$5$l$k(B */
#ifdef DEBUG
  j_printf("first words=%d\n", nwnum);
#endif
  for (w = 0; w < nwnum; w++) {
    new = newnode();
    start_word(new,nextword[w],param,backtrellis);
    genectr++;
    if (put_to_stack(new, &start, &bottom, &stacknum, stacksize) != -1) {
      pushctr++;
    }
  }

  /**************/
  /* $BC5:w%k!<%W(B */
  /**************/

  for (;;) {
    
    /* 
     * $BJ82>@b%9%?%C%/$+$i:G$b%9%3%"$N9b$$2>@b$r<h$j=P$9(B
     */
#ifdef DEBUG
    VERMES("get one hypothesis\n");
#endif
    
    now = get_best_from_stack(&start,&stacknum);
    if (now == NULL) {  /* stack empty ---> $BC5:w=*N;(B*/
      j_printf("stack empty\n");
      break;
    }
    if (now->score <= LOG_ZERO) continue;
#ifdef WORST_BEAM
    /* $BF1$8D9$5$N2>@b$r0lDj?t0J>eE83+$7$F$$$?$i!"$b$&$=$l0J>e$OE83+$7$J$$(B */
    if (worst_beam_width >= 0) {
      if (!wb_ok(now)) {
	free_node(now);
	continue;
      }
    }
#endif
    popctr++;

    /* 
     * $B<h$j=P$7$?2>@b$H$=$N%9%3%"$r%m%0$K=PNO(B
     */

    if (debug2_flag) {
      j_printf("--- %3d: s=%f g=%f lastw=[%3d-%3d]\n",popctr, now->score,
	     now->g[now->bestt], now->estimated_next_t + 1, now->bestt);
      j_printf("         "); put_hypo_woutput(now);
      j_printf("         "); put_hypo_wname(now);
    }
    
#ifdef SCAN_BEAM_1
    /* $B<h$j=P$7$?2>@b$N%9%3%"$r85$K(Benvelope$B$r99?7(B */
    envl_update(now, peseqlen);
#endif

    /* 
     * $B<h$j=P$7$?2>@b$N<uM}%U%i%0$,4{$KN)$C$F$$$l$P!$(B
     * $B$=$N2>@b$OC5:w=*N;$H$_$J$7!$7k2L$H$7$F=PNO$7$F<!$N%k!<%W$X!%(B
     */
#ifdef DEBUG
    VERMES("endflag check\n");
#endif
    
    if (now->endflag) {
      /* hack */
      if (now->score == last_score) {
	/* $BI=<($7$J$$(B */
	free_node(now);
	continue;
      } else {
	last_score = now->score;
      }
      finishnum++;

      if (result_reorder_flag) {
	/* $B0lDj?t$N2>@b$,F@$i$l$?$"$H%9%3%"$G%=!<%H$9$k$?$a!$(B
	 * $B0l;~E*$KJL$N%9%?%C%/$K3JG<$7$F$*$/(B */
	put_to_stack(now, &r_start, &r_bottom, &r_stacknum, r_stacksize);
	/* $B;XDj?t$NJ82>@b$,F@$i$l$?$J$iC5:w$r=*N;$9$k(B */
	if (finishnum >= ncan) {
	  if (debug2_flag) VERMES("%d\n",finishnum);
	  break;
	} else {
	  if (debug2_flag) VERMES("%d..", finishnum);
	  continue;
	}
      } else {
	/* $BJ82>@b$rI=<((B */
	putsentence(now,finishnum);
      /* Viterbi$B7PO)$r7W;;$7D>$73FC18l$,$I$N2;@<6h4V$K%^%C%A$9$k$+$rI=<($9$k(B */
	if (align_result_word_flag) word_rev_align(now->seq, now->seqnum, tparam);
	if (align_result_phoneme_flag) phoneme_rev_align(now->seq, now->seqnum, tparam);
	/* $B$3$N2>@b$O$b$&=*$o$j(B */
	free_node(now);
	/* $B;XDj?t$NJ82>@b$,F@$i$l$?$J$iC5:w$r=*N;$9$k(B */
	if (finishnum >= ncan) break;
	else continue;
      }
      
    }

    /* 
     * $BC5:w<:GT$r8!=P$9$k!%(B
     * $B2>@b?t$,(BMAXHYPO$B0J>eE83+$5$l$?$i<:GT$H8+$J$9!%(B
     */
#ifdef DEBUG
    VERMES("loop end check\n");
#endif
    
    if (popctr >= maxhypo) {
      j_printf("num of hypotheses overflow\n");
      /* $BC5:w<:GT;~$K!"%9%?%C%/$K;D$C$?>pJs$rEG$-=P$9(B */
      if (debug2_flag) put_all_in_stack(&start, &stacknum);
      free_node(now);
      break;			/* $BC5:w=*N;(B */
    }
    
    /* $B2>@bD9$,0lDjCM$r1[$($?$H$-!$$=$N2>@b$rGK4~$9$k(B */
    if (now->seqnum >= MAXSEQNUM) {
      j_printerr("sentence length exceeded ( > %d)\n",MAXSEQNUM);
      free_node(now);
      continue;
    }

    /*
     * $B<h$j=P$7$?2>@b$NA08~$-%9%3%"$r7W;;$9$k!%(B
     * $B:G8e$NC18l$NItJ,$NA08~$-%9%3%"$r7W;;$9$k!%(B
     */
#ifdef DEBUG
    VERMES("scan_word\n");
#endif

    scan_word(now, param);
    if (now->score < backmax + LOG_ZERO) {
      j_printf("now->score = %f?\n",now->score);
      put_hypo_woutput(now);
      free_node(now);
      break;
    }
    /* $B2>@b$N=*C<;~4V$r?dDj$9$k(B */
    /* $B%W%C%7%eA0$N(B next_word (start_word) $B$G(Bnow->estimated_next_t$B$K3JG<:Q(B */
    
    /* 
     * $B<h$j=P$7$?2>@b$,J8$H$7$F<uM}2DG=$G$"$l$P!$(B
     * $B<uM}%U%i%0$rN)$F$?$b$N$r%9%?%C%/$K$$$lD>$7$F$*$/!%(B
     * ($B<!$K<h$j=P$5$l$?$i=*N;(B)
     */
#ifdef DEBUG
    VERMES("accept check\n");
#endif
    if ((*gm_acceptable)(now) && now->estimated_next_t <= 5) {
      new = newnode();
      last_next_word(now, new, param);
      new->endflag = TRUE;
      put_to_stack(new, &start, &bottom, &stacknum, stacksize);
    }

    /* 
     * $B<h$j=P$7$?2>@b$K@\B3$7F@$kC18l$rIU2C$7$??7$?$JJ82>@b$r@8@.$7!$(B
     * $B%9%?%C%/$K$$$l$k!%(B($B:G=*C18l$N%9%3%"$O%R%e!<%j%9%F%#%C%/%9%3%"$N$^$^(B)
     */

#ifdef DEBUG
    VERMES("generate hypo\n");
#endif
    nwnum = (*gm_get_next_words)(now, nextword);

    i = pushctr;
    for (w = 0; w < nwnum; w++) {
      new = newnode();
      next_word(now, new, nextword[w], param, backtrellis);
      if (new->score <= LOG_ZERO) { /* not on trellis */
	free_node(new);
	continue;
      }
      genectr++;
      if (put_to_stack(new, &start, &bottom, &stacknum, stacksize) != -1) {
	if (debug2_flag) {
	  j = new->seq[new->seqnum-1];
	  j_printf("  %15s %15s(%f) [%d] pushed\n",winfo.wname[j], winfo.woutput[j], new->score, new->bestt);
	}
	pushctr++;
      }
    }
    if (debug2_flag) {
      j_printf("%d pushed\n",pushctr-i);
    }

    /* 
     * $B<h$j=P$7$?2>@b$r<N$F$k(B
     */
    free_node(now);

    /* $BCfCG%A%'%C%/(B */
    if (catch_intr_flag) break;
    
  }
  /***************/
  /* End of Loop */
  /***************/

  /* $B=PNO(B */
  /* $BC5:w<:GT$G0l$D$b8uJd$,F@$i$l$J$1$l$P!$(B */
  /* $BBh(B1$B%Q%9$N7k2L$r$=$N$^$^=PNO$9$k(B */
  if (finishnum == 0) {
    if (debug_flag) {
      j_printf("got no candidates, output 1st pass result as a final result\n",finishnum);
    }
    now = newnode();
    for (i=0;i<pass1_wnum;i++) {
      now->seq[i] = pass1_wseq[i];
    }
    now->seqnum = pass1_wnum;
    now->score = pass1_score;
    putsentence(now, 1);
    if (align_result_word_flag) word_rev_align(now->seq, now->seqnum, tparam);
    if (align_result_phoneme_flag) phoneme_rev_align(now->seq, now->seqnum, tparam);
    free_node(now);
  } else {
    if (debug2_flag) {
      j_printf("got %d candidates\n",finishnum);
    }
    if (result_reorder_flag) {
      if (debug2_flag) VERMES("done\n");
      /* $BJ88uJdMQ%9%?%C%/Fb$r%=!<%H$7$F!$7k2L$r0l5$$K=PNO$9$k(B */
      /* $B$D$$$G$K%9%?%C%/$b%/%j%"$9$k(B */
      result_reorder_and_output(&r_start, &r_bottom, &r_stacknum, output_hypo_maxnum);
    }
  }

  /* $B3F<o%+%&%s%?$r=PNO(B */
  if (debug_flag) {
    j_printf("%d generated, %d pushed, %d nodes popped in %d\n",
	   genectr, pushctr, popctr, param->samplenum);
  }
  
  /* $B=*N;=hM}(B */
  nw_free(nextword);
  free_all_nodes(start);
  free_wordtrellis();

}
