/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* search_bestfirst_v1.c --- hypothesis handling and scoring functions
   for backscan algorithm */


#include <julius.h>
#ifndef PASS2_STRICT_IWCD

#undef TCD			/* define for CCD debug messages */

/**********************************************************************/
/* $B2>@b%N!<%I$N4pK\A`:n(B */
/**********************************************************************/

/* $B2rJ|(B */
void
free_node(NODE *node)
{
  if (node == NULL) return;
  free(node->g);
#ifdef CCD_ON_2PASS
  if (ccd_flag) free(node->g_prev);
#endif

  free(node);
}

/* $B%3%T!<(B */
NODE *
cpy_node(NODE *dst, NODE *src)
{
  
  dst->next = src->next;
  dst->prev = src->prev;
  memcpy(dst->g, src->g, sizeof(LOGPROB) * peseqlen);
  memcpy(dst->seq, src->seq, sizeof(WORD_ID) * MAXSEQNUM);
#ifdef WORD_ALIGNMENT
  memcpy(dst->seq, src->seq, sizeof(short) * MAXSEQNUM);
#endif  
  dst->seqnum = src->seqnum;
  dst->score = src->score;
  dst->bestt = src->bestt;
  dst->estimated_next_t = src->estimated_next_t;
  dst->endflag = src->endflag;
#ifdef WORD_GRAPH
  dst->tre = src->tre;
#endif
#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    memcpy(dst->g_prev, src->g_prev, sizeof(LOGPROB)*peseqlen);
    dst->last_ph = src->last_ph;
    dst->lscore = src->lscore;
    dst->totallscore = src->totallscore;
  }
#endif
  return(dst);
}

/* $B?75,:n@.(B */
NODE *
newnode()
{
  NODE *tmp;
  int i;

  if ((tmp=(NODE *)mymalloc(sizeof(NODE)))==NULL) {
    j_printerr("can't malloc\n");
    exit(1);
  }
  /*bzero(tmp,sizeof(NODE));*/
  tmp->next=NULL;
  tmp->prev=NULL;
  tmp->g = (LOGPROB *)mymalloc(sizeof(LOGPROB)*peseqlen);
#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    tmp->g_prev = (LOGPROB *)mymalloc(sizeof(LOGPROB)*peseqlen);
    tmp->last_ph = NULL;
    tmp->lscore = LOG_ZERO;
    tmp->totallscore = LOG_ZERO;
  }
#endif
  tmp->endflag = FALSE;
  tmp->seqnum = 0;
  for(i=0;i<peseqlen;i++) {
    tmp->g[i] = LOG_ZERO;
#ifdef CCD_ON_2PASS
    if (ccd_flag) tmp->g_prev[i] = LOG_ZERO;
#endif
  }

  return(tmp);
}


/**********************************************************************/
/* $BA08~$-L`EY7W;;MQ%(%j%"(B */
/**********************************************************************/

static LOGPROB **wordtrellis;
static LOGPROB *g;
static HMM_Logical **phmmseq;
static int phmmlen_max;

void
malloc_wordtrellis()
{
  int i;
  int maxwn;
  LOGPROB *p;

  maxwn = winfo.maxwn + 10;	/* CCD$B$K$h$kJQF0$r9MN8(B */
  wordtrellis = (LOGPROB **)mymalloc(sizeof(LOGPROB *) * peseqlen);
  p = (LOGPROB *)mymalloc(sizeof(LOGPROB) * peseqlen * maxwn);
  for (i=0;i<peseqlen;i++) {
    wordtrellis[i] = p + i * maxwn;
  }
  wordtrellis[0][maxwn] = 1.0;

  g = (LOGPROB *)mymalloc(sizeof(LOGPROB) * peseqlen);

  phmmlen_max = winfo.maxwlen + 2;
  phmmseq = (HMM_Logical **)mymalloc(sizeof(HMM_Logical *) * phmmlen_max);
}
void
free_wordtrellis()
{
  free(wordtrellis[0]);
  free(wordtrellis);
  free(g);
  free(phmmseq);
}



/**********************************************************************/
/* $B2>@b$NL`EY7W;;(B */
/**********************************************************************/

#ifdef CCD_ON_2PASS
/* triphone$B$G$O(B
   $B!&2>@bE83+;~(B: right-context$B$N$_9MN8(B($B4{CN$@$+$i(B)$B$7$F%9%3%"7W;;(B
   $B!&2>@b<h$j$@$7;~(B: $B$5$+$N$\$C$F(B $BA08e$N(Bcontext$B$r9MN8$7$FA08~$-%9%3%"7W;;(B
   
   $B$=$N$?$a$K!$(B
   $B!&(BNODE$B$K$O:G8e$N(Bg[]$B$@$1$G$J$/!$$=$N(B1$BC18lA0$N(Bg[]$B$b5-O?$7$F$*$/!#(B
   $B!&(Bnext_word()$B$G$O!$0l%U%l!<%`J,$@$1(Bviterbi$B%Q%9$r7W;;$9$k$N$K!$(B
     $BE83+C18l$N:G8e$N2;AG$r$=$N$^$^;H$&$N$G$O$J$/!$$=$N1&C18l4V%3%s%F%-%9%H(B
     $B$r9MN8$7$?2;AG$G7W;;$7!$2>@b$N%9%3%"$H$9$k!#(B
   $B!&(Bscan_word()$B$G$O!$:#$^$G$O2>@b:G8e$NC18lJ,$K$D$$$F%H%l%j%9$r7W;;$7!$(B
     $BD>A0$N(Bg[]$B$+$i?7$?$J(Bg[]$B$r7W;;$7$F$$$?!#$3$3$+$i!$D>A0C18l$H$N6-3&It$N(B
     $B2;AG$,JQ$o$k$N$G!$(B2$BC18l$^$G$5$+$N$\$C$F%H%l%j%97W;;$r9T$&!#6qBNE*$K$O!$(B
     1)$BD>A0$N(Bg[]$B$r(Bg[]$B!$$=$N$b$&0l$DA0$N(Bg[]$B$r(Bg2[]$B$H$7$F%9%H%C%/$9$k!#(B
     2)$B:G8e$N(B2$BC18l$+$i!$C18l4V%3%s%F%-%9%H$r9MN8$7$D$D(BHMM$B$r:n$k!#(B
       $B!J(B1$BC18l$+$i$J$k2>@b$N>l9g!$(B1$BC18lJ,(B($B%3%s%F%-%9%H9MN8L5$7(B)$B!K(B
     3)g2[]$B$+$i(Bviterbi$B%Q%9$r7W;;$9$k!JD>A0C18lJ,$O?75,!$$=$N$b$&0l$DA0$N(B
       $BC18lJ,$O%3%s%F%-%9%H$r9MN8$7$?:F7W;;$H$$$&$3$H$K$J$k!K!#(B
       $B!J(B1$BC18l2>@b$N>l9g$O(Bg[]$B$+$i$G$h$$!K(B
     4)HMM$B$N:G8e$NItJ,$r?7$?$J(Bg[]$B!$C18l$H$N6-3&ItJ,$r?7$?$J(Bg2[]$B$H$7$F(B
       NODE$B$K3JG<$9$k!#(B
       $B!J(B1$BC18l2>@b$N>l9g$O%9%H%C%/$7$F$*$$$?(Bg[]$B$r(Bg2[]$B$H$9$k!K(B
   $B!&(Bstart_word()$B$G$O!$DL>oDL$j:G8e$N2;AGJ,$K$D$$$F%9%3%"$r7W;;$9$k$@$1!#(B
 */
#endif

/* $BJ82>@b$NA08~$-L`EY$r99?7$9$k(B */
/* $B:G8e$N(B1$BC18l$NA08~$-%H%l%j%9$r7W;;$9$k(B */
void
scan_word(NODE *now, HTK_Param *param)
{
  int   i,t;
  HMM *whmm;
  A_CELL *ac;
  WORD_ID word;
  LOGPROB tmpmax, tmptmp, score1;
  int startt = 0;
  int wordhmmnum;
#ifdef CCD_ON_2PASS
  LOGPROB tmpmax2 = LOG_ZERO;
  int phmmlen;
  int store_point = 0;
  int crossword_point = 0;
  boolean back_rescan = FALSE;
#endif
  boolean node_exist_p;
  
  /* ----------------------- prepare HMM ----------------------- */

#ifdef CCD_ON_2PASS
  if (ccd_flag) {
  /* $BD>A0$N2;AG$,$"$l$P!$$=$3$^$G$5$+$N$\$C$F(B scan $B$9$k(B */
    if (now->last_ph == NULL) {
      /* g$B$N3+;O%9%3%"(B: now->g[] */
      /* scan$BHO0O(B: $BC18l(B now->seq[now->seqnum-1]$BJ,(B */
      back_rescan = FALSE;
    } else {
      /* g$B$N3+;O%9%3%"(B: now->g_prev[] ($B0l2;AGA0$N(B g[])*/
      /* scan$BHO0O(B: $BC18l(B now->seq[now->seqnum-1] + now->last_ph */
      back_rescan = TRUE;
    }
  }
#endif
#ifdef TCD
  if (now->last_ph != NULL) {
    j_printf("inherited last_ph: %s\n", (now->last_ph)->name);
  } else {
    j_printf("no last_ph inherited\n");
  }
#endif

  /* scan $BHO0OJ,$N(BHMM$B$r=`Hw(B */
  word = now->seq[now->seqnum-1];
#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    if (back_rescan) {
      /* scan$BHO0O(B: $BC18l(B now->seq[now->seqnum-1] + now->last_ph */
      phmmlen = winfo.wlen[word] + 1;
      if (phmmlen > phmmlen_max) {
	j_printerr("short of phmmlen\n");
	exit(1);
      }
      for (i=0;i<phmmlen - 2;i++) {
	phmmseq[i] = winfo.wseq[word][i];
      }
      /* word$B$N:G8e$N2;AG$H(B now->last_ph $B$O4D6-0MB8$K$h$jJQ2=(B */
      phmmseq[phmmlen-2] = get_right_context_HMM(winfo.wseq[word][winfo.wlen[word]-1], now->last_ph, &hmminfo);
      phmmseq[phmmlen-1] = get_left_context_HMM(now->last_ph, winfo.wseq[word][winfo.wlen[word]-1], &hmminfo);

#ifdef TCD
      j_printf("w=");
      for(i=0;i<winfo.wlen[word];i++) {
	j_printf(" %s",(winfo.wseq[word][i])->name);
      }
      j_printf(" | %s\n", (now->last_ph)->name);
      j_printf("scan for:");
      for (i=0;i<phmmlen;i++) {
	j_printf(" %s", phmmseq[i]->name);
      }
      j_printf("\n");
#endif

      /* $BC18l(BHMM$B$r:n$k(B */
      whmm = new_make_word_hmm(&hmminfo, phmmseq, phmmlen);
      /* $B7W;;A0$N(B g[] $B$O!$(B1$BC18lA0$N(B g[] == g_prev[] */
      for (t=0;t<peseqlen;t++) {
	g[t]=now->g_prev[t];
      }
      /* $B<!CJMQ$N(Bg_prev$B$r3JG<$9$k%N!<%I0LCV$O(B
	 $B@hF,$N(B1$B2;AG$N:G8e$N%N!<%I(B */
      store_point = (phmmseq[0])->def->state_num - 2 - 1;
      /* $BD>A0C18l$H$3$NC18l$r$^$?$0ItJ,$O8@8l%9%3%"$rB-$9(B */
      crossword_point = whmm->len - ((phmmseq[phmmlen-1])->def->state_num - 2) - 1;
    } else {
      /* scan$BHO0O(B: $BC18l(B now->seq[now->seqnum-1]$BJ,(B */
#ifdef TCD
      j_printf("scan(org):");
      for (i=0;i<winfo.wlen[word];i++) {
	j_printf(" %s", (winfo.wseq[word][i])->name);
      }
      j_printf("\n");
#endif
      whmm = new_make_word_hmm(&hmminfo, winfo.wseq[word], winfo.wlen[word]);
      /* g[]$B$K$O!$D>A0C18l$+$iC18l4VA+0\$7$?>l9g$N%9%3%"$,4{$K<}$a$i$l$F$$$k(B */
      for (t=0;t<peseqlen;t++) {
	g[t]=now->g[t];
      }
      /* $B<!CJ7W;;MQ$K!$(Bg_prev$B$r3JG<$7$F$*$/%N!<%I0LCV$r$"$i$+$8$a7W;;(B */
      store_point = (winfo.wseq[word][0])->def->state_num - 2 - 1;
      crossword_point = -1;
    }
  } else {			/* ccd_flag == FALSE */
    whmm = new_make_word_hmm(&hmminfo, winfo.wseq[word], winfo.wlen[word]);
    /* g[]$B$K$O!$D>A0C18l$+$iC18l4VA+0\$7$?>l9g$N%9%3%"$,4{$K<}$a$i$l$F$$$k(B */
    for (t=0;t<peseqlen;t++) {
      g[t]=now->g[t];
    }
  }
#else  /* not CCD_ON_2PASS */
  whmm = new_make_word_hmm(&hmminfo, winfo.wseq[word], winfo.wlen[word]);
  /* g[]$B$K$O!$D>A0C18l$+$iC18l4VA+0\$7$?>l9g$N%9%3%"$,4{$K<}$a$i$l$F$$$k(B */
  for (t=0;t<peseqlen;t++) {
    g[t]=now->g[t];
  }
#endif /* CCD_ON_2PASS */

#ifdef TCD
  j_printf("whmm len	  = %d\n",whmm->len);
  j_printf("crossword_point = %d\n", crossword_point);
  j_printf("g[] store point = %d\n", store_point);
#endif

  wordhmmnum = whmm->len;
  if (wordhmmnum >= winfo.maxwn + 10) {
    j_printerr("scan_word: word too long\n");
    exit(1);
  }

  /* ----------------------- do scan ----------------------- */
#if 0
  printf("-------\n");
  for(t=0;t<peseqlen;t++) {
    printf("%d %f %f\n",t,framemaxscore[t],g[t]);
  }
  printf("-------\n");
#endif
  
  /* scan$B3+;OE@$r8!:w(B -> startt$B$X(B*/
  for(t = peseqlen-1; t >=0 ; t--) {
    if (
#ifdef SCAN_BEAM
	g[t] > framemaxscore[t] - scan_beam_thres &&
#endif
	g[t] > LOG_ZERO) {
      break;
    }
  }
  if (t < 0) {			/* no node has score > LOG_ZERO */
    for(t=0;t<peseqlen;t++) {
#ifdef CCD_ON_2PASS
      if (ccd_flag) now->g_prev[t] = LOG_ZERO;
#endif
      now->g[t] = LOG_ZERO;
    }
    goto end_of_scan;
  }

  startt = t;
  /* [startt+1..peseqlen-1] $B$N7k2LCM$r%/%j%"(B */
  for(t=peseqlen-1;t>startt;t--) {
#ifdef CCD_ON_2PASS
    if (ccd_flag) now->g_prev[t] = LOG_ZERO;
#endif
    now->g[t] = LOG_ZERO;
  }
  /* [startt] $B%N!<%I$H7k2LCM$r=i4|2=(B */
  for(i=0;i<wordhmmnum-1;i++) wordtrellis[startt][i] = LOG_ZERO;
  wordtrellis[startt][wordhmmnum-1] = g[startt] + outprob(startt, whmm->state[wordhmmnum-1].out, param);
#ifdef CCD_ON_2PASS
  if (ccd_flag) now->g_prev[startt] = LOG_ZERO;
#endif
  now->g[startt] = LOG_ZERO;

  /* [0..startt-1] $B$r7W;;(B */
  
  for(t=startt-1;t>=0;t--) {
    node_exist_p = FALSE;

    /* node[wordhmmnum-1]$B$O!$<+8JA+0\(B $B$+(B g[]$B$N9b$$J}$K$J$k(B */
    tmptmp = LOG_ZERO;
    for (ac=whmm->state[wordhmmnum-1].ac;ac;ac=ac->next) {
      if (ac->arc == wordhmmnum - 1) { /* $B<+8JA+0\(B */
	tmptmp = wordtrellis[t+1][wordhmmnum-1] + ac->a;
	break;
      }
    }
    tmpmax = (g[t] > tmptmp) ? g[t] : tmptmp;
    if (
#ifdef SCAN_BEAM
	tmpmax <= framemaxscore[t] - scan_beam_thres ||
#endif
	tmpmax <= LOG_ZERO
	) {
      /* invalid node */
      wordtrellis[t][wordhmmnum-1] = LOG_ZERO;
    } else {
      node_exist_p = TRUE;
      wordtrellis[t][wordhmmnum-1] = tmpmax + outprob(t, whmm->state[wordhmmnum-1].out, param);
    }

    /* node[wordhmmnum-2..0]$B$K$D$$$F%H%l%j%9$r7W;;(B */
    for(i=wordhmmnum-2;i>=0;i--) {

#ifdef CCD_ON_2PASS
      if (ccd_flag) {
	if (i == store_point) {
	  tmpmax2 = LOG_ZERO;	/* $B<+8JA+0\$r=|$/(Bmax */
	}
	tmpmax = LOG_ZERO;
	for (ac=whmm->state[i].ac;ac;ac=ac->next) {
	  score1 = wordtrellis[t+1][ac->arc] + ac->a;
	  if (i <= crossword_point && ac->arc > crossword_point) {
	    /* $BC18l$r1[$($kA+0\(B */
	    score1 += now->lscore;
	  }
	  if (i == store_point && i != ac->arc) {
	    if (tmpmax2 < score1) tmpmax2 = score1;
	  }
	  if (tmpmax < score1) tmpmax = score1;
	}
	if (
#ifdef SCAN_BEAM
	    tmpmax <= framemaxscore[t] - scan_beam_thres ||
#endif
	    tmpmax <= LOG_ZERO
	    ) {
	  /* invalid node */
	  wordtrellis[t][i] = LOG_ZERO;
	  if (i == store_point) now->g_prev[t] = LOG_ZERO;
	} else {
	  /* survived node */
	  if (i == store_point) now->g_prev[t] = tmpmax2;
	  node_exist_p = TRUE;
	  tmptmp = outprob(t, whmm->state[i].out, param);
	  wordtrellis[t][i] = tmpmax + tmptmp;
	}
      } else {
	tmpmax = LOG_ZERO;
	for (ac=whmm->state[i].ac;ac;ac=ac->next) {
	  score1 = wordtrellis[t+1][ac->arc] + ac->a;
	  if (tmpmax < score1) tmpmax = score1;
	}
	if (
#ifdef SCAN_BEAM
	    tmpmax <= framemaxscore[t] - scan_beam_thres ||
#endif
	    tmpmax <= LOG_ZERO
	    ) {
	  /* invalid node */
	  wordtrellis[t][i] = LOG_ZERO;
	} else {
	  /* survived node */
	  node_exist_p = TRUE;
	  wordtrellis[t][i] = tmpmax + outprob(t, whmm->state[i].out, param);
	}
      }
#else  /* not CCD_ON_2PASS */
      tmpmax = LOG_ZERO;
      for (ac=whmm->state[i].ac;ac;ac=ac->next) {
	score1 = wordtrellis[t+1][ac->arc] + ac->a;
	if (tmpmax < score1) tmpmax = score1;
      }
      if (
#ifdef SCAN_BEAM
	  tmpmax <= framemaxscore[t] - scan_beam_thres ||
#endif
	  tmpmax <= LOG_ZERO
	  ) {
	/* invalid node */
	wordtrellis[t][i] = LOG_ZERO;
      } else {
	/* survived node */
	node_exist_p = TRUE;
	wordtrellis[t][i] = tmpmax + outprob(t, whmm->state[i].out, param);
      }
#endif /* CCD_ON_2PASS */
    } /* end of node loop */
    
    now->g[t] = wordtrellis[t][0];

/*    if (startt - t > BEAM_STEP && (!node_exist_p)) {*/
    if (t < now->estimated_next_t && (!node_exist_p)) {
      /* terminate remaining scan */
      for (i=t-1;i>=0;i--) {
	now->g[i] = LOG_ZERO;
#ifdef CCD_ON_2PASS
	if (ccd_flag) now->g_prev[i] = LOG_ZERO;
#endif
      }
      break;			/* terminate loop */
    }
    
  } /* end of time loop */
  if (debug2_flag) j_printf("scanned: [%3d-%3d]\n",t+1,startt);

 end_of_scan:
#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    if (store_point == wordhmmnum - 1) {
      /* last_ph$BL5$7!$$+$DC18l$N2;AGD9(B=1$B$N;~$KH/@8(B */
      /* g_prev $B$O!$$3$N(Bscan_word$B$N3+;O%9%3%"!J$9$J$o$A(Bg[]$B$r$=$N$^$^F~$l$k(B */
      for (t = startt; t>=0; t--) {
	now->g_prev[t] = g[t];
      }
    }
    /* $B<!2s$N$?$a$K(B now->last_ph $B$r99?7(B */
    if (back_rescan) {
      now->last_ph = phmmseq[0];
    } else {
      now->last_ph = winfo.wseq[word][0];
    }
  }
#endif
  
  free_hmm(whmm);
#ifdef TCD
  j_printf("last_ph = %s\n", (now->last_ph)->name);
#endif
}


/* $B2>@b$rE83+$7!$(B($B%R%e!<%j%9%F%#%C%/$r4^$`(B)$B2>@b%9%3%"$r5a$a$k(B */
void
next_word(
     NODE *now, NODE *new,
     NEXTWORD *nword,
     HTK_Param *param,
     BACKTRELLIS *backtrellis)
{
  int   t;
  HMM_Logical *newphone;
  int lastword;
  int   i;
  LOGPROB a_value;
  int   startt;
  int word;
  LOGPROB totalscore;
  TRELLIS_ATOM *tre;

  new->score = LOG_ZERO;
  
  word = nword->id;
  for (i=0;i< now->seqnum;i++){
    new->seq[i] = now->seq[i];
  }
  new->seq[i] = word;
  new->seqnum = now->seqnum+1;

  new->totallscore = now->totallscore + nword->lscore;

  lastword=now->seq[now->seqnum-1];

#ifdef CCD_ON_2PASS
  /* $B1&$N2;AG4D6-$N$_$r9MN8(B */
  if (ccd_flag) {
    newphone = get_right_context_HMM(winfo.wseq[word][winfo.wlen[word]-1], now->last_ph, &hmminfo);
    /* $BMzNr>pJs$r%3%T!<(B */
    for (t=0;t<peseqlen;t++) {
      new->g_prev[t] = now->g_prev[t];
    }
    new->last_ph = now->last_ph;
    new->lscore = nword->lscore;
  } else {
    newphone = winfo.wseq[word][winfo.wlen[word]-1];
  }
#else
  newphone = winfo.wseq[word][winfo.wlen[word]-1];
#endif /* CCD_ON_2PASS */
  
#ifdef LOGTABLE
  a_value = newphone->def->tr->a[newphone->def->state_num-2][newphone->def->state_num-1];
#else
  a_value = log10(newphone->def->tr->a[newphone->def->state_num-2][newphone->def->state_num-1]);
#endif
  
  startt = peseqlen-2;
  new->g[startt+1] = LOG_ZERO;
  
#ifdef WORD_GRAPH
  /* $BF1$8C18l$G$b8uJdKh$K%9%3%"$,0c$&(B */
  /* bestt $B$O(B $B8GDj(B */
  for(t=startt;t>=0;t--) {
    new->g[t] = now->g[t+1] + a_value
      + nword->lscore
      ;
  }
  /* $B:G8e$K;2>H$7$?(BTRELLIS_ATOM$B$N=*C<$K8GDj(B */
  new->bestt = (nword->tre)->endtime;
  /* new$B$N?dDj=*C<;~4V$O!$>e5-$N;OC<(B */
  new->estimated_next_t = (nword->tre)->begintime - 1;
  new->score = new->g[new->bestt] + (nword->tre)->backscore
#ifdef DISCOUNT_BACKSCORE
    * discount_backscore_rate
#endif
      + outprob(new->bestt, newphone->def->s[newphone->def->state_num-2], param);
  new->tre  =  nword->tre;
  
#else  /* not WORD_GRAPH */
  
  /* $B:GL`C<$r:F7W;;(B */
  /* $B:G8e$K;2>H$7$?(BTRELLIS_ATOM$B$N=*C<;~4V$NA08e(B */
  /* new$B$N?dDj;~4V$O!$>e5-$G:NMQ$7$?(BTRELLIS_ATOM$B$N;OC<;~4V(B */
  for(t=startt;t>=0;t--) {
    new->g[t] = now->g[t+1] + a_value
      + nword->lscore
      ;
  }

  /* $BE83+$7$?%H%l%j%9>e$NC18l$N=*C<$NA08e$N$_$G(Bscan$B$9$k(B */
  /* 1. $BA0J}$X8!:w(B */
  for(t = (nword->tre)->endtime; t >= 0; t--) {
    tre = bt_binsearch_atom(backtrellis, t, (WORD_ID) word);
    if (tre == NULL) break;	/* $B%H%l%j%9$K;D$C$F$J$1$l$P=*N;(B */
    totalscore = new->g[t] + tre->backscore
#ifdef DISCOUNT_BACKSCORE
      * discount_backscore_rate
#endif
	+ outprob(t, newphone->def->s[newphone->def->state_num-2], param);
    if (new->score < totalscore) {
      new->score = totalscore;
      new->bestt = t;
      new->estimated_next_t = tre->begintime - 1;
    }
  }
  /* 2. $B8eJ}$X8!:w(B */
  for(t = (nword->tre)->endtime + 1; t <= startt; t++) {
    tre = bt_binsearch_atom(backtrellis, t, (WORD_ID) word);
    if (tre == NULL) break;	/* $B%H%l%j%9$K;D$C$F$J$1$l$P=*N;(B */
    totalscore = new->g[t] + tre->backscore 
#ifdef DISCOUNT_BACKSCORE
      * discount_backscore_rate
#endif
	+ outprob(t, newphone->def->s[newphone->def->state_num-2], param);
    if (new->score < totalscore) {
      new->score = totalscore;
      new->bestt = t;
      new->estimated_next_t = tre->begintime - 1;
    }
  }

  
#endif /* WORD_GRAPH */
}


/* $B=i4|2>@b$r@8@.(B --- $BC5:w$N:G=i$K$N$_8F$P$l$k(B */
void
start_word(
     NODE *new,
     NEXTWORD *nword,
     HTK_Param *param,
     BACKTRELLIS *backtrellis)
{
  HMM_Logical *newphone;
  WORD_ID word;
  TRELLIS_ATOM *tre = NULL;
  int t;

  /* $B=i4|2=(B */
  word = nword->id;
  new->score = LOG_ZERO;
  new->seqnum = 1;
  new->seq[0] = word;

  new->totallscore = nword->lscore;

  newphone = winfo.wseq[word][winfo.wlen[word]-1];
#ifdef CCD_ON_2PASS
  /* $B=i4|2>@b$G$O2;AG4D6-$O9MN8$9$kI,MW$,$J$$(B */
  if (ccd_flag) {
    new->last_ph = NULL;
    new->lscore = nword->lscore;
  }
#endif /* CCD_ON_2PASS */
  
  new->g[peseqlen-1] = nword->lscore;
  
#ifdef WORD_GRAPH
  new->score = new->g[peseqlen-1] + (nword->tre)->backscore
#ifdef DISCOUNT_BACKSCORE
    * discount_backscore_rate
#endif
      + outprob(peseqlen-1, newphone->def->s[newphone->def->state_num-2], param);
  new->tre  =  nword->tre;
  new->bestt = peseqlen-1;
  new->estimated_next_t = (nword->tre)->begintime - 1;
  
#else

  for (t=peseqlen-1; t>=0; t--) {
    tre = bt_binsearch_atom(backtrellis, t, word);
    if (tre != NULL) {
      new->bestt = t;
      new->score = new->g[peseqlen-1] + tre->backscore
#ifdef DISCOUNT_BACKSCORE
	* discount_backscore_rate
#endif
        + outprob(peseqlen-1, newphone->def->s[newphone->def->state_num-2], param);
      new->estimated_next_t = tre->begintime - 1;
      break;
    }
  }
  if (tre == NULL) {		/* no word in backtrellis */
    new->score = LOG_ZERO;
  }
#endif /* WORD_GRAPH */
}



/* $B:G=*2>@b$N%9%3%"7W;;(B --- $B<uM}2>@b$N:G=*%9%3%"$r@_Dj(B */
void
last_next_word(NODE *now, NODE *new, HTK_Param *param)
{
  cpy_node(new, now);
  new->score = now->g[0];
}


#endif /* PASS2_STRICT_IWCD */
