/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* search_bestfirst_v2.c --- hypothesis handling and scoring functions
   for nextscan algorithm */
/* for only triphones with option "--enable-strict-pass2iwcd" */

#include <julius.h>
#ifdef PASS2_STRICT_IWCD

#undef TCD			/* define for CCD debug messages */

/**********************************************************************/
/* $B2>@b%N!<%I$N4pK\A`:n(B */
/**********************************************************************/

/* $B2rJ|(B */
void
free_node(NODE *node)
{
  if (node == NULL) return;
  free(node->g);
  free(node);
}

/* $B%3%T!<(B */
NODE *
cpy_node(NODE *dst, NODE *src)
{
  
  dst->next = src->next;
  dst->prev = src->prev;
  memcpy(dst->g, src->g, sizeof(LOGPROB) * peseqlen);
  memcpy(dst->seq, src->seq, sizeof(WORD_ID) * MAXSEQNUM);
  dst->seqnum = src->seqnum;
  dst->score = src->score;
  dst->bestt = src->bestt;
  dst->estimated_next_t = src->estimated_next_t;
  dst->endflag = src->endflag;
#ifdef WORD_GRAPH
  dst->tre = src->tre;
#endif
#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    dst->last_ph = src->last_ph;
    dst->totallscore = src->totallscore;
  }
#endif
  return(dst);
}

/* $B?75,:n@.(B */
NODE *
newnode()
{
  NODE *tmp;
  int i;

  if ((tmp=(NODE *)mymalloc(sizeof(NODE)))==NULL) {
    j_printerr("can't malloc\n");
    exit(1);
  }
  /*bzero(tmp,sizeof(NODE));*/
  tmp->next=NULL;
  tmp->prev=NULL;
  tmp->g = (LOGPROB *)mymalloc(sizeof(LOGPROB)*peseqlen);
#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    tmp->last_ph = NULL;
    tmp->totallscore = LOG_ZERO;
  }
#endif
  tmp->endflag = FALSE;
  tmp->seqnum = 0;
  for(i=0;i<peseqlen;i++) {
    tmp->g[i] = LOG_ZERO;
  }

  return(tmp);
}


/**********************************************************************/
/* $BA08~$-L`EY7W;;MQ%(%j%"(B */
/**********************************************************************/

static LOGPROB **wordtrellis;
static LOGPROB *g;
static HMM_Logical **phmmseq;
static int phmmlen_max;
static HMM_Logical *tailph;

void
malloc_wordtrellis()
{
  int i;
  int maxwn;
  LOGPROB *p;

  maxwn = winfo.maxwn + 10;	/* CCD$B$K$h$kJQF0$r9MN8(B */
  wordtrellis = (LOGPROB **)mymalloc(sizeof(LOGPROB *) * peseqlen);
  p = (LOGPROB *)mymalloc(sizeof(LOGPROB) * peseqlen * maxwn);
  for (i=0;i<peseqlen;i++) {
    wordtrellis[i] = p + i * maxwn;
  }
  wordtrellis[0][maxwn] = 1.0;

  g = (LOGPROB *)mymalloc(sizeof(LOGPROB) * peseqlen);

  phmmlen_max = winfo.maxwlen + 2;
  phmmseq = (HMM_Logical **)mymalloc(sizeof(HMM_Logical *) * phmmlen_max);
}
void
free_wordtrellis()
{
  free(wordtrellis[0]);
  free(wordtrellis);
  free(g);
  free(phmmseq);
}




/**********************************************************************/
/* $B2>@b$NL`EY7W;;(B */
/**********************************************************************/

/* $BM?$($i$l$?2;AG$N$J$i$S$KBP$7$F(Bviterbi$B7W;;$r9T$&!%(B
   g[0..framelen-1] $B$N%9%3%"$r(B, g_new[0..framelen-1]$B$X99?7(B
   $B:GDc(B least_frame $B$^$G$O(Bscan$B$9$k!%(B
 */
static void
do_viterbi(LOGPROB *g, LOGPROB *g_new, HMM_Logical **phmmseq, int phmmlen, HTK_Param *param, int framelen, int least_frame)
{
  HMM *whmm;			/* HMM */
  int wordhmmnum;		/* length of above */
  int startt;			/* scan start frame */
  LOGPROB tmpmax,tmpscore;	/* variables for Viterbi process */
  A_CELL *ac;
  int t,i;
  boolean node_exist_p;

#ifdef TCD
  j_printf("scan for:");
  for (i=0;i<phmmlen;i++) {
    j_printf(" %s", phmmseq[i]->name);
  }
  j_printf("\n");
#endif
  
  /* HMM$B$r9=@.(B */
  whmm = new_make_word_hmm(&hmminfo, phmmseq, phmmlen);
  wordhmmnum = whmm->len;
  if (wordhmmnum >= winfo.maxwn + 10) {
    j_printerr("scan_word: word too long\n");
    exit(1);
  }

  /* scan$B3+;OE@$r8!:w(B -> startt$B$X(B*/
  for(t = framelen-1; t >=0 ; t--) {
    if (
#ifdef SCAN_BEAM
	g[t] > framemaxscore[t] - scan_beam_thres &&
#endif
	g[t] > LOG_ZERO) {
      break;
    }
  }
  if (t < 0) {			/* no node has score > LOG_ZERO */
    for(t=0;t<framelen;t++) {	/* reset all and end */
      g_new[t] = LOG_ZERO;
    }
    free_hmm(whmm);
    return;
  }
  startt = t;
  
  /* $B3+;OE@0J9_(B[startt+1..framelen-1] $B$N(B g_new[] $B$r%j%;%C%H(B */
  for(t=framelen-1;t>startt;t--) g_new[t] = LOG_ZERO;

  /* viterbi$B3+;O(B */
  /* $B:G=i$N(B[startt]$B$O(B g[]$B>e$N%N!<%I$N$_7W;;(B */
  for(i=0;i<wordhmmnum-1;i++) wordtrellis[startt][i] = LOG_ZERO;
  wordtrellis[startt][wordhmmnum-1] = g[startt] + outprob(startt, whmm->state[wordhmmnum-1].out, param);
  g_new[startt] = LOG_ZERO;
  
  /* [0..startt-1] $B$r7W;;(B */
  for(t=startt-1; t>=0; t--) {
    node_exist_p = FALSE;

    /* node[wordhmmnum-1]$B$O!$<+8JA+0\(B $B$+(B g[] */
    tmpscore = LOG_ZERO;
    for (ac=whmm->state[wordhmmnum-1].ac;ac;ac=ac->next) {
      if (ac->arc == wordhmmnum - 1) { /* $B<+8JA+0\(B */
	tmpscore = wordtrellis[t+1][wordhmmnum-1] + ac->a;
	break;
      }
    }
    tmpmax = (g[t] > tmpscore) ? g[t] : tmpscore;
    /* $BM-8z$J%N!<%I$,B8:_$9$k$+%A%'%C%/(B */
    if (
#ifdef SCAN_BEAM
	tmpmax <= framemaxscore[t] - scan_beam_thres ||
#endif
	tmpmax <= LOG_ZERO
	) {
      /* invalid node */
      wordtrellis[t][wordhmmnum-1] = LOG_ZERO;
    } else {
      node_exist_p = TRUE;
      wordtrellis[t][wordhmmnum-1] = tmpmax + outprob(t, whmm->state[wordhmmnum-1].out, param);
    }

    /* node[wordhmmnum-2..0]$B$K$D$$$F%H%l%j%9$rE83+(B */
    for(i=wordhmmnum-2;i>=0;i--) {
      tmpmax = LOG_ZERO;
      for (ac=whmm->state[i].ac;ac;ac=ac->next) {
	tmpscore = wordtrellis[t+1][ac->arc] + ac->a;
	if (tmpmax < tmpscore) tmpmax = tmpscore;
      }
      if (
#ifdef SCAN_BEAM
	  tmpmax <= framemaxscore[t] - scan_beam_thres ||
#endif
	  tmpmax <= LOG_ZERO
	  ) {
	/* invalid node */
	wordtrellis[t][i] = LOG_ZERO;
      } else {
	/* survived node */
	node_exist_p = TRUE;
	wordtrellis[t][i] = tmpmax + outprob(t, whmm->state[i].out, param);
      }
    } /* end of node loop */

    /* g_new $B$r99?7(B */
    g_new[t] = wordtrellis[t][0];

    /* $B%U%l!<%`Fb$K%N!<%I$,;D$C$F$$$J$1$l$P(B scan $B$r=*N;(B */
    if (t < least_frame && (!node_exist_p)) {
      /* terminate remaining scan */
      for (i=t-1;i>=0;i--) {
	g_new[i] = LOG_ZERO;
      }
      break;			/* terminate loop */
    }
  }

  free_hmm(whmm);
}

/* next_word$BFb$G(B, now $B:G8e$N(B1$B2;AG(B(lastphone)$BJ,$r(Bscan$B$7!$99?7$7$?%9%3%"$r(B */
/* new $B$KJ]B8(B */
static void
do_viterbi_next_word(NODE *now, NODE *new, HMM_Logical *lastphone, HTK_Param *param)
{
  LOGPROB a_value;
  int t;
  
  if (winfo.wlen[now->seq[now->seqnum-1]] > 1) {
    /* $BE83+852>@b$,(Bscan$B$5$l$F$$$k$J$i(Bg[]$B$r?7$?$K=`Hw(B */
#ifdef LOGTABLE
    a_value = lastphone->def->tr->a[lastphone->def->state_num-2][lastphone->def->state_num-1];
#else
    a_value = log10(lastphone->def->tr->a[lastphone->def->state_num-2][lastphone->def->state_num-1]);
#endif
    for(t=0; t<peseqlen-1; t++) g[t] = now->g[t+1] + a_value;
    g[peseqlen-1] = LOG_ZERO;
  } else {
    /* $B$9$G$K=`Hw:Q$_!J(B2$B2sA0$N$b$N$rN.MQ!K(B */
    for(t=0; t<peseqlen; t++) g[t] = now->g[t];
  }
  
  /* viterbi $B<B9T(B */
  do_viterbi(g, new->g, &lastphone, 1, param, peseqlen, now->estimated_next_t);
}

/* $BJ82>@b$NA08~$-L`EY$r99?7$9$k(B */
/* $B:G8e$N(B1$BC18l$NA08~$-%H%l%j%9$r7W;;$9$k(B */
void
scan_word(NODE *now, HTK_Param *param)
{
  int   i,t;
  WORD_ID word;
  int phmmlen;
  
  /* ----------------------- prepare phoneme sequence ------------------ */
  /* triphone$B$J$i@hF,$N(B1$B2;AG$O$3$3$G$OBP>]30(B($B$"$H$G(Bnext_word$B$G$d$k(B) */
  /*             $BKvHx$N(B1$B2;AG$O%3%s%F%-%9%H$K$7$?$,$C$FCV49(B */
  word = now->seq[now->seqnum-1];
#ifdef TCD
    j_printf("w=");
    for(i=0;i<winfo.wlen[word];i++) {
      j_printf(" %s",(winfo.wseq[word][i])->name);
    }
# ifdef CCD_ON_2PASS
    if (ccd_flag) {
      if (now->last_ph != NULL) {
	j_printf(" | %s", (now->last_ph)->name);
      }
    }
# endif/* CCD_ON_2PASS */
    j_printf("\n");
#endif /* TCD */
#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    /* word$B$NKvHx$N2;AG$O4D6-0MB8$K$h$jJQ2=(B */
    if (now->last_ph != NULL) {
      tailph = get_right_context_HMM(winfo.wseq[word][winfo.wlen[word]-1], now->last_ph, &hmminfo);
    } else {
      tailph = winfo.wseq[word][winfo.wlen[word]-1];
    }
    /* $BD9$5(B1$B$NC18l$O$3$3$G$O(Bscan$B$7$J$$(B */
    if (winfo.wlen[word] == 1) {
      now->last_ph = tailph;
#ifdef TCD
      j_printf("suspended as %s\n", (now->last_ph)->name);
#endif
      return;
    }

    /* scan$BHO0O$N2;AGNs$r=`Hw(B */
    phmmlen = winfo.wlen[word] - 1;
    if (phmmlen > phmmlen_max) {
      j_printerr("short of phmmlen\n");
      exit(1);
    }
    for (i=0;i<phmmlen-1;i++) {
      phmmseq[i] = winfo.wseq[word][i+1];
    }
    phmmseq[phmmlen-1] = tailph;
  } else {
    phmmlen = winfo.wlen[word]; 
    for (i=0;i<phmmlen;i++) phmmseq[i] = winfo.wseq[word][i];
  }
#else  /* not CCD_ON_2PASS */
  phmmlen = winfo.wlen[word]; 
  for (i=0;i<phmmlen;i++) phmmseq[i] = winfo.wseq[word][i];
#endif /* CCD_ON_2PASS */

  /* $B85$N(Bg[]$B$r$$$C$?$sBTHr$7$F$*$/(B */
  for (t=0;t<peseqlen;t++) g[t] = now->g[t];

  /* viterbi$B$r<B9T$7$F(B g[] $B$+$i(B now->g[] $B$r99?7$9$k(B */
  do_viterbi(g, now->g, phmmseq, phmmlen, param, peseqlen, now->estimated_next_t);

#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    /* $B<!2s$N$?$a$K(B now->last_ph $B$r99?7(B */
    now->last_ph = winfo.wseq[word][0];
#ifdef TCD
    j_printf("last_ph = %s\n", (now->last_ph)->name);
#endif
  }
#endif
}

/* $B2>@b(B now $B$+$i!$C18l$rIU$12C$($?2>@b(B new $B$r?7$?$K:n$k(B */
/* triphone$B$N>l9g(B, $BE83+85$N2>@b$N:G8e(B($B@hF,(B)$B$N(B1$B2;AGJ,$r(B */
/*                 $BE83+C18l$4$H$K(B scan $B$9$k(B */
static HMM_Logical *lastphone, *newphone;
static LOGPROB *g_src;
void
next_word(
     NODE *now, NODE *new,
     NEXTWORD *nword,
     HTK_Param *param,
     BACKTRELLIS *backtrellis)
{
  int   t;
  int lastword;
  int   i;
  LOGPROB a_value;
  int   startt;
  int word;
  LOGPROB totalscore;
  TRELLIS_ATOM *tre;

  word = nword->id;
  lastword = now->seq[now->seqnum-1];

#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    /* $B@\B3$9$kC18l$K0MB8$7$?(B triphone $B$r8!:w(B */
    lastphone = get_left_context_HMM(now->last_ph, winfo.wseq[word][winfo.wlen[word]-1], &hmminfo);
  }
#endif /* CCD_ON_2PASS */

  newphone = winfo.wseq[word][winfo.wlen[word]-1];
#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    newphone = get_right_context_HMM(newphone, now->last_ph, &hmminfo);
  }
#endif
  
  
  /*------------------ new $B$X2>@b>pJs$r%3%T!<(B --------------*/
  new->score = LOG_ZERO;
  for (i=0;i< now->seqnum;i++){
    new->seq[i] = now->seq[i];
  }
  new->seq[i] = word;
  new->seqnum = now->seqnum+1;
  new->totallscore = now->totallscore + nword->lscore;
#ifdef CCD_ON_2PASS
  if (ccd_flag) {
    /* $B<!2>@b$NMzNr>pJs$H$7$FJ]B8(B */
    new->last_ph = lastphone;
  }
#endif /* CCD_ON_2PAS */

  startt = peseqlen-2;
#ifdef CCD_ON_2PASS
  /* --------------------- 1$B2;AGJ,(Bscan ----------------------------- */
  if (ccd_flag) {
    /* $B:G8e$N(B1$B2;AG(B(lastphone)$BJ,$r(Bscan$B$7!$99?7$7$?%9%3%"$r(B */
    /* new $B$KJ]B8(B */
    do_viterbi_next_word(now, new, lastphone, param);
    g_src = new->g;
  } else {
    g_src = now->g;
  }
#else  /* ~CCD_ON_2PASS */
  g_src = now->g;
#endif
      
  /* ---------------- $B<!2s$N(B scan $B$KHw$($F(B new->g[] $B$r=`Hw(B ----------*/
#ifdef LOGTABLE
  a_value = newphone->def->tr->a[newphone->def->state_num-2][newphone->def->state_num-1];
#else
  a_value = log10(newphone->def->tr->a[newphone->def->state_num-2][newphone->def->state_num-1]);
#endif
  for(t=0; t <= startt; t++) {
    new->g[t] = g_src[t+1] + a_value
      + nword->lscore
      ;
  }

  /* --------$B<!C18l$N%R%e!<%j%9%F%#%C%/$H@\B3$7$F2>@b%9%3%"$r5a$a$k(B -- */
#ifdef WORD_GRAPH
  /* $BF1$8C18l$G$b8uJdKh$K%9%3%"$,0c$&(B */
  /* bestt $B$O(B $B8GDj(B */
  /* $B:G8e$K;2>H$7$?(BTRELLIS_ATOM$B$N=*C<$K8GDj(B */
  new->bestt = (nword->tre)->endtime;
  /* new$B$N?dDj=*C<;~4V$O!$>e5-$N;OC<(B */
  new->estimated_next_t = (nword->tre)->begintime - 1;
  new->score = new->g[new->bestt] + (nword->tre)->backscore
#ifdef DISCOUNT_BACKSCORE
    * discount_backscore_rate
#endif
    + outprob(new->bestt, newphone->def->s[newphone->def->state_num-2], param);
  new->tre  =  nword->tre;
  
#else  /* not WORD_GRAPH */
  
  /* $B:GL`C<$r:F7W;;(B */
  /* $B:G8e$K;2>H$7$?(BTRELLIS_ATOM$B$N=*C<;~4V$NA08e(B */
  /* new$B$N?dDj;~4V$O!$>e5-$G:NMQ$7$?(BTRELLIS_ATOM$B$N;OC<;~4V(B */

  /* $BE83+$7$?%H%l%j%9>e$NC18l$N=*C<$NA08e$N$_$G(Bscan$B$9$k(B */
  /* 1. $BA0J}$X8!:w(B */
  for(t = (nword->tre)->endtime; t >= 0; t--) {
    tre = bt_binsearch_atom(backtrellis, t, (WORD_ID) word);
    if (tre == NULL) break;	/* $B%H%l%j%9$K;D$C$F$J$1$l$P=*N;(B */
    totalscore = new->g[t] + tre->backscore
#ifdef DISCOUNT_BACKSCORE
      * discount_backscore_rate
#endif
      + outprob(t, newphone->def->s[newphone->def->state_num-2], param);
    if (new->score < totalscore) {
      new->score = totalscore;
      new->bestt = t;
      new->estimated_next_t = tre->begintime - 1;
    }
  }
  /* 2. $B8eJ}$X8!:w(B */
  for(t = (nword->tre)->endtime + 1; t <= startt; t++) {
    tre = bt_binsearch_atom(backtrellis, t, (WORD_ID) word);
    if (tre == NULL) break;	/* $B%H%l%j%9$K;D$C$F$J$1$l$P=*N;(B */
    totalscore = new->g[t] + tre->backscore 
#ifdef DISCOUNT_BACKSCORE
      * discount_backscore_rate
#endif
      + outprob(t, newphone->def->s[newphone->def->state_num-2], param);
    if (new->score < totalscore) {
      new->score = totalscore;
      new->bestt = t;
      new->estimated_next_t = tre->begintime - 1;
    }
  }

  
#endif /* WORD_GRAPH */
}

/* 1$BC18l$+$i$J$k2>@b$r?7$?$K:n$k(B */
void
start_word(
     NODE *new,
     NEXTWORD *nword,
     HTK_Param *param,
     BACKTRELLIS *backtrellis)
{
  HMM_Logical *newphone;
  WORD_ID word;
  TRELLIS_ATOM *tre = NULL;
  int t;

  /* $B=i4|2=(B */
  word = nword->id;
  new->score = LOG_ZERO;
  new->seqnum = 1;
  new->seq[0] = word;

  new->totallscore = nword->lscore;

  newphone = winfo.wseq[word][winfo.wlen[word]-1];
#ifdef CCD_ON_2PASS
  /* $B=i4|2>@b$G$O2;AG4D6-$O9MN8$9$kI,MW$,$J$$(B */
  if (ccd_flag) {
    new->last_ph = NULL;
  }
#endif /* CCD_ON_2PASS */
  
  new->g[peseqlen-1] = nword->lscore;
  
#ifdef WORD_GRAPH
  new->score = new->g[peseqlen-1] + (nword->tre)->backscore
#ifdef DISCOUNT_BACKSCORE
    * discount_backscore_rate
#endif
    ;
  new->tre  =  nword->tre;
  new->bestt = peseqlen-1;
  new->estimated_next_t = (nword->tre)->begintime - 1;
  
#else
  
  for (t=peseqlen-1; t>=0; t--) {
    tre = bt_binsearch_atom(backtrellis, t, word);
    if (tre != NULL) {
      new->bestt = t;
      new->score = new->g[peseqlen-1] + tre->backscore
#ifdef DISCOUNT_BACKSCORE
	* discount_backscore_rate
#endif
        + outprob(peseqlen-1, newphone->def->s[newphone->def->state_num-2], param);
      new->estimated_next_t = tre->begintime - 1;
      break;
    }
  }
  if (tre == NULL) {		/* no word in backtrellis */
    new->score = LOG_ZERO;
  }
#endif /* WORD_GRAPH */
}

/* $B:G=*2>@b$N%9%3%"7W;;(B --- $B<uM}2>@b$N:G=*%9%3%"$r@_Dj(B */
void
last_next_word(NODE *now, NODE *new, HTK_Param *param)
{
  cpy_node(new, now);
  if (ccd_flag) {
    do_viterbi_next_word(now, new, now->last_ph, param);
    new->score = new->g[0];
  } else {
    new->score = now->g[0];
  }
}

#endif /* PASS2_STRICT_IWCD */
