/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* $Id: */

/* trevis_x11.c --- Trellis Visualization functions for X Window System */
/* a quick hack, old style functions, uses raw Xlib functions */
/* should work on X11R5 or later */

#include <julius.h>

#ifdef TRELLIS_VISUALIZE

#include <X11/Xlib.h>
#include <X11/Xutil.h>


/* parameter */
static int swidth = 750;	/* canvas width */
static int sheight = 400;	/* canvas height */
static int smergin = 10;	/* screen mergin */
static int swordheight = 20;
#define WTITLE "Word Trellis Index View"	/* title of window name */

/* Window Info */
static Display *dis = NULL;
static Window win;
static XSetWindowAttributes     att;
static XWMHints xwmh;
static XSizeHints  xsh;        /* Size set for window manager */

/* GC to draw */
static GC gcgraph, gcword;

/* Color handling */
/* #define NCOLS 30	*/	/* num of colors to use for gradation*/
#define MAXRGBVAL 0xffff	/* max rgb value */
static Colormap cmap;
static unsigned long c_border, c_canvas, c_begin, c_end;
static unsigned long colors[10];

/* fonts */
static Font word_kfont, word_afont; /* fonts */
#define TITLE_KFONT "k14"
#define TITLE_AFONT "a14"


/* Allocate Color for given R,G,B value */
static unsigned long
MyColor(Display *display, int r, int g, int b)
{
  XColor col;

  col.red = r;
  col.green = g;
  col.blue = b;
  col.flags = (DoGreen | DoRed | DoBlue);
  if (!XAllocColor(dis, cmap, &col)) {
    j_printerr("can't allocate color for R,G,B=%4x,%4x,%4x. use White insted\n",
	    r, g, b);
    return(WhitePixel(dis, DefaultScreen(dis)));
  } else {
    return(col.pixel);
  }
}

/* Allocate Color for given name */
static unsigned long
MyNamedColor(Display *display, char *p)
{
  XColor col1, col2;
    
  if (!XAllocNamedColor(dis, cmap, p, &col2, &col1)) {
    j_printerr("can't allocate color name %s. use White insted\n", p);
    return(WhitePixel(dis, DefaultScreen(dis)));
  } else {
    return(col2.pixel);
  }
}

/* allocate/free GC */
static void
x11_alloc_GCs()
{
  gcgraph = XCreateGC(dis,win,0,0);
  XSetLineAttributes(dis, gcgraph,
		     1, LineSolid, CapProjecting, JoinMiter);
  gcword = XCreateGC(dis,win,0,0);
  XSetFillStyle(dis,gcword,FillSolid);
  word_afont = XLoadFont(dis,TITLE_AFONT);
  XSetFont(dis,gcword,word_afont);
  
}
static void
x11_free_GCs()
{
  XFreeGC(dis,gcgraph);
  XFreeGC(dis,gcword);
}

/* Draw String which may include Multi-Byte Japanese Character */
/* kanji-code must be EUC */
/* args are like "XDrawString" with "kfont, kfont_width" added */
static void
XDrawMultiString(Display *display, Drawable d, GC gc,
		 int x, int y, char *string,
		 Font afont, int afont_width, Font kfont, int kfont_width)
{
  XChar2b       kanji;
  char *str;
  int which_flag;

  which_flag = 0;
  XSetFont(dis,gc,kfont);
  str = string;
  while ( *str ) {
    if ( *str & 0x80 ) {
      /* $B4A;z(B */
      kanji.byte1 = *str++ & 0x7f;
      kanji.byte2 = *str & 0x7f;
      if ( *str )
        *str++;
      if (which_flag == 1) {
        XSetFont(dis,gc,kfont);
        which_flag = 0;
      }
      XDrawString16( display, d, gc, x, y, &kanji, 1 );
      x+=kfont_width;
    } else {
      /* $BHs4A;z(B */
      if (which_flag == 0) {
        XSetFont(dis,gc,afont);
        which_flag = 1;
      }
      XDrawString( display, d, gc, x, y, str, 1 );
      str++;
      x+=afont_width;
    }
  }
}



/* initialize X11 environment */
void
TrevisInit()
{
  int i;

  /* open DISPLAY */
  if ((dis=XOpenDisplay(NULL)) == NULL) {
    j_printerr("can't open display. X11 feature disabled\n");
    trevis_flag = FALSE;
    return;
  }
  /* alloc colors */
  cmap = DefaultColormap(dis, 0);
#if 0
  for(i=0;i<NCOLS;i++) {
    grad[i] = MyColor(dis, 0, (i * MAXRGBVAL) / NCOLS, 0);
  }
#endif
  c_border = MyNamedColor(dis, "orange");
  c_canvas = MyNamedColor(dis, "navyblue");
  c_begin = MyNamedColor(dis, "yellow");
  c_end = MyNamedColor(dis, "green");
  colors[0] = MyNamedColor(dis, "DarkKhaki"); /* basic bt */
  colors[1] = MyNamedColor(dis, "orange");	/* best path */
  colors[2] = MyNamedColor(dis, "red"); /* specified word */

  /* create window */
  xsh.flags  = (PPosition | PSize );
  xsh.width  = swidth + smergin * 2;
  xsh.height = sheight + smergin * 2 + swordheight;
  xsh.x      = 20;
  xsh.y      = 20;
  win = XCreateSimpleWindow(dis, DefaultRootWindow(dis),
			    xsh.x, xsh.y, xsh.width, xsh.height,
			    2, /*border width*/
			    c_border, c_canvas); /* border color,fg */
  /* set window manager hint */
  xwmh.flags = (InputHint|StateHint);
  xwmh.input = True;
  xwmh.initial_state = NormalState;
  XSetStandardProperties(dis, win,
			 WTITLE,     /* Window name */
			 WTITLE,     /*  icon  name */
			 None,          /* icon pixmap */
			 NULL,0,     /* argc, argv */
			 &xsh);         /* normal hints */
  XSetWMHints(dis, win, &xwmh);
  att.backing_store = Always;
  XChangeWindowAttributes(dis, win, CWBackingStore, &att);
  XMapWindow(dis, win);
  x11_alloc_GCs();
  VERMES("X11 init ok\n");
}


/*************************/
/*** Drawing Functions ***/
/*************************/

static int xmax;
static float *ymin, *ymax;
static float yrange;
static boolean scaled=FALSE;

/* prepare for drawing (re-allocate GC)*/
void
TrevisPrepare()
{
  x11_free_GCs();
  XClearWindow(dis,win);
  x11_alloc_GCs();
  if (scaled) {
    scaled = FALSE;
    free(ymin);
    free(ymax);
  }
}

/* compute scaling factors */
static void
calc_scaler(BACKTRELLIS *bt)
{
  int t,i;
  float y;
  /* compute scale factor */
  xmax = bt->framelen;
  ymin = (float *)mymalloc(sizeof(float) * bt->framelen);
  ymax = (float *)mymalloc(sizeof(float) * bt->framelen);
    
  for (t=0;t<bt->framelen;t++) {
    ymin[t] = - (LOG_ZERO);
    ymax[t] = 0.0;
    for (i=0;i<bt->num[t]; i++) {
      y = - (bt->rw[t][i])->backscore;
      if (ymax[t] < y + 1.0) {
	ymax[t]  = y + 1.0;
      }
      if (ymin[t] > y) {
	ymin[t]  = y;
      }
    }
  }
  yrange = 0.0;
  for (t=20;t<bt->framelen;t++) {
    if (ymax[t] != 0) {
      if (yrange < ymax[t] - ymin[t]) yrange = ymax[t]-ymin[t];
    }
  }
  /*yrange /= 5.0;*/
}

static int
scale_x(int x)
{
  return(x * swidth  / xmax + smergin);
}
static int
scale_y(float y, int t)
{
  return((y - ymin[t]) * (float)sheight / yrange + smergin + swordheight);
}

static void
set_coordinates(TRELLIS_ATOM *ttmp, int t, BACKTRELLIS *bt,
		int *x1, int *y1, int *x2, int *y2)
{
  if (scaled == FALSE) {
    calc_scaler(bt);
    scaled = TRUE;
  }

  *x2 = scale_x(t);
  *y2 = scale_y(- (ttmp->backscore), t);
  if (ttmp->begintime > 0) {
    *x1 = scale_x(ttmp->begintime-1);
    *y1 = scale_y(- (ttmp->last_tre->backscore), ttmp->last_tre->endtime);
  } else {
    *x1 = smergin;
    *y1 = smergin + swordheight;
  }
}

void
TrevisPutAll(BACKTRELLIS *bt)
{
  int t, i;
  static int x1, y1, x2, y2;
  unsigned long col;

  /* put lines & ends */
  XSetLineAttributes(dis, gcgraph,
		     1, LineSolid, CapProjecting, JoinMiter);
  for (t = bt->framelen-1; t >= 0; t--) {
    for (i=0;i<bt->num[t]; i++) {
      set_coordinates(bt->rw[t][i], t, bt, &x1, &y1, &x2, &y2);
      XSetForeground(dis, gcgraph, colors[0]);
      XDrawLine(dis,win,gcgraph, x1, y1, x2, y2);
      XSetForeground(dis, gcgraph, c_end);
      XFillRectangle(dis,win,gcgraph, x2-1, y2-1, 3, 3);
    }
  }
  /* put beginning-of-word */
  for (t = bt->framelen-1; t >= 0; t--) {
    for (i=0;i<bt->num[t]; i++) {
      set_coordinates(bt->rw[t][i], t, bt, &x1, &y1, &x2, &y2);
      XSetForeground(dis, gcgraph, c_begin);
      XFillRectangle(dis,win,gcgraph, x1-2, y1-2, 5, 5);
    }
  }

  XFlush(dis);
}

void
TrevisPutWord(TRELLIS_ATOM *atom, int t, int cidx, BACKTRELLIS *bt)
{
  static  int x1, y1, x2, y2;

  XSetLineAttributes(dis, gcgraph,
		     1, LineSolid, CapProjecting, JoinMiter);

  set_coordinates(atom, t, bt, &x1, &y1, &x2, &y2);
  XSetForeground(dis, gcgraph, colors[cidx]);
  XFillRectangle(dis,win,gcgraph, x2-2, y2-2, 5, 5);
  XDrawLine(dis,win,gcgraph, x1, y1, x2, y2);
  XFillRectangle(dis,win,gcgraph, x1-3, y1-3, 7, 7);
  XFlush(dis);
}
  
void
TrevisPutPathTrail(TRELLIS_ATOM *atom, int t, BACKTRELLIS *bt, int cidx, int width)
{
  int x1,x2,y1,y2;

  j_printf("%d: ",t);
  put_voca(&winfo, atom->wid);

  XSetLineAttributes(dis, gcgraph,
		     width, LineSolid, CapProjecting, JoinMiter);

  if (scaled == FALSE) {
    calc_scaler(bt);
    scaled = TRUE;
  }

  x2 = scale_x(t);
  y2 = scale_y(- (atom->backscore), t);
  if (atom->begintime > 0) {
    x1 = scale_x(atom->begintime-1);
    y1 = scale_y(- (atom->last_tre->backscore), atom->last_tre->endtime);
  } else {
    x1 = smergin;
    y1 = smergin + swordheight;
  }

  /* put lines */
  XSetForeground(dis, gcgraph, colors[cidx]);
  XFillRectangle(dis,win,gcgraph, x2-1, y2-1, 3, 3);
  XDrawLine(dis,win,gcgraph, x1, y1, x2, y2);
  XFillRectangle(dis,win,gcgraph, x1-3, y1-3, 7, 7);

  /* put word seq */
#if 0
  {
    int tx1;
    int i;
    int plen;

    XDrawLine(dis,win,gcword, x1, smergin, x1, smergin+swordheight-6);
    tx1 = x1 + 4;
    for (i = 0; i < winfo.wlen[atom->wid]; i++) {
      plen = strlen(winfo.wseq[atom->wid][i]->name);
      XDrawString(dis, win, gcword, tx1, smergin+(swordheight-12),
		  winfo.wseq[atom->wid][i]->name, plen);
      tx1 += 7 * plen + 4;
    }
  }
#endif

  /* draw next (recursive call) */
  if (atom->begintime > 0) {
    TrevisPutPathTrail(atom->last_tre, atom->last_tre->endtime,bt,cidx,width);
  } else {
    XFlush(dis);
  }
}

#endif /* TRELLIS_VISUALIZE */
