/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* wchmm.c --- word-conjunction HMM */

/* $Id: wchmm.c,v 1.2 1998/09/02 15:46:55 ri Exp $ */

#include <julius.h>


static int bogus_arc = 0;	/* reduced arc by tree*/
static int homophone_word_num;	/* number of homophone words in vocabulary */
static int dupcount = 0;


/* $B=i4|2=(B */
static void
wchmm_init(WCHMM_INFO *wchmm)
{
  wchmm->maxwcn = MAXWCNSTEP;
  wchmm->state = (WCHMM_STATE *)mymalloc(sizeof(WCHMM_STATE)*wchmm->maxwcn);
  wchmm->ststart = (WORD_ID *)mymalloc(sizeof(WORD_ID)*wchmm->maxwcn);
  wchmm->stend = (WORD_ID *)mymalloc(sizeof(WORD_ID)*wchmm->maxwcn);
  /* no need to expand */
  wchmm->offset = (int **)mymalloc(sizeof(int *)*winfo.num);
  wchmm->wordend = (int *)mymalloc(sizeof(int)*winfo.num);
  wchmm->startnode = (int *)mymalloc(sizeof(int)*winfo.num);
  wchmm->wordend_a = (LOGPROB *)mymalloc(sizeof(LOGPROB)*winfo.num);
  
  wchmm->n = 0;
}
/* $BLZ$NBg$-$5$K1~$8$FNN0h$r?-D9(B */
static void
wchmm_extend(WCHMM_INFO *wchmm)
{
  wchmm->maxwcn += MAXWCNSTEP;
  wchmm->state = (WCHMM_STATE *)myrealloc(wchmm->state, sizeof(WCHMM_STATE)*wchmm->maxwcn);
  wchmm->ststart = (WORD_ID *)myrealloc(wchmm->ststart, sizeof(WORD_ID)*wchmm->maxwcn);
  wchmm->stend = (WORD_ID *)myrealloc(wchmm->stend, sizeof(WORD_ID)*wchmm->maxwcn);
}

/* $BC18l$r2;AG$N$J$i$S$G%=!<%H$9$k(B */
static int
compare_wseq(WORD_ID *widx1, WORD_ID *widx2)
{
  int len1, len2, n;
  int p=0;
  
  len1 = winfo.wlen[*widx1];
  len2 = winfo.wlen[*widx2];

  n=0;
  /*  while (n < len1 && n < len2 && (p = (int)winfo.wseq[*widx1][n] - (int)winfo.wseq[*widx2][n]) == 0 ) n++;*/
  while (n < len1 && n < len2 && (p = strcmp((winfo.wseq[*widx1][n])->name, (winfo.wseq[*widx2][n])->name)) == 0 ) n++;
  if (n < len1) {
    if (n < len2) {
      /* differ */
      return(p);
    } else {
      /* 2 is part of 1 */
      return(1);
    }
  } else {
    if (n < len2) {
      /* 1 is part of 2 */
      return(-1);
    } else {
      /* same */
      return(0);
    }
  }
}
static void
wchmm_sort_idx_by_wseq(WORD_ID *windex, WORD_ID bgn, WORD_ID len)
{
  qsort(&(windex[bgn]), len, sizeof(WORD_ID), (int (*)(const void *, const void *))compare_wseq);
}

  

/* 2$BC18l$N@hF,$b$7$/$OKvHx$+$i$N6&M-2;AG?t$rD4$Y$k(B
   return value: match length of phmm */

static int		   /* $BC18l(B(ID=i)$B$HC18l(B(ID=j)$B$N%^%C%A%s%0$rD4$Y$k(B */
wchmm_check_match(int i, int j)
{
  int k,tmplen;

  for (tmplen=0,k=0;k<winfo.wlen[i];k++) {
    if (k > winfo.wlen[j]-1)
      break;
    if (! (strmatch(winfo.wseq[i][k]->name, winfo.wseq[j][k]->name)))
      break;
    tmplen++;
  }
  return(tmplen);
}

/*
 * arc $B$rDI2C(B
 */
static void
add_wacc(int node, LOGPROB a, int arc)
{
  A_CELL *ac;
  ac       = (A_CELL *) mymalloc(sizeof(A_CELL));
  ac->a    = a;
  ac->arc  = arc;
  ac->next = wchmm.state[node].ac;
  wchmm.state[node].ac   = ac;
}  

/*
 * $B2;AG(BHMM$B$I$&$7$r$D$J$0(B
 */
static void
wchmm_link_hmm(int from_node, int to_node, HTK_HMM_Trans *tinfo)
{     
  A_CELL *actmp;
  LOGPROB a, atmp;
  int i;
  boolean tflag;

  /* endstate$B$X$N(Blink$B$r8!:w(B */
  for(i = tinfo->statenum - 2; i >= 0; i--) {
#ifdef LOGTABLE
    if ((a = tinfo->a[i][tinfo->statenum-1]) != LOG_ZERO) {
      atmp = a;
#else
    if ((a = tinfo->a[i][tinfo->statenum-1]) != 0.0) {
      atmp = log10(a);
#endif
      /* $B;^$NEPO?$N=EJ#$rHr$1$k(B */
      tflag = FALSE;
      for (actmp = wchmm.state[from_node].ac; actmp; actmp = actmp->next) {
	if (actmp->arc == to_node && actmp->a == atmp) {
	  bogus_arc++;
	  tflag = TRUE;
	  break;
	}
      }
      if (tflag) break;

      /* link$B$rEPO?(B */
      add_wacc(from_node, atmp, to_node);
      return;
    }
  }      
  j_printerr("Error: No arc to endstate?\n");
  exit(1);
}
/*
 * $BC18l(B[from_word]$B$N(B[from_seq]$BHVL\$N2;AG$N:G8e$H(B
 * $BC18l(B[ to_word ]$B$N(B[ to_seq ]$BHVL\$N2;AG$N:G=i$r$D$J$0!%(B
 */
static void
wchmm_link_subword(int from_word, int from_seq, int to_word, int to_seq)
{     
  /* matchword[add_to]$B$N=*C<%N!<%I$H(Bword[add_head]$B$N;OC<$r$D$J$0(B */
  /* int last,lastp; */
  HMM_Logical *last;
  int lastp;

  last=winfo.wseq[from_word][from_seq];
  lastp=wchmm.offset[from_word][from_seq] + last->def->state_num-2 -1;

  wchmm_link_hmm(lastp, wchmm.offset[to_word][to_seq],
		 last->def->tr);
}

/*
 * node $B$r(B word $B$N:G8e(B/$B:G=i$H$7$FFs=E2=$9$k(B
 */
static void
wchmm_duplicate_state(
     int node,			/* $BFs=E2=$9$k%N!<%I(B */
     int word)			/* $BFs=E2=$9$kC18l(B */
{
  int n;
  int n_src, n_prev;
  A_CELL	*ac;
  HMM_Logical *lastphone;
  
  /* $BLZ$N:GBg%5%$%:$rI,MW$K1~$8$F?-D9(B */
  if (wchmm.n + 1 >= wchmm.maxwcn) {
    wchmm_extend(&wchmm);
  }
  /* $BG[Ns$N:G8e$KIU$12C$($k(B */
  n = wchmm.n;

  n_src = node;

  /* $B=PNO>pJs$N%3%T!<(B */
#ifdef PASS1_IWCD
  {
    RC_INFO *rcnew;
    LRC_INFO *lrcnew;
    wchmm.state[n].outstyle = wchmm.state[n_src].outstyle;
    if (wchmm.state[n].outstyle == AS_RSET) {
      /* duplicate RC_INFO because it has its own cache */
      rcnew = (RC_INFO *)mymalloc(sizeof(RC_INFO));
      memcpy(rcnew, wchmm.state[n_src].out.rset, sizeof(RC_INFO));
      wchmm.state[n].out.rset = rcnew;
    } else if (wchmm.state[n].outstyle == AS_LRSET) {
      /* duplicate LRC_INFO because it has its own cache */
      lrcnew = (LRC_INFO *)mymalloc(sizeof(LRC_INFO));
      memcpy(lrcnew, wchmm.state[n_src].out.lrset, sizeof(LRC_INFO));
      wchmm.state[n].out.lrset = lrcnew;
    } else {
      /* share same info, simply copy the pointer */
      memcpy(&(wchmm.state[n].out), &(wchmm.state[n_src].out), sizeof(ACOUSTIC_SPEC));
    }
  }
#else  /* ~PASS1_IWCD */
  memcpy(&(wchmm.state[n].out), &(wchmm.state[n_src].out), sizeof(HTK_HMM_State *));
#endif

  /* $B<+8JA+0\$N%3%T!<(B */
  wchmm.state[n].ac = NULL;
  for(ac=wchmm.state[n_src].ac; ac; ac=ac->next) {
    if (ac->arc == n_src) {
      add_wacc(n, ac->a, n);
    }
  }
  /* $B>uBV4V$NA+0\$N%3%T!<(B */
  /* $B0l$DA0$N>uBV$+$i?7>uBV$X$NA+0\$rIU$12C$($k(B */
  lastphone = winfo.wseq[word][winfo.wlen[word]-1]; /* n_src $B$O$3$l$N:G=*>uBV$H$$$&A0Ds$"$j(B */
  
  if (lastphone->def->state_num == 3) { /* $B<B<A(B1 */
    if (winfo.wlen[word] == 1) { /* $B$3$N2;AG$@$1$+$i$J$kC18l(B */
      /* $B%3%T!<$7$J$/$FNI$$(B */
      /* $B$3$N%N!<%I$OF1;~$KC18l$N@hF,%N!<%I$G$b$"$k(B */
      wchmm.ststart[n] = word;
      wchmm.offset[word][0] = n;
    } else {
      /* $B$^$($N2;AG$N:G8e$N>uBV$+$i$NA+0\$r%3%T!<(B */
      n_prev = wchmm.offset[word][winfo.wlen[word]-2]
	+ winfo.wseq[word][winfo.wlen[word]-2]->def->state_num - 3;
      for (ac=wchmm.state[n_prev].ac; ac; ac=ac->next) {
	if (ac->arc == n_src) {
	  add_wacc(n_prev, ac->a, n);
	}
      }
      /* $BC18l$N@hF,%N!<%I$G$O$"$jF@$J$$(B */
      wchmm.ststart[n] = WORD_INVALID;
    }
  } else {
    /* $B$=$N>uBV$r4^$`2;AGFb$N!$$=$N>uBV$X$NA+0\D>A0$N>uBV$+$i%3%T!<(B */
    for (n_prev = wchmm.offset[word][winfo.wlen[word]-1]; n_prev < n_src; n_prev++) {
      for (ac=wchmm.state[n_prev].ac; ac; ac=ac->next) {
	if (ac->arc == n_src) {
	  add_wacc(n_prev, ac->a, n);
	}
      }
    }
    /* $BC18l$N@hF,%N!<%I$G$O$"$jF@$J$$(B */
    wchmm.ststart[n] = WORD_INVALID;
  }

  /* $BC18l(B<->$B>uBV$NBP1~(B */
  wchmm.stend[n]   = word;
  wchmm.wordend[word] = n;

  /* $B>uBV$O(B1$B8DA}$($?(B
   */
  wchmm.n++;
  
}

/*
 * $B3FC18l$,(Bunique $B$J(B leaf node $B$r;}$D$h$&!$(B2$B=E2=$r9T$&(B
 *  $B%Q%?!<%s(B1: $BF12;8l(B
 *  $B%Q%?!<%s(B2: $BJL$NC18l$N0lIt(B
 */

static void
wchmm_duplicate_leafnode()
{
  int w, nlast, n;
  boolean *dupw;
  A_CELL *actmp;

  nlast = wchmm.n;
  dupw = (boolean *)mymalloc(sizeof(boolean) * nlast);
  for(n=0;n<nlast;n++) dupw[n] = FALSE;

  for (w=0;w<winfo.num;w++) {
    n = wchmm.wordend[w];
    if (dupw[n]) {		/* 2$B$D$a0J9_(B */
      wchmm_duplicate_state(n, w); dupcount++;
    } else {			/* $B=i2s(B */
      for(actmp=wchmm.state[n].ac;actmp;actmp=actmp->next) {
	if (actmp->arc != n) {	/* $BB>>uBV$X$NA+0\$,$"$l$P(B */
	  /* $BC18l$N0lIt$,JL$NC18l$K$J$C$F$$$k(B */
	  /* $B=E$J$C$F$$$k?t$@$1A4ItB?=E2=(B */
	  wchmm_duplicate_state(n, w); dupcount++;
	  /* $BC18l=*C<%U%i%0$r%j%;%C%H(B */
	  wchmm.stend[n] = WORD_INVALID;
	  break;
	}
      }
      if (actmp == NULL) {	/* $B1&$X$NA+0\$,B8:_$7$J$$(B */
	/* $BJL$NC18l$KKd$a9~$^$l$F$O$$$J$$(B */
	/* $B=E$J$C$F$$$k?t(B -1 $B$rB?=E2=$9$k(B */
	/* $B=i2s$O(Bstend$B$NIU$1$+$($@$1$G:Q$`(B */
	wchmm.stend[n] = w;
      }
      dupw[n] = TRUE;
    }
  }
  free(dupw);
}


/*
 * $BC18l(B word ($BC18l(Bmatchword$B$H(Bmatchlen$B?t$N2;AG?t%^%C%A$9$k(B)$B$r(B wchmm $B$KIU$12C$($k(B */
static void
wchmm_add_word(
     int word,			/* $BIU$12C$($k%+%F%4%jFb$NC18lHV9f(B */
     int matchlen,		/* $B%^%C%A$9$kD9$5(B(0:$BA4ItIU$12C$($k(B) */
     int matchword)		/* $B:G$bD9$/%^%C%A$9$kEPO?:Q$_C18lHV9f(B */
{
  int   j,k,n;
  int   add_head, add_tail, add_to;
  int   word_len, matchword_len;
  
/* 
 *   printf("--\n");
 *   put_voca(&winfo, word);
 *   put_voca(&winfo, matchword);
 *   printf("matchlen=%d\n", matchlen);
 */
  
  /* variable abbreviations */
  n = wchmm.n;
  word_len      = winfo.wlen[word];
  matchword_len = winfo.wlen[matchword];
  
  if((wchmm.offset[word]=(int *)mymalloc(sizeof(int)*word_len))==NULL){ 
    j_printerr("malloc failed at wchmm_add_word()\n");
    exit(1);
  }
  
  /* $B?7$?$KIU$12C$($kItJ,$r7hDj(B */
  /* word[matchlen...word_len-1] -> matchword[matchlen-1]  */
  add_head = matchlen;
  add_tail = word_len - 1;
  add_to   = matchlen - 1;
  
  if (add_tail - add_head + 1 > 0) {
    /* $B?7$?$JItJ,$r(B wchmm $B$K2C$($k(B */
    {
      HMM_Logical *ltmp;
      int ato;
      LOGPROB prob;
      int ntmp = n;
#ifdef PASS1_IWCD
      LC_Set *lcd = NULL;
#endif
      
      ntmp = n;
      for (j=add_head; j <= add_tail; j++) { /* $B3F2;AG(BHMM */
	ltmp = winfo.wseq[word][j];
#ifdef PASS1_IWCD
	if (ccd_flag) {
	  if (winfo.wlen[word] > 1 && j == winfo.wlen[word] - 1) { /* $B:G8e$N2;AG$J$i(B */
	    if ((lcd = lcdset_lookup_by_hmmname(&hmminfo, ltmp->name)) == NULL) {
	      j_printerr("Error: no lcdset found for [%s]\n",ltmp->name);
	      exit(1);
	    }
	  }
	}
#endif /* PASS1_IWCD */
	for (k = 1; k < ltmp->def->state_num - 1; k++) { /* $B3F>uBV(B */
	  /* $B=PNO$N@_Dj(B */
#ifdef PASS1_IWCD
	  if (ccd_flag) {
	    if (winfo.wlen[word] == 1) { /* 1$B2;AG$N$_$NC18l(B */
	      wchmm.state[ntmp].outstyle = AS_LRSET;
	      wchmm.state[ntmp].out.lrset = (LRC_INFO *)mymalloc(sizeof(LRC_INFO));
	      (wchmm.state[ntmp].out.lrset)->hmm       = ltmp;
	      (wchmm.state[ntmp].out.lrset)->state_loc = k;
	    } else if (j == 0) {	/* $B:G=i$N2;AG(B */
	      wchmm.state[ntmp].outstyle = AS_RSET;
	      wchmm.state[ntmp].out.rset = (RC_INFO *)mymalloc(sizeof(RC_INFO));
	      (wchmm.state[ntmp].out.rset)->hmm       = ltmp;
	      (wchmm.state[ntmp].out.rset)->state_loc = k;
	    } else if (j == winfo.wlen[word] - 1) { /* $B:G8e$N2;AG(B */
	      wchmm.state[ntmp].outstyle = AS_LSET;
	      wchmm.state[ntmp].out.lset = &(lcd->stateset[k]);
	    } else {
	      wchmm.state[ntmp].outstyle = AS_STATE;
	      wchmm.state[ntmp].out.state = ltmp->def->s[k];
	    }
	  } else {		/* monophone */
	    wchmm.state[ntmp].outstyle = AS_STATE;
	    wchmm.state[ntmp].out.state = ltmp->def->s[k];
	  }
#else  /* ~PASS1_IWCD */
	  wchmm.state[ntmp].out = ltmp->def->s[k];
#endif /* PASS1_IWCD */
	  /* $BA+0\$N@_Dj(B */
	  wchmm.state[ntmp].ac = NULL;
	  for (ato = 1; ato < ltmp->def->state_num; ato++) {
	    prob = ltmp->def->tr->a[k][ato];
	    if
#ifdef LOGTABLE
	      (prob != LOG_ZERO)
#else
	      (prob != 0.0)
#endif
	      {
	      if (j == add_tail && k == ltmp->def->state_num - 2 && ato == ltmp->def->state_num - 1) {
		/* $BItJ,$N:G8e$+$i30$X$NA+0\$O$"$H$G(Blink$B$9$k(B */
	      } else {
#ifdef LOGTABLE
		add_wacc(ntmp, prob, ntmp + ato - k);
#else
		add_wacc(ntmp, log10(prob), ntmp + ato - k);
#endif
	      }
	    }
	  }
	  /* $B$=$NB>$N(Bwchmm$B>pJs$O=i4|2=(B */
	  wchmm.ststart[ntmp] = WORD_INVALID;
	  wchmm.stend[ntmp] = WORD_INVALID;
	  ntmp++;
	  /* $BLZ$N%5%$%:$,8=:_$N@)8B$r1[$($?$i!$3JG<NN0h$r?-D9(B */
	  if (ntmp >= wchmm.maxwcn) wchmm_extend(&wchmm);
        }
      }
    }
  }
  
  /* $BC18l>pJs(B <-> wchmm$B>pJs$NBP1~$r5-O?(B */
  for (j=0;j<word_len;j++) {
    if (j < add_head) {	/* $B@hF,$N6&M-ItJ,(B */
      wchmm.offset[word][j] = wchmm.offset[matchword][j];
    } else if (add_tail < j) { /* $BKvHx$N6&M-ItJ,(B */
      wchmm.offset[word][j] = wchmm.offset[matchword][j+(matchword_len-word_len)];
    } else {			/* $B?7$?$KIU$12C$($?ItJ,(B */
      wchmm.offset[word][j] = n;
      n += winfo.wseq[word][j]->def->state_num - 2;
    }
  }
  wchmm.n = n;
  
  /* $BC18l$N;OC<$rEPO?(B */
  wchmm.ststart[wchmm.offset[word][0]] = word;
  /* $BC18l$N=*C<$rEPO?(B */
  k = wchmm.offset[word][word_len-1] + winfo.wseq[word][word_len-1]->def->state_num-2 -1;
  wchmm.wordend[word] = k;
  wchmm.stend[k] = word;
  
  if (matchlen != 0 && add_tail - add_head + 1 > 0) {
    /* $B6&M-ItJ,$H?75,DI2CItJ,$NA+0\$rIU$12C$($k(B */
    wchmm_link_subword(matchword,add_to,word,add_head);	
  }


}

/*----------------------------------------------------------------------*/
/* $BC18l$N@hF,>uBV$N%$%s%G%C%/%9$r:n@.(B */
static void
wchmm_index_ststart()
{
  int n;
  int id;

  id = 0;
  for (n=0;n<wchmm.n;n++) {
    if (wchmm.ststart[n] != WORD_INVALID) {
      /* $B@hF,C18l$N;OC<$X$OA+0\$5$;$J$$(B=$B%$%s%G%C%/%9$K4^$a$J$$(B */
      if (wchmm.ststart[n] == winfo.head_silwid) continue;
      wchmm.startnode[id] = n;
      id++;
      if (id >= winfo.num) {
	j_printerr("Error: start node num exceeded %d\n", winfo.num);
      }
    }
  }
  wchmm.startnum = id;
}

/* $BC18l$N=*C<>uBV$+$i$N<!A+0\3NN($r5a$a$F$*$/(B */
static void
wchmm_calc_wordend_arc()
{
  WORD_ID w;
  HTK_HMM_Trans *tr;
  LOGPROB a;

  for (w=0;w<winfo.num;w++) {
    tr = winfo.wseq[w][winfo.wlen[w]-1]->def->tr;
    a = tr->a[tr->statenum-2][tr->statenum-1];
#ifdef LOGTABLE
    wchmm.wordend_a[w] = a;
#else
    wchmm.wordend_a[w] = log10(a);
#endif
  }
}

/*----------------------------------------------------------------------*/
#ifdef SEPARATE_BY_UNIGRAM
/* 1-gram$B%9%3%"$N>e0L(B N $BHVL\$NCM$r5a$a$k(B */
static int
compare_prob(LOGPROB *a, LOGPROB *b)
{
  if (*a < *b)  return (1);
  if (*a > *b)  return (-1);
  return(0);
}
static LOGPROB
get_nbest_uniprob(int n)
{
  LOGPROB *u_p;
  WORD_ID w;
  LOGPROB x;

  if (n < 1) n = 1;
  if (n > winfo.num) n = winfo.num;
  
  /* 1-gram$B%9%3%"$GC18l(BID$B$N%$%s%G%C%/%9$r%=!<%H(B */
  u_p = (LOGPROB *)mymalloc(sizeof(LOGPROB) * winfo.num);
  for(w=0;w<winfo.num;w++) u_p[w] = uni_prob(&ngram, winfo.wton[w]);
  qsort(u_p, winfo.num, sizeof(LOGPROB),
	(int (*)(const void *,const void *))compare_prob);

  /* $B>e$+$i(B n $BHVL\$N%9%3%"$,5a$a$kCM(B */
  x = u_p[n-1];
  free(u_p);
  return(x);
}
#endif
  
/*----------------------------------------------------------------------*/


/* $BLZ9=B$2=<-=q(B(word-conjunction tree hmm)$B$r:n$k(B */
#define COUNT_STEP 500		/* count log period */

static int separated_word_count; /* words num actually separated */
void
build_wchmm()
{
  int i,j;
  int matchword=0, sharelen=0, maxsharelen=0;
  int counter = COUNT_STEP;
#ifdef SEPARATE_BY_UNIGRAM
  LOGPROB separate_thres;
#endif
  
#ifdef SEPARATE_BY_UNIGRAM
  /* $B>e0L(B[separate_wnum]$BHVL\$N(B1-gram$B%9%3%"$r5a$a$k(B */
  /* 1-gram$B%9%3%"$,$3$NCM0J>e$N$b$N$OLZ$+$iJ,$1$k(B */
  separate_thres = get_nbest_uniprob(separate_wnum);
#endif
  /* wchmm$B$r=i4|2=(B */
  wchmm_init(&wchmm);

  /* $BF12;8l%+%&%s%?%j%;%C%H(B */
  homophone_word_num = 0;
  separated_word_count=0;

  j_printerr("Building HMM lexicon tree (left-to-right)...\n");
  for (i=0;i<winfo.num;i++) {
    if (verbose_flag) {
      if (i >= counter) {
	j_printerr("\r %5d words proceeded (%6d nodes)",i,wchmm.n);
	counter += COUNT_STEP;
      }
    }
    if (i == winfo.head_silwid || i == winfo.tail_silwid) {
      /* $B@hF,(B/$BKvHx$NL52;%b%G%k$OLZ9=B$2=$;$:!$(B
       * $B@hF,$NL52;C18l$N@hF,$X$NA+0\!$KvHxC18l$NKvHx$+$i$NA+0\$O:n$i$J$$(B*/
      wchmm_add_word(i,0,0); /* sharelen=0$B$G$=$N$^$^(B */
      continue;
    }
#ifndef NO_SEPARATE_SHORT_WORD
    if (winfo.wlen[i] <= SHORT_WORD_LEN) {
      /* $BD9$5$NC;$$C18l$rLZ9=B$2=$7$J$$(B($B$3$3$G$O(B1$B2;@a(B) */
      wchmm_add_word(i,0,0); /* sharelen=0$B$G$=$N$^$^(B */
      separated_word_count++;
      continue;
    }
#endif
#ifdef SEPARATE_BY_UNIGRAM
    if (uni_prob(&ngram, winfo.wton[i]) >= separate_thres) {
      /* $BIQEY$N9b$$C18l$rLZ9=B$2=$7$J$$(B */
      /* separate_thres $B$O>e0L(Bseparate_wnum$BHVL\$N%9%3%"(B */
      wchmm_add_word(i,0,0);
      separated_word_count++;
      continue;
    }
#endif
    /* $B:G$bD9$/2;AG$r6&M-=PMh$kC18l$rC5$9(B */
    maxsharelen=0;
    for (j=0;j<i;j++) {
      sharelen = wchmm_check_match(i, j);
      if (sharelen == winfo.wlen[i] && sharelen == winfo.wlen[j]) {
	/* word $B$KF12;8l$,B8:_$9$k(B */
	homophone_word_num++;
	/* $BI,$::GBg$ND9$5$G$"$j!$=EJ#%+%&%s%H$rHr$1$k$?$a$3$3$GH4$1$k(B */
	maxsharelen = sharelen;
	matchword = j;
	break;
      }
      if (sharelen > maxsharelen) {
	matchword = j;
	maxsharelen = sharelen;
      }
    }
    wchmm_add_word(i,maxsharelen,matchword);
  }
  j_printerr("\r %5d words ended     (%6d nodes)\n",i,wchmm.n);

#if 0
  /* $BLZ9=B$$r:n$i$J$$(B */
  for (i=0;i<winfo.num;i++) {
    if (verbose_flag) {
      if (i >= counter) {
	j_printerr("  %5d words proceeded (%6d nodes)\n",i,wchmm.n);
	counter += COUNT_STEP;
      }
    }
    wchmm_add_word(i,0,0); /* sharelen=0$B$G$=$N$^$^(B */
  }
  j_printerr("  %5d words ended     (%6d nodes)\n",i,wchmm.n);
#endif  
  
  /* $BF10l2;AG7ONs$r;}$DC18lF1;N$N(B leaf node $B$r(B2$B=E2=$7$F6hJL$9$k(B */
  wchmm_duplicate_leafnode();
  VERMES("  %d leaf nodes are made unshared\n",dupcount);
  
  /* $BC18l$N=*C<$+$i30$X$NA+0\3NN($r5a$a$F$*$/(B */
  wchmm_calc_wordend_arc();

  /* wchmm$B$N@09g@-$r%A%'%C%/$9$k(B */
  check_wchmm();

  /* $BC18l$N@hF,%N!<%I(B ststart $B$NHV9f$r(B $BJL$K3JG<(B */
  wchmm_index_ststart();

  /* factoring$BMQ$K3F>uBV$K8eB3C18l$N%j%9%H$rIU2C$9$k(B */
  make_successor_list();
#ifdef UNIGRAM_FACTORING
  /* $BA0$b$C$F(Bfactoring$BCM$r7W;;(B */
  /* $BKvC<0J30$N(Bsc$B$OI,MW$J$$$N$G%U%j!<$9$k(B */
  calc_all_unigram_factoring_values();
  j_printerr("  1-gram factoring values has been pre-computed\n");
#endif /* UNIGRAM_FACTORING */

  j_printerr("done\n");

  /* $B5/F0;~(B -check $B$G%A%'%C%/%b!<%I$X(B */
  if (wchmm_check_flag) {
    wchmm_check_interactive();
  }
}

/* $B9bB.HG(B: $B%+%F%4%jC10L$OL$<BAu(B */
void
build_wchmm2()
{
  int i,j, last_i;
  int count_step, counter;
  WORD_ID *windex;
#ifdef SEPARATE_BY_UNIGRAM
  LOGPROB separate_thres;
#endif

  separated_word_count = 0;
  count_step = winfo.num / 10;
  counter = count_step;
  j_printerr("Building HMM lexicon tree");
#ifdef SEPARATE_BY_UNIGRAM
  /* $B>e0L(B[separate_wnum]$BHVL\$N(B1-gram$B%9%3%"$r5a$a$k(B */
  /* 1-gram$B%9%3%"$,$3$NCM0J>e$N$b$N$OLZ$+$iJ,$1$k(B */
  separate_thres = get_nbest_uniprob(separate_wnum);
#endif
  
  /* wchmm$B$r=i4|2=(B */
  wchmm_init(&wchmm);
  
  /* $B2;AG$N$J$i$S$GC18l$r%=!<%H$7$?%$%s%G%C%/%9$rF@$k(B */
  windex = (WORD_ID *)mymalloc(sizeof(WORD_ID) * winfo.num);
  for(i=0;i<winfo.num;i++) windex[i] = i;
  /* $BA4BN$G%=!<%H(B */
  wchmm_sort_idx_by_wseq(windex, 0, winfo.num);

  last_i = WORD_INVALID;
  for (j=0;j<winfo.num;j++) {
    i = windex[j];
      if (j >= counter) {
	/*j_printerr("\r %5d words proceeded (%6d nodes)",j, wchmm.n);*/
	j_printerr(".");
	counter += count_step;
      }
    if (i == winfo.head_silwid || i == winfo.tail_silwid) {
      /* $B@hF,(B/$BKvHx$NL52;%b%G%k$OLZ9=B$2=$;$:!$(B
       * $B@hF,$NL52;C18l$N@hF,$X$NA+0\!$KvHxC18l$NKvHx$+$i$NA+0\$O:n$i$J$$(B*/
      wchmm_add_word(i,0,0); /* sharelen=0$B$G$=$N$^$^(B */
      continue;
    }
#ifndef NO_SEPARATE_SHORT_WORD
    if (winfo.wlen[i] <= SHORT_WORD_LEN) {
      /* $BC;$$C18l$rLZ9=B$2=$7$J$$(B */
      wchmm_add_word(i,0,0); /* sharelen=0$B$G$=$N$^$^(B */
      separated_word_count++;
      continue;
    }
#endif
#ifdef SEPARATE_BY_UNIGRAM
    if (uni_prob(&ngram, winfo.wton[i]) >= separate_thres) {
      /* $BIQEY$N9b$$C18l$rLZ9=B$2=$7$J$$(B */
      /* separate_thres $B$O>e0L(Bseparate_wnum$BHVL\$N%9%3%"(B */
      wchmm_add_word(i,0,0);
      separated_word_count++;
      continue;
    }
#endif
    if (last_i == WORD_INVALID) {
      wchmm_add_word(i,0,0);
    } else {
      /* $B:G$bD9$/2;AG$r6&M-=PMh$kC18l!a%=!<%H8e$ND>A0C18l$H$9$k(B */
      wchmm_add_word(i, wchmm_check_match(i, last_i), last_i);
    }
    last_i = i;
  }
  /*j_printerr("\r %5d words ended     (%6d nodes)\n",j,wchmm.n);*/
  free(windex);
  
  /* $BF10l2;AG7ONs$r;}$DC18lF1;N$N(B leaf node $B$r(B2$B=E2=$7$F6hJL$9$k(B */
  j_printerr("%d", wchmm.n);
  wchmm_duplicate_leafnode();
  j_printerr("+%d=%d nodes\n",dupcount, wchmm.n);
  
  /* $BC18l$N=*C<$+$i30$X$NA+0\3NN($r5a$a$F$*$/(B */
  wchmm_calc_wordend_arc();

  /* wchmm$B$N@09g@-$r%A%'%C%/$9$k(B */
  check_wchmm();

  /* $BC18l$N@hF,%N!<%I(B ststart $B$NHV9f$r(B $BJL$K3JG<(B */
  wchmm_index_ststart();

  /* factoring$BMQ$K3F>uBV$K8eB3C18l$N%j%9%H$rIU2C$9$k(B */
  make_successor_list();
#ifdef UNIGRAM_FACTORING
  /* $BA0$b$C$F(Bfactoring$BCM$r7W;;(B */
  /* $BKvC<0J30$N(Bsc$B$OI,MW$J$$$N$G%U%j!<$9$k(B */
  calc_all_unigram_factoring_values();
  j_printerr("  1-gram factoring values has been pre-computed\n");
#endif /* UNIGRAM_FACTORING */

  j_printerr("done\n");

  /* $B5/F0;~(B -check $B$G%A%'%C%/%b!<%I$X(B */
  if (wchmm_check_flag) {
    wchmm_check_interactive();
  }
}


/* wchmm$B$N>pJs$rI8=`=PNO$K=PNO$9$k(B */
void
print_wchmm_info()
{
  int n,i, rootnum;

  rootnum = wchmm.startnum + 1;	/* winfo.head_silwid$B$NJ,(B */
  
  j_printf("Lexicon tree info:\n");
  j_printf("\t total node num = %6d\n", wchmm.n);
  j_printf("\t  root node num = %6d", rootnum);
#ifdef NO_SEPARATE_SHORT_WORD
#ifdef SEPARATE_BY_UNIGRAM
  j_printf(" (%d hi-freq. words are separated from tree)\n", separated_word_count);
#else
  j_printf(" (no words are separated from tree)\n");
#endif /* SEPARATE_BY_UNIGRAM */
#else
  j_printf(" (%d short words (<= %d phonemes) are separated from tree)\n", separated_word_count, SHORT_WORD_LEN);
#endif /* NO_SEPARATE_SHORT_WORD */
  for(n=0,i=0;i<wchmm.n;i++) {
    if (wchmm.stend[i] != WORD_INVALID) n++;
  }
  j_printf("\t  leaf node num = %6d\n", n);
  j_printf("\t     branch num = %6d\n", wchmm.scnum);
}
