/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* defines for word-conjunction HMM */
/* original from ~speech/src/sent/sentence.h */

/* $Id: wchmm.h,v 1.3 1998/09/02 15:47:03 ri Exp $ */

#ifndef __SENT_WORD_CONJ_HMM__
#define __SENT_WORD_CONJ_HMM__

#define		MAXWCNSTEP  40000
enum{HEAD,TAIL};                /* tree direction */

/* typedef of A_CELL is in HMM.h */

typedef struct s_cell {		/* succesor word list */
  WORD_ID word;
  struct s_cell *next;
} S_CELL;

#ifdef PASS1_IWCD
typedef struct {
  HMM_Logical  *hmm;		/* original HMM */
  short		state_loc;	/* state location (1-) */
  HTK_HMM_State *cache;		/* last assigned state */
  WORD_ID	lastwid_cache;	/* wid when above is assigned */
} RC_INFO;
typedef struct {		/* for 1-phoneme LRSET */
  HMM_Logical  *hmm;		/* original HMM */
  short		state_loc;	/* state location (1-) */
  union {
    HTK_HMM_State *state;	/* last assigned state */
    LC_State_Set  *lset;	/* last assigned lset */
  } cache;
  boolean	last_is_lset;
  WORD_ID	lastwid_cache;	/* wid when above is assigned */
} LRC_INFO;
typedef union {
  /* for AS_STATE */
  HTK_HMM_State *state;		/* HMM state for normal node */
  /* for AS_LSET */
  LC_State_Set  *lset;		/* left-context dependent HMM state set */
  /* for AS_RSET */
  RC_INFO	*rset;		/* right-context dependent state info */
  /* for AS_LRSET */
  LRC_INFO	*lrset;		/* right-context dependent state info */
} ACOUSTIC_SPEC;
typedef enum {
  AS_STATE,			/* normal state */
  AS_LSET,			/* left-context-dependent HMM state set */
  AS_RSET,			/* right-context-dependent HMM state set */
  AS_LRSET			/* both (for 1-phoneme word) */
} AS_Style;
#endif
  
typedef struct wchmm_state {
  A_CELL	*ac;
#ifdef PASS1_IWCD
  ACOUSTIC_SPEC out;
  AS_Style	outstyle;
#else  /* ~PASS1_IWCD */
  HTK_HMM_State *out;
#endif /* ~PASS1_IWCD */
  S_CELL  *sc;			/* successor word list for factoring */
#ifdef UNIGRAM_FACTORING
  /* max 1-gram prob of all words in "sc" */
  /* for 1-gram factoring, simply use this value */
  LOGPROB fscore;
#endif
} WCHMM_STATE;

typedef struct wchmm_info {
  int maxwcn;			/* $B8=:_$NNN0h3NJ]?t(B(=$B:GBg?t(B) */
  int		n;		/* $B>uBV?t(B */
  WCHMM_STATE	*state;		/* $B>uBV>pJs(B * n */
  WORD_ID   *ststart;		/* $B>uBV$r3+;O>uBV$H$9$kC18l$N(BID * n */
  WORD_ID   *stend;		/* $B>uBV$r=*N;>uBV$H$9$kC18l$N(BID * n */
  int	**offset;		/* $B3FC18l$N3F2;AG$N3+;O0LCV(B * wnum */
  int	*wordend;		/* $B3FC18l$N:G=*>uBV(B * wnum */
  int	startnum;		/* $BLZ$N:,%N!<%I?t(B */
  int	*startnode;		/* $BLZ$N:,%N!<%I$N%$%s%G%C%/%9(B($B:GBg(Bwnum) */
  LOGPROB *wordend_a;		/* $BC18l$N30$X$NA+0\3NN((B * wnum */
  int   scnum;			/* sc!=NULL$B$N>uBV?t(B */
} WCHMM_INFO;

/* $BBh(B1$B%Q%9$NC18l%H%l%j%99=B$BN(B */
typedef struct __trellis_atom__ {
  LOGPROB backscore;		/* $B%9%3%"9g7W(B */
  LOGPROB lscore;		/* $B$3$NC18l$N8@8l%9%3%"(B */
  WORD_ID wid;			/* $BC18lHV9f(B */
  short begintime;		/* $BC18l$N6-3&;~9o(B($B$3$NC18l$N3+;O;~9o(B)*/
  short endtime;		/* $B=*N;;~4V(B */
#ifdef WORDGRAPH
  boolean within_wordgraph;	/* $BC18l%0%i%UFb$K$"$l$P(BTrue */
#endif
  struct __trellis_atom__ *last_tre; /* $BD>A0$NC18l$N(Bid */
  struct __trellis_atom__ *next; /* 1pass$BCf(B:$B<!$N(BTRE$B$X$N%]%$%s%?(B */
} TRELLIS_ATOM;

/* $B7k2LA4BN(B */
typedef struct __backtrellis__ {
  int framelen;			/* $BAm%U%l!<%`?t(B($B5l(Bpeseqlen$B$KF1$8(B) */
  int *num;			/* $B%U%l!<%`$4$H$N3JG<?t(B */
  TRELLIS_ATOM ***rw;		/* $B%U%l!<%`$4$H$K%S!<%`Fb$K;D$C$?C18l>pJs(B */
  TRELLIS_ATOM *list;		/* $B0l;~E*$K(BTRELLIS$B$r3JG<$7$F$*$/(B */
} BACKTRELLIS;

  
#endif /* __SENT_WORD_CONJ_HMM__ */
