/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* wchmm_check.c --- check wchmm structure  */

/* $Id: wchmm_check.c,v 1.1.1.1 1998/07/16 07:40:25 ri Exp $ */

#include <julius.h>


static void
print_winfo_w(WORD_ID word)
{
  int i;
  if (word >= winfo.num) return;
  j_printf("--winfo\n");
  j_printf("wname   = %s\n",winfo.wname[word]);
  j_printf("woutput = %s\n",winfo.woutput[word]);
  j_printf("wlen    = %d\n",winfo.wlen[word]);
  j_printf("wseq    =");
  for (i=0;i<winfo.wlen[word];i++) {
    j_printf(" %s",winfo.wseq[word][i]->name);
  }
  j_printf("\nwseq_def=");
  for (i=0;i<winfo.wlen[word];i++) {
    j_printf(" %s",winfo.wseq[word][i]->def->name);
  }
  j_printf("\nwton  = %d\n",winfo.wton[word]);
}

static void
print_wchmm_w(WORD_ID word)
{
  int i;
  if (word >= winfo.num) return;
  j_printf("--wchmm (word)\n");
  j_printf("offset  =");
  for (i=0;i<winfo.wlen[word];i++) {
    j_printf(" %d",wchmm.offset[word][i]);
  }
  j_printf("\n");
  j_printf("wordend = %d\n",wchmm.wordend[word]);
}

static void
print_wchmm_s(int node)
{
  j_printf("--wchmm (node)\n");
  j_printf("ststart = %d\n",wchmm.ststart[node]);
  j_printf("stend   = %d\n",wchmm.stend[node]);
#ifdef PASS1_IWCD
  j_printf("outstyle= ");
  switch(wchmm.state[node].outstyle) {
  case AS_STATE:
    j_printf("AS_STATE (id=%d)\n", (wchmm.state[node].out.state)->id);
    break;
  case AS_LSET:
    j_printf("AS_LSET  (%d variants)\n", (wchmm.state[node].out.lset)->num);
    break;
  case AS_RSET:
    j_printf("AS_RSET  (name=\"%s\", defname=\"%s\", loc=%d)\n",
	     (wchmm.state[node].out.rset)->hmm->name,
	     (wchmm.state[node].out.rset)->hmm->def->name,
	     (wchmm.state[node].out.rset)->state_loc);
    break;
  case AS_LRSET:
    j_printf("AS_LRSET (name=\"%s\", defname=\"%s\", loc=%d)\n",
	     (wchmm.state[node].out.lrset)->hmm->name,
	     (wchmm.state[node].out.lrset)->hmm->def->name,
	     (wchmm.state[node].out.lrset)->state_loc);
    break;
  default:
    j_printf("UNKNOWN???\n");
  }
#endif /* PASS1_IWCD */
    
}

static void
print_wchmm_s_arc(int node)
{
  A_CELL *ac;
  int i = 0;
  j_printf("arcs:\n");
  for (ac=wchmm.state[node].ac;ac;ac=ac->next) {
    j_printf(" %d %f(%f)\n",ac->arc,ac->a,pow(10.0, ac->a));
    i++;
  }
  j_printf(" total %d arcs\n",i);
}

static void
print_wchmm_s_successor(int node)
{
  S_CELL *sc;
  int i = 0;
  j_printf("successors:\n");
  for (sc=wchmm.state[node].sc;sc;sc=sc->next) {
    j_printf(" %d\n",sc->word);
    i++;
  }
  j_printf(" total %d successors\n",i);
}

static void
print_hmminfo(char *name)
{
  HMM_Logical *l;

  l = htk_hmmdata_lookup_logical(&hmminfo, name);
  if (l == NULL) {
    j_printf("no HMM named \"%s\"\n", name);
  } else {
    j_printf("HMM logical name: %s\n", l->name);
    j_printf("HMM defined name: %s\n", l->def->name);
    put_htk_hmm(l->def);
  }
}

static void
print_ngraminfo(int nid)
{
  j_printf("-- N-gram entry --\n");
  j_printf("nid  = %d\n", nid);
  j_printf("name = %s\n", ngram.wname[nid]);
}
    


void
wchmm_check_interactive() /* interactive check */
{
  char buf1[24], buf2[24];
  int arg;
  WORD_ID argw;
  boolean endflag;

  j_printf("interactive hmminfo & wchmm check begin\n");

  for (endflag = FALSE; endflag == FALSE;) {
    j_printf("===== check_command arg (\"H H\" for help) > ");
    scanf("%s %s",buf1, buf2);
    if (strlen(buf1)==0) continue;

    arg = atoi(buf2);
    
    switch(buf1[0]) {
    case 'w':			/* word info */
      argw = arg;
      print_winfo_w(argw);
      print_wchmm_w(argw);
      break;
    case 'n':			/* node info */
      print_wchmm_s(arg);
      break;
    case 'a':			/* arc list */
      print_wchmm_s_arc(arg);
      break;
#if 0
    case 'r':			/* reverse arc list */
      print_wchmm_r_arc(arg);
      break;
#endif
    case 's':			/* successor word list */
      print_wchmm_s_successor(arg);
      break;
    case 't':			/* node total info of above */
      print_wchmm_s(arg);
      print_wchmm_s_arc(arg);
#if 0
      print_wchmm_r_arc(arg);
#endif
      print_wchmm_s_successor(arg);
      break;
    case 'h':			/* hmm state info */
      print_hmminfo(buf2);
      break;
    case 'l':			/* N-gram language model info */
      print_ngraminfo(arg);
      break;
    case 'H':			/* help */
      j_printf("syntax: [command_character] [number(#)]\n");
      j_printf("  w [word_id] ... show word info\n");
      j_printf("  n [state]   ... show wchmm state info\n");
      j_printf("  a [state]   ... show arcs from the state\n");
#if 0
      j_printf("  r [state]   ... show arcs  to  the state\n");
#endif
      j_printf("  s [state]   ... show successor list of the state\n");
      j_printf("  h [hmmname] ... show HMM info of the name\n");
      j_printf("  l [nwid]    ... N-gram entry info\n");
      j_printf("  H           ... print this help\n");
      break;
    case 'q':			/* quit */
      endflag = TRUE;
      break;
    }
  }
}

/* wchmm$B$N@09g@-$r%A%'%C%/$9$k(B (for debug) */
void
check_wchmm()
{
  int i,j=0,n;
  boolean ok_flag;
  A_CELL *ac;

  ok_flag = TRUE;
  /* $BC18l(B->$B>uBV(B $B$H(B $B>uBV(B->$BC18l$N@09g$,$H$l$F$$$k$+(B */
  for (i=0;i<winfo.num;i++) {
    if (wchmm.stend[wchmm.wordend[i]]!=i) {
      j_printf("end ga awanai!!!: word=%d, node=%d, value=%d\n",
	       i, wchmm.wordend[i], wchmm.stend[wchmm.wordend[i]]);
      ok_flag = FALSE;
    }
    if (wchmm.ststart[wchmm.offset[i][0]] == WORD_INVALID) {
      j_printf("word start node is WORD_INVALID:word=%d, node=%d, value=%d\n",
	       i, wchmm.offset[i][0], wchmm.ststart[wchmm.offset[i][0]]);
      ok_flag = FALSE;
    }
  }

  /* $B=*C<>uBV$,$+$J$i$:FHN)$+$I$&$+(B (for HEAD)*/
  /* $BFHN)(B := ($BA+0\$,<+J,5Z$SB>$NC18l$N@hF,$X$N$_(B) */
  /* $B$D$$$G$K(Bstend$B$N?t$N%A%'%C%/$b9T$&(B */

  i = 0;
  for (n=0;n<wchmm.n;n++) {
    if (wchmm.stend[n] != WORD_INVALID) {
      i++;
      for (ac=wchmm.state[n].ac; ac; ac=ac->next) {
	if (ac->arc == n) continue;
	if (wchmm.ststart[ac->arc] != WORD_INVALID) continue;
	break;
      }
      if (ac != NULL) {
	j_printf("node %d is shared?\n",n);
	ok_flag = FALSE;
      }
    }
  }
  if (i != winfo.num ) {
    j_printf("num of heads of words in wchmm not match word num!!\n");
    j_printf("from wchmm->stend:%d != from winfo:%d ?\n",j,winfo.num);
    ok_flag = FALSE;
  }

#if 0
  /* trace check */
  /* $B:G=*>uBV$+$iAL$C$F:G=i$N>uBV$^$G!$(B
     $B<+8JA+0\0J30$N7PO)$,0l0U$KDj$^$k$+$I$&$+(B */
  /* skip arc $B$r5v$7$F$$$k$N$G:#$O0UL#$,$J$$(B */
  for (i=0;i<winfo.num;i++) {
    n = wchmm.wordend[i];
    while (n != wchmm.offset[i][0]) { /* $BC18l$N@hF,$KE~C#$9$k$^$G(B */
      for (count=0,ac=rev_state[n].ac;ac;ac=ac->next) {
	if (ac->arc != n) j = ac->arc;
	count++;
      }
      if (count != 2) {		/* $B<+8JA+0\$H?F$X$NA+0\0J30$K$"$k(B */
	j_printf("not 2 arcs for rev_state??: word=%d, node=%d\n", i, n);
	ok_flag = FALSE;
      }
      n = j;
    }
  }
#endif

  /* $B%A%'%C%/$K0z$C3]$+$C$?$i<jF0%A%'%C%/$X(B */
  if (!ok_flag) {
    wchmm_check_interactive();
  }

  VERMES("  coordination check passed\n");
}

