/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/***** revision 3 *****/
/* ngram2.h .. n-gram language model for speech recognition */
/* Third revision: sequencial allocation */
/* for disk-LM and rapid read-in */

#ifndef __SENT_NGRAM2_H__
#define __SENT_NGRAM2_H__

#include <sent/stddefs.h>
#include <sent/ptree.h>


/* MACROs */
#define MAX_N 3			/* trigram$B$^$G$K8GDj(B */
#define UTT_END 3		/* $BJ8$N=*N;(B($B6h@Z$j(B)$B5-9f(B($B!V!#!W(B) */
#define UNK 0			/* $BL$CN8l(B */
#define LOGPROB_INVALID 1.0;	/* N-gram$B8+$D$+$i$J$$>l9g$N%(%i!<CM(B */
				/* logprob = log10(prob) <= 0 */
/*#define FLOOR 0.00003*/ /* flooring$B$N$7$-$$CM(B(1/($B8lWC?t!\2>Dj$9$kL$CN8l?t(B)) */
#define LOGFLOOR -4.522879	/* FLOOR$B$NBP?tCM(B */

#define BIGRAM_UNITSIZE 10	/* unit size of bigram malloc */
#define TRIGRAM_UNITSIZE 3	/* unit size of trigram malloc */

/* data structures */
/* revision 3 has flat structure */

typedef int NNID;
#define NNID_INVALID -1

typedef struct ngram_data {
  WORD_ID max_word_num;		/* N-gram$B$N8lWC?t(B */
  NNID ngram_num[MAX_N];		/* N-gram$B$N(Btuple$B?t(B */
  NNID total_ngram_num[MAX_N];	/* n-gram$B$N=P8=?tAm7W(B */
  WORD_ID unk_id;		/* $B<-=qCf$NL$CN8l$N(BID */
  int unk_num;			/* $B!J>e5-$N?t!K(B */
  LOGPROB unk_num_log;		/* $B!J>e5-$N?t$NBP?tCM!K(B */
  boolean isopen;		/* open vocabulary$B$+$I$&$+(B */

  /* $B4pK\%(%s%H%j(B + 1-gram ( nid: 0 - ngram_num[0]-1 ) */
  char **wname;			/* $BC18lL>J8;zNs(B[nid] */
  PATNODE *root;		/* $BC18lL>8!:wMQ(Btree$B$N:,%N!<%I(B */
  
  LOGPROB *p;			/* 1-gram$B3NN((B[nid] */
  LOGPROB *bo_wt_lr;		/* LR2-gram$BMQ(Bback-off$B78?t(B[nid] */
  LOGPROB *bo_wt_rl;		/* RL2-gram$BMQ(Bback-off$B78?t(B[nid] */
  NNID *n2_bgn;
  WORD_ID *n2_num;		/* [nid] -> n2 bgn--end-1 nid$B$O:8%3%s%F%-%9%H(B */
  
  /* 2-gram (n2: 0 - ngram_num[1] - 1) */
  WORD_ID *n2tonid;		/* [n2] -> nid */
  LOGPROB *p_lr;		/* LR2-gram$B3NN((B[n2] */
  LOGPROB *p_rl;		/* RL2-gram$B3NN((B[n2] */
  LOGPROB *bo_wt_rrl;		/* 3-gram$BMQ(BRL2-gram back-off$B78?t(B[n2] */
  NNID *n3_bgn;
  WORD_ID *n3_num;		/* [n2] -> n3 ($BHO0O(B) n2$B$O1&%3%s%F%-%9%H(B */
  
  /* 3-gram (n3: 0 - ngram_num[2] - 1) */
  WORD_ID *n3tonid;		/* $BC18l(BID$BJQ49(B[n3] = nid */
  LOGPROB *p_rrl;		/* RL3-gram$B3NN((B[n3] */
} NGRAM_INFO;



#define BINGRAM_IDSTR "julius_bingram_v3"
#define BINGRAM_HDSIZE 512
#define BEGIN_WORD_DEFAULT "<s>"
#define END_WORD_DEFAULT "</s>"



/* function declare */
NNID search_bigram(NGRAM_INFO *ndata, WORD_ID w_l, WORD_ID w_r);
NNID search_trigram(NGRAM_INFO *ndata,  NNID n2, WORD_ID wkey);
LOGPROB uni_prob(NGRAM_INFO *ndata, WORD_ID w);
LOGPROB bi_prob_lr(NGRAM_INFO *ndata, WORD_ID w1, WORD_ID w2);
LOGPROB bi_prob_rl(NGRAM_INFO *ndata, WORD_ID w1, WORD_ID w2);
LOGPROB tri_prob_rl(NGRAM_INFO *ndata, WORD_ID w1, WORD_ID w2, WORD_ID w3);

boolean ngram_read_arpa(FILE *fp, NGRAM_INFO *ndata, int direction);
void set_unknown_id(NGRAM_INFO *ndata);
boolean ngram_read_bin(FILE *fp, NGRAM_INFO *ndata);
boolean ngram_write_bin(FILE *fp, NGRAM_INFO *ndata, char *header_str);

void ngram_make_lookup_tree(NGRAM_INFO *ndata);
WORD_ID ngram_lookup_word(NGRAM_INFO *ndata, char *wordstr);
WORD_ID make_ngram_ref(NGRAM_INFO *, char *);

void init_ngram_bin(NGRAM_INFO *ndata, char *ngram_file);
void init_ngram_arpa(NGRAM_INFO *ndata, char *lrfile, char *rlfile);

void print_ngram_info(NGRAM_INFO *ndata);

#include <sent/vocabulary.h>
void make_voca_ref(NGRAM_INFO *ndata, WORD_INFO *winfo);

#endif /* __SENT_NGRAM2_H__ */
