/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* check_hmm_restriction.c --- check if the hmmdef is appropriate for J */

/* $Id: check_hmm_restriction.c,v 1.1.1.1 1998/07/16 07:41:55 ri Exp $ */

/* $B<+J,$HNY$j0J30$NA+0\$,$"$k$+$I$&$+%A%'%C%/(B */
/* $B$H$j$"$($:IT@5$JA+0\$NA+0\3NN($rJ,G[$9$k7A$G(B
   $B=$@5$7$J$,$iFI$_9~$_$rB3$1$k(B */
/* $BA+0\3NN($O$^$@BP?t2=$5$l$F$$$J$$(B */
#include <sent/stddefs.h>
#include <sent/htk_hmm.h>
#include <sent/htk_param.h>
#include <sent/hmm.h>

/* 
 * (HTK origin)
 * $B=i4|>uBV$X$N(Barc$B$OB8:_$7$J$$(B
 * $B:G=*>uBV$+$i$N(Barc$B$OB8:_$7$J$$(B
 */

/* 
 * (J)
 * $B=i4|>uBV$+$i$NA+0\$O0l$D$N$_$H$9$k(B
 * $B:G=*>uBV$X$NA+0\$b0l$D$H$9$k!%(B
 * $BFbIt$G$O(B skip, loop $B2D(B
 */
/* G=good,F=fixed,B=bad */
/*                      */
/*  |G|F|B */
/* -+-+-+- */
/* G|G|F|B */
/* F|F|F|B */
/* B|B|B|B */
#define GOOD 0
#define FIXED 1
#define BAD 3

static int
trans_ok_p(t)
     HTK_HMM_Trans *t;
{
  int i, j;
  int tflag = BAD;
  int retflag = BAD;
  PROB maxprob;
  int maxid = -1;
  
  /* $B=i4|>uBV$+$i$NA+0\$O0l$D$N$_$H$9$k(B */
  tflag = BAD;
  for (i=0;i<t->statenum;i++) {
    if (t->a[0][i] != (PROB)0.0) {
      if (tflag == BAD) {
	tflag = GOOD;
      } else {			/* 2nd time */
	j_printerr("Warning: initial state has more than one arc\n");
	tflag = BAD;
	break;
      }
    }
  }
  if (tflag == BAD) {
    if (i >= t->statenum) {	/* no arc */
      j_printerr("Error: initial state has no arc\n");
    } else {
      /* $B:G$bA+0\3NN($N9b$$(B1$BA+0\$K$^$H$a$k(B */
      maxprob = 0.0; maxid = -1;
      for (j=0;j<t->statenum;j++) {
	if (maxprob < t->a[0][j]) {
	  maxprob = t->a[0][j];
	  maxid = j;
	}
      }
      if (maxid == -1) {
	j_printerr("Error: trans_ok_p: no transition in a state?\n");
	exit(1);
      }
      t->a[0][maxid] = 1.0;
      for (j=0;j<t->statenum;j++) {
	if (j == maxid) continue;
	t->a[0][j] = 0.0;
      }
      tflag = FIXED;
    }
  }

  retflag = tflag;
  
  /* $B:G=*>uBV$X$NA+0\$b0l$D$H$9$k!%(B*/
  tflag = BAD;
  for (i=0;i<t->statenum;i++) {
    if (t->a[i][t->statenum-1] != (PROB)0.0) {
      if (tflag == BAD) {
	tflag = GOOD;
      } else {			/* 2nd time */
	j_printerr("Warning: more than one arc to end state\n");
	tflag = BAD;
	break;
      }
    }
  }
  if (tflag == BAD) {
    if (i >= t->statenum) {	/* no arc */
      j_printerr("Error: no arc to end state\n");
    } else {
      /* $B:G$bA+0\3NN($N9b$$(B1$BA+0\$K$^$H$a$k(B */
      maxprob = (PROB)0.0;
      for (j=0;j<t->statenum;j++) {
	if (maxprob < t->a[j][t->statenum-1]) {
	  maxprob = t->a[j][t->statenum-1];
	  maxid = j;
	}
      }
      for (i=0;i<t->statenum;i++) {
	if (t->a[i][t->statenum-1] == (PROB)0.0) continue;
	if (i == maxid) continue;
	for (j=t->statenum-2;j>=0;j--) {
	  if (t->a[i][j] != (PROB)0.0) {
	    t->a[i][j] += t->a[i][t->statenum-1];
	    t->a[i][t->statenum-1] = (PROB)0.0;
	    break;
	  }
	}
      }
      tflag = FIXED;
    }
  }
    
  return(retflag | tflag);
}

boolean
check_hmm_limit(HTK_HMM_Data *dt)
{
  boolean return_flag = TRUE;
  int tflag;

  tflag = trans_ok_p(dt->tr);
  if (tflag == BAD) {
    return_flag = FALSE;
    j_printerr("Limit: HMM \"%s\" has unsupported arc.\n", dt->name);
    put_htk_trans(dt->tr);
  } else if (tflag == FIXED) {
    j_printerr("Warning: HMM \"%s\" has unsupported arc, fixed.\n", dt->name);
    put_htk_trans(dt->tr);
  }
#if 0
  if (dt->state_num <= 3) {
    return_flag = FALSE;
    j_printerr("Limit: HMM \"%s\" too short (only %d states)\n", dt->name,
	    dt->state_num);
  }
#endif
  return(return_flag);
}

boolean
check_all_hmm_limit(HTK_HMM_INFO *hmminfo)
{
  HTK_HMM_Data *dt;
  boolean return_flag = TRUE;

  for (dt = hmminfo->start; dt; dt = dt->next) {
    if (check_hmm_limit(dt) == FALSE) {
      return_flag = FALSE;
    }
  }
  return(return_flag);
}

