/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* hmm-lookup.c --- HMM name lookup with logical->physical mapping */
/* physical HMMs ... already indexed when reading in hmmdefs */
/* logical HMMs  ... already indexed when reading in HMMList */

/* $Id: hmm_lookup.c,v 1.1.1.1 1998/07/16 07:41:55 ri Exp $ */

#include <sent/stddefs.h>
#include <sent/htk_hmm.h>
#include <sent/ptree.h>

HTK_HMM_Data *
htk_hmmdata_lookup_physical(HTK_HMM_INFO *hmminfo, char *keyname)
{
  HTK_HMM_Data *tmp;
  tmp = aptree_search_data(keyname, hmminfo->physical_root);
  if (strmatch(tmp->name, keyname)) {
    return tmp;
  } else {
    return NULL;
  }
}

HMM_Logical *
htk_hmmdata_lookup_logical(HTK_HMM_INFO *hmminfo, char *keyname)
{
  HMM_Logical *tmp;
  tmp = aptree_search_data(keyname, hmminfo->logical_root);
  if (strmatch(tmp->name, keyname)) {
    return tmp;
  } else {
    return NULL;
  }
}

static void
hmm_count_logical_num(HTK_HMM_INFO *hmminfo)
{
  HMM_Logical *lg;
  int n;

  n = 0;
  for (lg = hmminfo->lgstart; lg; lg = lg->next) n++;
  hmminfo->totallogicalnum = n;
}

/* physical HMM $B$NL>A0$r(B Logical $B$H$7$FDI2CEPO?$9$k(B */
/* $B=E$J$k>l9g$OEPO?$7$J$$(B(HMMList$B$G$N;XDj$,M%@h(B */
void
hmm_add_physical_to_logical(HTK_HMM_INFO *hmminfo)
{
  HMM_Logical *new, *match;
  HTK_HMM_Data *ph;

  for (ph = hmminfo->start; ph; ph = ph->next) {

    /* check logical mapping overriding */
    if (hmminfo->logical_root != NULL) {
      match = aptree_search_data(ph->name, hmminfo->logical_root);
      if (strmatch(match->name, ph->name)) {
	/* the physcal name was already mapped to other HMMs in HMMList */
	j_printerr("Warning: defined HMM \"%s\": mapped to \"%s\" in HMMList, NEVER used\n", ph->name, match->def->name);
	continue;
      }
    }
    /* create new HMM_Logical */
    new = (HMM_Logical *)mybmalloc(sizeof(HMM_Logical));
    new->name = (char *)mybmalloc(strlen(ph->name) + 1);
    strcpy(new->name, ph->name);
    new->def = ph;
    new->next = hmminfo->lgstart;
    hmminfo->lgstart = new;
    if (hmminfo->logical_root == NULL) {
      hmminfo->logical_root = aptree_make_root_node(new);
    } else {
      aptree_add_entry(new->name, new, match->name, &(hmminfo->logical_root));
    }
  }

  /* re-count total number */
  hmm_count_logical_num(hmminfo);
}
