/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* init_phmm.c --- read in hmmdefs file & initialize */

/* $Id: init_phmm.c,v 1.1.1.1 1998/07/16 07:41:59 ri Exp $ */

#include <sent/stddefs.h>
#include <sent/htk_hmm.h>

/* initialize phoneme HMM */
void
init_hmminfo(HTK_HMM_INFO *hmminfo, char *hmmfilename, char *namemapfile)
{
  FILE *fp;

  j_printerr("Reading in HMM definition...");
  /* read hmmdef file */
  if ((fp = fopen_readfile(hmmfilename)) == NULL) {
    j_printerr("failed to open %s\n",hmmfilename);
    exit(1);
  }
  if (rdhmmdef(fp, hmminfo) == FALSE) {
    j_printerr("read error in %s\n",hmmfilename);
    exit(1);
  }
  if (fclose_readfile(fp) < 0) {
    j_printerr("failed to close %s\n", hmmfilename);
    exit(1);
  }

  j_printerr("   defined HMMs: %5d\n", hmminfo->totalhmmnum);

  /* read in logical name -> physical HMM name map file */
  if (namemapfile != NULL) {
    if ((fp = fopen_readfile(namemapfile)) == NULL) {
      j_printerr("failed to open %s\n",namemapfile);
      exit(1);
    }
    if (rdhmmlist(fp, hmminfo) == FALSE) {
      j_printerr("HMMList \"%s\" read error\n",namemapfile);
      exit(1);
    }
    if (fclose_readfile(fp) < 0) {
      j_printerr("failed to close %s\n", namemapfile);
      exit(1);
    }
    j_printerr("  logical names: %5d in HMMList\n", hmminfo->totallogicalnum);
  }


  /* add the names of physical HMMs as logical names (ignore if duplicated) */
  /* also set totallogicalnum */
  hmm_add_physical_to_logical(hmminfo);

  j_printerr("   merged total: %5d\n", hmminfo->totallogicalnum);

  /* Guess we need to handle context dependency */
  /* (word-internal CD is done statically, cross-word CD dynamically */
  if (guess_if_cd_hmm(hmminfo)) {
    hmminfo->is_triphone = TRUE;
  } else {
    hmminfo->is_triphone = FALSE;
  }

  j_printerr("done\n");
}
