/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* lcdset.c --- left-context dependent HMM state set for inter-word CD */

#include <sent/stddefs.h>
#include <sent/htk_param.h>
#include <sent/htk_hmm.h>

#define LC_STATE_SET_STEP 10


static int maxlcnum;
static void
lcdset_init(HTK_HMM_INFO *hmminfo)
{
  hmminfo->lcd_info.lcroot = NULL;
}

static LC_Set *
lcdset_new(HTK_HMM_INFO *hmminfo)
{
  return((LC_Set *)mybmalloc(sizeof(LC_Set)));
}
static void
lcdset_add(HTK_HMM_INFO *hmminfo, LC_Set *new)
{
  new->next = hmminfo->lcd_info.lcroot;
  hmminfo->lcd_info.lcroot = new;
}


/* lookup function */
LC_Set *
lcdset_lookup(HTK_HMM_INFO *hmminfo, char *lcdstr)
{
  LC_Set *lc;
  lc = aptree_search_data(lcdstr, hmminfo->lcd_info.lctree);
  if (strmatch(lcdstr, lc->name)) {
    return lc;
  } else {
    return NULL;
  }
}

/* wrapper for any hmmname */
LC_Set *
lcdset_lookup_by_hmmname(HTK_HMM_INFO *hmminfo, char *hmmname)
{
  char *buf;
  LC_Set *ret;

  buf = strcpy((char *)mymalloc(strlen(hmmname)+1), hmmname);
  ret = lcdset_lookup(hmminfo, leftcenter_name(hmmname, buf));
  free(buf);
  return(ret);
}


/* output given lcdset info to stdout */
static void
put_lcdset(LC_Set *a)
{
  int i;
  printf("name: %s\n", a->name);
  /*printf("state_num: %d\n", a->state_num);*/
  for(i=0;i<a->state_num;i++) {
    if (a->stateset[i].num == 0) {
      printf("\t[state %d]  no\n", i);
    } else {
      printf("\t[state %d]  %d variants\n", i, a->stateset[i].num);
    }
    /* for(j=0;j<a->stateset[i].num;j++) {
       put_htk_state(a->stateset[i].s[j]);
       }*/
  }
}

/* output all lcdset info to stdout */
void
put_all_lcdinfo(HTK_HMM_INFO *hmminfo)
{
  LC_Set *lc;

  for (lc = hmminfo->lcd_info.lcroot; lc; lc = lc->next) {
    put_lcdset(lc);
  }
}


/* construct whole lcdsets from HMM info */
boolean
make_lcdset(HTK_HMM_INFO *hmminfo)
{
  LC_Set *lset, *lmatch;
  HTK_HMM_Data *d;
  HMM_Logical *lg;
  char buf[50], *p;
  int i,n,j;
  LC_State_Set *tmp;
  boolean need_new;

  lcdset_init(hmminfo);

  for(lg = hmminfo->lgstart; lg; lg = lg->next) {
    d = lg->def;
    /*j_printerr("[%s]-[%s]...",lg->name, d->name);*/
    /* make lcdset name from logical HMM name: "a-k+i"->"a-k", "a+o"->"a" */
    /* simply strip off "+.." */
    strcpy(buf, lg->name);
    if ((p = strchr(buf, HMM_RC_DLIM_C)) != NULL) *p = '\0';

    /* check if the lcdset already exist */
    need_new = TRUE;
    if (hmminfo->lcd_info.lctree != NULL) {
      lmatch = aptree_search_data(buf, hmminfo->lcd_info.lctree);
      if (strmatch(lmatch->name, buf)) {
	/* if exist, add to it */
	lset = lmatch;
	need_new = FALSE;
	/* expand state num if needed */
	if (d->state_num > lset->state_num) {
	  lset->stateset = (LC_State_Set *)myrealloc(lset->stateset, sizeof(LC_State_Set) * d->state_num);
	  /* 0 1 ... (lset->state_num-1) */
	  /* N A ... N                   */
	  /* 0 1 ...                     ... (d->state_num-1) */
	  /* N A ... A ..................... N                */
	  /* malloc new area to expanded state (N to A above) */
	  for(j = lset->state_num - 1; j < d->state_num - 1; j++) {
	    lset->stateset[j].maxnum = LC_STATE_SET_STEP;
	    lset->stateset[j].s = (HTK_HMM_State **)mymalloc(sizeof(HTK_HMM_State *) * lset->stateset[j].maxnum);
	    lset->stateset[j].num = 0;
	  }
	  lset->stateset[d->state_num-1].s = NULL;
	  lset->stateset[d->state_num-1].num = 0;
	  lset->stateset[d->state_num-1].maxnum = 0;

	  lset->state_num = d->state_num;
	}
      }
    }

    if (need_new) {
      /* if not exist, allocate as new */
      lset = lcdset_new(hmminfo);
      lset->name = mybstrdup(buf);
      lset->state_num = d->state_num;
      lset->stateset = (LC_State_Set *)mymalloc(sizeof(LC_State_Set) * lset->state_num);
      /* assume first and last state has no outprob */
      lset->stateset[0].s = lset->stateset[lset->state_num-1].s = NULL;
      lset->stateset[0].num = lset->stateset[lset->state_num-1].num = 0;
      lset->stateset[0].maxnum = lset->stateset[lset->state_num-1].maxnum = 0;
      for(j=1;j<lset->state_num-1; j++) {
	lset->stateset[j].maxnum = LC_STATE_SET_STEP;
	lset->stateset[j].s = (HTK_HMM_State **)mymalloc(sizeof(HTK_HMM_State *) * lset->stateset[j].maxnum);
	lset->stateset[j].num = 0;
      }
      lcdset_add(hmminfo, lset);
      /* add to search index tree */
      if (hmminfo->lcd_info.lctree == NULL) {
	hmminfo->lcd_info.lctree = aptree_make_root_node(lset);
      } else {
	aptree_add_entry(lset->name, lset, lmatch->name, &(hmminfo->lcd_info.lctree));
      }
    }
    
    /*j_printerr("add to \"%s\"\n", lset->name);*/
    /* register each HMM states to the lcdset */
    for (j=1;j<d->state_num-1;j++) {
      tmp = &(lset->stateset[j]);
      /* check if the state has already registered */
      for(n = 0; n < tmp->num ; n++) {
	if (tmp->s[n] == d->s[j]) { /* compare by pointer */
	  /*j_printerr("\tstate %d has same\n", n);*/
	  break;
	}
      }
      if (n < tmp->num ) continue;	/* found, no more regist. */

      /* expand storage area if necessary */
      if (tmp->num >= tmp->maxnum) {
	tmp->maxnum += LC_STATE_SET_STEP;
	tmp->s = (HTK_HMM_State **)myrealloc(tmp->s, sizeof(HTK_HMM_State *) * tmp->maxnum);
      }

      tmp->s[tmp->num] = d->s[j];
      tmp->num++;
    }
  }

  return(TRUE);
}
