/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* rdhmmdef_data.c --- read in hmmdefs (data = HMM) */

/* $Id: rdhmmdef_data.c,v 1.1.1.1 1998/07/16 07:41:56 ri Exp $ */

#include <sent/stddefs.h>
#include <sent/htk_hmm.h>

extern char *rdhmmdef_token;

/* malloc new */
HTK_HMM_Data *
htk_hmmdata_new()
{
  HTK_HMM_Data *new;

  new = (HTK_HMM_Data *)mybmalloc(sizeof(HTK_HMM_Data));

  new->name = NULL;
  new->state_num = 0;
  new->s = NULL;
  new->tr = NULL;
  new->next = NULL;

  return(new);
}

static void
htk_hmmdata_add(HTK_HMM_INFO *hmm, HTK_HMM_Data *new)
{
  HTK_HMM_Data *match;
  /* link data structure */
  new->next = hmm->start;
  hmm->start = new;

  if (new->name == NULL) {
    /* HMM must have a name */
    rderr("Error: HMM has no name");
  } else {
    /* add index to search index tree */
    if (hmm->physical_root == NULL) {
      hmm->physical_root = aptree_make_root_node(new);
    } else {
      match = aptree_search_data(new->name, hmm->physical_root);
      if (strmatch(match->name, new->name)) {
	/* HMM of the same name should not be defined */
	j_printerr("Error: HMM \"%s\" is defined more than twice\n", new->name);
	rderr(NULL);
      } else {
	aptree_add_entry(new->name, new, match->name, &(hmm->physical_root));
      }
    }
  }
}

static HTK_HMM_Data *
htk_hmmdata_read(FILE *fp, HTK_HMM_INFO *hmm)
{
  HTK_HMM_Data *new;
  int i;
  short sid;

  new = htk_hmmdata_new();

  /* begin tag */
  if (!currentis("BEGINHMM")) rderr("<BEGINHMM> not found");
  read_token(fp);

  /* read global opt if any */
  /* read_global_opt(fp, &(new->opt)); */

  /* num of state */
  if (!currentis("NUMSTATES")) rderr("<NUMSTATES> not found");
  read_token(fp);
  NoTokErr("state num not found\n");
  new->state_num = atoi(rdhmmdef_token);
  read_token(fp);

  /* malloc */
  new->s = (HTK_HMM_State **)mybmalloc(sizeof(HTK_HMM_State *) * new->state_num);
  for(i=0;i<new->state_num;i++) {
    new->s[i] = NULL;
  }

  /* read/set each state info */
  for (;;) {
    if (!currentis("STATE")) break;
    read_token(fp); NoTokErr("STATE id not found");
    sid = atoi(rdhmmdef_token) - 1;
    read_token(fp);
    new->s[sid] = get_state_data(fp, hmm);
  }

  /* read/set transition info */
  new->tr = get_trans_data(fp, hmm);
  if ((new->tr)->statenum != new->state_num) {
    rderr("# of transition != # of state");
  }

  /* read/set duration */

  /* end tag */
  if (!currentis("ENDHMM")) rderr("<ENDHMM> not found");
  read_token(fp);

  return(new);
}  


/* HMM $BDj5A(B */
void
def_HMM(
	char *name,		/* should not be freed after */
	FILE *fp,
	HTK_HMM_INFO *hmm)
{
  HTK_HMM_Data *new;

  /* fp$B$+$i(B HMM $B%G!<%?$rFI$_9~$_!$?7$?$K(Bmalloc$B$7$?(BHTK_HMM_Data$B$rJV$9(B */
  new = htk_hmmdata_read(fp, hmm);

  /* $BL>A0$r%;%C%H$7!$(BHMM$B9=B$BN$KIU$12C$($k(B */
  new->name = name;
  htk_hmmdata_add(hmm, new);
}
