/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* rdhmmdef_state.c --- read in hmmdefs (state) */

/* $Id: rdhmmdef_state.c,v 1.2 1998/08/31 05:20:47 ri Exp $ */

#include <sent/stddefs.h>
#include <sent/htk_hmm.h>

extern char *rdhmmdef_token;

/* malloc new */
static HTK_HMM_State *
state_new()
{
  HTK_HMM_State *new;

  new = (HTK_HMM_State *)mybmalloc(sizeof(HTK_HMM_State));

  new->name = NULL;
  new->mix_num = 0;
  new->b = NULL;
  new->bweight = NULL;
  new->id = 0;
  new->next = NULL;

  return(new);
}

static void
state_add(HTK_HMM_INFO *hmm, HTK_HMM_State *new)
{
  HTK_HMM_State *match;

  /* link data structure */
  new->next = hmm->ststart;
  hmm->ststart = new;

  if (new->name != NULL) {
    /* add index to search index tree */
    if (hmm->st_root == NULL) {
      hmm->st_root = aptree_make_root_node(new);
    } else {
      match = aptree_search_data(new->name, hmm->st_root);
      if (strmatch(match->name, new->name)) {
	j_printerr("Error: ~s \"%s\" is already defined\n", new->name);
	rderr(NULL);
      } else {
	aptree_add_entry(new->name, new, match->name, &(hmm->st_root));
      }
    }
  }
  
}

static HTK_HMM_State *
state_lookup(HTK_HMM_INFO *hmm, char *keyname)
{
  HTK_HMM_State *s;

  s = aptree_search_data(keyname, hmm->st_root);
  if (strmatch(s->name, keyname)) {
    return s;
  } else {
    return NULL;
  }
}

/* read in state data and return new (malloced) HTK_HMM_State */
static HTK_HMM_State *
state_read(FILE *fp, HTK_HMM_INFO *hmm)
{
  HTK_HMM_State *new;
  int mid;
  int i;

  new = state_new();

  if (currentis("NUMMIXES")) {
    read_token(fp);
    NoTokErr("missing NUMMIXES value");
    new->mix_num = atoi(rdhmmdef_token);
    read_token(fp);
  } else {
    new->mix_num = 1;
  }

  if (currentis("TMIX")) {
    read_token(fp);
    /* read in TMIX */
    tmix_read(fp, new, hmm);
  } else {
    
    new->b = (HTK_HMM_Dens **) mybmalloc(sizeof(HTK_HMM_Dens *) * new->mix_num);
    new->bweight = (PROB *) mybmalloc(sizeof(PROB) * new->mix_num);
    for (i=0;i<new->mix_num;i++) {
      new->b[i] = NULL;
#ifdef LOGTABLE
      new->bweight[i] = LOG_ZERO;
#else
      new->bweight[i] = 0.0;
#endif
    }

    if (new->mix_num == 1) {
      mid = 0;
#ifdef LOGTABLE
      new->bweight[mid] = 0.0;
#else
      new->bweight[mid] = 1.0;
#endif
      new->b[mid] = get_dens_data(fp, hmm);
    } else {
      for (;;) {
	if (!currentis("MIXTURE")) break;
	read_token(fp);
	NoTokErr("missing MIXTURE id");
	mid = atoi(rdhmmdef_token) - 1;
	read_token(fp);
	NoTokErr("missing MIXTURE weight");
#ifdef LOGTABLE
	new->bweight[mid] = log(atof(rdhmmdef_token));
#else
	new->bweight[mid] = atof(rdhmmdef_token);
#endif
	read_token(fp);
	new->b[mid] = get_dens_data(fp, hmm);
      }
    }
  }

#if 0
  /* diabled */
  for (i=0;i<new->mix_num;i++) {
    /* make error if no mixture was associated */
    if (new->b[i] == NULL) {
      j_printerr("Warning: lack of mixture component found, use previous insted...\n");
      new->b[i] = new->b[i-1];
    }
  }
#endif
  
  
  return (new);
}

HTK_HMM_State *
get_state_data(FILE *fp, HTK_HMM_INFO *hmm)
{
  HTK_HMM_State *tmp;

  if (currentis("NUMMIXES")||currentis("MEAN")) {
    /* $B%G!<%?FI$_9~$_(B->$BDI2C(B->$B%]%$%s%?JV$9(B */
    tmp = state_read(fp, hmm);
    tmp->name = NULL; /* $BL>A0$OL5$7(B */
    state_add(hmm, tmp);
    return tmp;
  } else if (currentis("~s")) {
    /* $B%G!<%?8!:w(B->$B%]%$%s%?JV$9(B */
    read_token(fp);
    NoTokErr("missing state macro name");
    tmp = state_lookup(hmm, rdhmmdef_token);
    if (tmp == NULL) {
      j_printerr("Error: ~s \"%s\" not defined\n", rdhmmdef_token);
      rderr(NULL);
    }
    read_token(fp);
    return tmp;
  } else {
    rderr("no state data");
    return NULL;
  }
}



/* state $B%^%/%mDj5A(B */
void
def_state_macro(char *name, FILE *fp, HTK_HMM_INFO *hmm)
{
  HTK_HMM_State *new;

  /* fp$B$+$i(Bstate$B%G!<%?$rFI$_9~$_!$?7$?$K(Bmalloc$B$7$?(BHTK_HMM_State$B$rJV$9(B */
  new = state_read(fp, hmm);

  /* $BL>A0$r%;%C%H$7!$(BHMM$B9=B$BN$KIU$12C$($k(B */
  new->name = name;
  state_add(hmm, new);
}
