/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* rdhmmlist.c --- read in HMMList file and set into hmminfo */
/* no need to define physical names only */
/* NOTE: logical mapping overrides physical definition */

#include <sent/stddefs.h>
#include <sent/htk_hmm.h>
#include <sent/ptree.h>

#define MAXLINEINHMMLIST 256	/* max length per line in HMMList */

boolean
rdhmmlist(FILE *fp, HTK_HMM_INFO *hmminfo)
{
  char *buf, *lname, *pname;
  HMM_Logical *new, *match;
  HTK_HMM_Data *mapped;
  boolean ok_flag = TRUE;
  int n;
  /* 1 column ... ignore (physical names are added after this function) */
  /* 2 column ... define HMM_Logical of the name "$1" which has pointer to $2 */

  buf = (char *)mymalloc(MAXLINEINHMMLIST);
  n = 0;
  while (getl(buf, MAXLINEINHMMLIST, fp) != NULL) {
    n++;
    lname = first_token(buf);
    if ((pname = next_token_if_any()) == NULL) {
      /* only check existence */
      if (htk_hmmdata_lookup_physical(hmminfo, lname) == NULL) {
	j_printerr("Error: HMMList: physical HMM \"%s\" not found\n", lname);
	ok_flag = FALSE;	/* continue reading even on error, exit at last*/
      }
      continue;
    }
    /* check existence */
    if ((mapped = htk_hmmdata_lookup_physical(hmminfo, pname)) == NULL) {
      j_printerr("Error: HMMList: physical HMM \"%s\" not found\n", pname);
      ok_flag = FALSE;	/* continue reading even on error, exit at last*/
      continue;
    }
    /* create new HMM_Logical */
    new = (HMM_Logical *)mybmalloc(sizeof(HMM_Logical));
    new->name = mybstrdup(lname);
    new->def = mapped;
    new->next = hmminfo->lgstart;
    hmminfo->lgstart = new;
    /* add index to search index tree */
    if (hmminfo->logical_root == NULL) {
      hmminfo->logical_root = aptree_make_root_node(new);
    } else {
      match = aptree_search_data(new->name, hmminfo->logical_root);
      if (strmatch(match->name, new->name)) {
	j_printerr("Error: HMMList: logical HMM \"%s\" is defined twice\n", new->name);
	ok_flag = FALSE;
      } else {
	aptree_add_entry(new->name, new, match->name, &(hmminfo->logical_root));
      }
    }
    
  }

  hmminfo->totallogicalnum = n;
  return(ok_flag);
}
