/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* ngram_read_arpa.c --- read in N-gram data in ARPA standard format */

/* $Id: ngram_read_arpa.c,v 1.2 1998/09/02 15:47:22 ri Exp $ */

/* ARPA-standard format $B$N(B2-gram,3-gram$B$^$G$rFI$_9~$a$k(B */
/* $BC18l$NJB$S$O!$(Balphabetically sorted $B$G$J$1$l$P$J$i$J$$(B */
/* RL 3-gram $B$NA0$KI,$:(B LR 2-gram $B$rFI$_9~$s$G$*$+$J$/$F$O$J$i$J$$(B */

#include <sent/stddefs.h>
#include <sent/ngram2.h>

static char buf[800];			/* $BFI$_9~$_MQ%P%C%U%!(B */
static char pbuf[800];			/* $B%(%i!<=PNOMQ%9%H%C%/(B */


/* subfunctions */

static WORD_ID
lookup_word(NGRAM_INFO *ndata, char *str)
{
  WORD_ID wid;
  
  if ((wid = ngram_lookup_word(ndata, str)) == WORD_INVALID) {
    j_printerr("word %s not in N-gram vocabulary.\n",str);
    exit(1);
  }
  return wid;
}


void
set_unknown_id(NGRAM_INFO *ndata)
{
#if 0
  ndata->unk_id = ngram_lookup_word(ndata, unkword);
  if (ndata->unk_id == WORD_INVALID) {
    j_printerr("word %s not found, so assume this is a closed vocabulary model\n",
	    unkword);
    ndata->isopen = FALSE;
  } else {
    ndata->isopen = TRUE;
  }
#endif
  ndata->isopen = TRUE;
  ndata->unk_id = 0;		/* unknown (OOV) words are always mapped to
				   the number 0 (by CMU-TK)*/
}


/* N-gram$B$NA4BN>pJs$rFI$_9~$`(B (LR)*/
/* \data\ */
static void
set_total_info(FILE *fp, NGRAM_INFO *ndata)
{
  char *p;
  int n;

  while (getl(buf, sizeof(buf), fp) != NULL && buf[0] != '\\') {
    if (strnmatch(buf, "ngram", 5)) { /* n-gram num */
      p = strtok(buf, "=");
      n = p[strlen(p)-1] - '0' - 1;
      p = strtok(NULL, "=");
      ndata->ngram_num[n] = atoi(p);
    }
  }
}

/* N-gram$B$NA4BN>pJs$rEPO?:Q$N$b$N$H%A%'%C%/$9$k(B (RL) */
static void
set_and_check_total_info(FILE *fp, NGRAM_INFO *ndata)
{
  char *p;
  int n;

  while (getl(buf, sizeof(buf), fp) != NULL && buf[0] != '\\') {
    if (strnmatch(buf, "ngram", 5)) { /* n-gram num */
      p = strtok(buf, "=");
      n = p[strlen(p)-1] - '0' - 1;
      p = strtok(NULL, "=");
/* 
 *	 if (n <= 2 && ndata->ngram_num[n] != atoi(p)) {
 *	   j_printerr("LR and RL don't match at ngram_num!\n");
 *	   j_printerr("cut-off value when building LM differ?\n");
 *	   exit(1);
 *	 }
 */
      if (n == 2) {		/* 3-gram */
	ndata->ngram_num[n] = atoi(p);
      } else {
	if (n <= 1 && ndata->ngram_num[n] != atoi(p)) {
	  j_printerr("Warning: %d-gram total num differ! may cause read error\n",n+1);
	}
      }
    }
  }
}

/* 1-gram$B$rFI$_9~$`(B (LR)*/
static void
set_unigram(FILE *fp, NGRAM_INFO *ndata)
{
  WORD_ID read_word_num;	/* $B$3$l$^$G$KFI$_9~$s$@(Bunigram$B$NC18l$N?t(B */
  WORD_ID nid, resid;
  LOGPROB prob, bo_wt;
  char *name, *p;

  ndata->wname = (char **)mybmalloc(sizeof(char *)*ndata->ngram_num[0]);
  ndata->p = (LOGPROB *)mybmalloc(sizeof(LOGPROB)*ndata->ngram_num[0]);
  ndata->bo_wt_lr = (LOGPROB *)mybmalloc(sizeof(LOGPROB)*ndata->ngram_num[0]);
  ndata->bo_wt_rl = (LOGPROB *)mybmalloc(sizeof(LOGPROB)*ndata->ngram_num[0]);
  ndata->n2_bgn = (NNID *)mybmalloc(sizeof(NNID)*ndata->ngram_num[0]);
  ndata->n2_num = (WORD_ID *)mybmalloc(sizeof(WORD_ID)*ndata->ngram_num[0]);
  read_word_num = 0;
  
  while (getl(buf, sizeof(buf), fp) != NULL && buf[0] != '\\') {
    prob = atof(first_token(buf));
    p = next_token();
    name = strcpy((char *)mymalloc(strlen(p)+1), p);
    bo_wt = atof(next_token());

  
    /* Unigram $B$rEPO?(B */
    nid = read_word_num;		/* nid $B$r%$%s%/%j%a%s%H(B */
    ndata->wname[nid] = name;
    /* $BL>A0$r8!:wLZ$KEPO?!u=EJ#%A%'%C%/(B */
    if (ndata->root == NULL) {
      ndata->root = ptree_make_root_node(nid);
    } else {
      resid = ptree_search_data(name, ndata->root);
      if (strmatch(name, ndata->wname[resid])) { /* $B=EJ#(B */
	j_printerr("Error: word \"%s\" multiply defined at (#%d and #%d)\n",
		   name, resid, nid);
	exit(1);
      } else {
	ptree_add_entry(name, nid, ndata->wname[resid], &(ndata->root));
      }
    }
    ndata->p[nid] = prob;
    ndata->bo_wt_lr[nid] = bo_wt;
    ndata->n2_bgn[nid] = NNID_INVALID;
    ndata->n2_num[nid] = 0;
  
    read_word_num++;
    if (read_word_num > ndata->max_word_num) {
      j_printerr("Error: actual n-gram word num exceeded header value\n");
      j_printerr("%d > %d\n", read_word_num, ndata->max_word_num);
      exit(1);
    }
  }

  if (read_word_num != ndata->ngram_num[0]) {
    j_printerr("Error: actual n-gram word num not match the header value\n");
    j_printerr("%d != %d ?\n", read_word_num, ndata->ngram_num[0]);
    exit(1);
  }
}

/* 1-gram$B$rDI2C$GFI$_9~$`(B (RL) */

static void
add_unigram(FILE *fp, NGRAM_INFO *ndata)
{
  WORD_ID read_word_num;	/* $B$3$l$^$G$KFI$_9~$s$@(Bunigram$B$NC18l$N?t(B */
  WORD_ID nid;
  LOGPROB prob, bo_wt;
  char *name, *p;

  read_word_num = 0;
  while (getl(buf, sizeof(buf), fp) != NULL && buf[0] != '\\') {
    prob = atof(first_token(buf));
    p = next_token();
    name = strcpy((char *)mymalloc(strlen(p)+1), p);
    bo_wt = atof(next_token());
  
    /* bo_wt_rl $B$r4{B8$N(B unigram $B$KDI2C(B */
    nid = lookup_word(ndata, name);
    if (nid == WORD_INVALID) {
      j_printerr("Warning: n-gram word \"%s\" in RL not exist in LR (ignored)\n", name);
    } else {
      ndata->bo_wt_rl[nid] = bo_wt;
    }
  
    read_word_num++;
    if (read_word_num > ndata->max_word_num) {
      j_printerr("Error: actual n-gram word num exceeded header value\n");
      j_printerr("%d > %d\n", read_word_num, ndata->max_word_num);
      exit(1);
    }
  }

}

/* LR 2-gram$B$rFI$_9~$`(B (LR) */
/* CMU$B$N7A$G%=!<%H$5$l$F$$$k$3$H$,A0Ds$H$J$k(B */

static void
set_bigram(FILE *fp, NGRAM_INFO *ndata)
{
  int w_l, w_r;
  int w_last, w_r_last;
  LOGPROB p;
  int bi_counter = 0;
  NNID n2;

  ndata->n2tonid = (WORD_ID *)mybmalloc(sizeof(WORD_ID)*ndata->ngram_num[1]);
  ndata->p_lr = (LOGPROB *)mybmalloc(sizeof(LOGPROB)*ndata->ngram_num[1]);
  ndata->p_rl = (LOGPROB *)mybmalloc(sizeof(LOGPROB)*ndata->ngram_num[1]);
  ndata->bo_wt_rrl = (LOGPROB *)mybmalloc(sizeof(LOGPROB)*ndata->ngram_num[1]);
  ndata->n3_bgn = (NNID *)mybmalloc(sizeof(NNID)*ndata->ngram_num[1]);
  ndata->n3_num = (WORD_ID *)mybmalloc(sizeof(WORD_ID)*ndata->ngram_num[1]);
  n2 = 0;
    
  /* read in LR 2-gram */
  w_last = -1; w_r_last = -1;
  for (;;) {
    /* $B0l9TFI$_9~$`(B */
    if (getl(buf, sizeof(buf), fp) == NULL || buf[0] == '\\') break;
    strcpy(pbuf, buf);
    if ( ++bi_counter % 100000 == 0) j_printerr("  2-gram read %d\n",bi_counter);

    /* N-gram$B3NN((B */
    p = atof(first_token(buf));
    /* $B:8C18l$N(B id $B$r8!:w(B */
    w_l = lookup_word(ndata, next_token());
    /* $B:G=i!$$b$7$/$O?7$7$$(B wl $B$@$C$?$i!$(Bn2_bgn, n2_num $B$r?75,@_Dj(B */
    if (w_l != w_last) {
      if (w_last != -1) ndata->n2_num[w_last] = n2 - ndata->n2_bgn[w_last];
      /* $B?75,EPO?$G$J$1$l$P%(%i!<(B */
      if (ndata->n2_bgn[w_l] != NNID_INVALID) {
	j_printerr("Error: entry not sorted (same left context not sequenced)\n");
	j_printerr("at 2-gram #%d: \"%s\"\n", n2+1, pbuf);
	exit(1);
      }
      ndata->n2_bgn[w_l] = n2;
      w_r_last = -1;
    }
    /* $B1&C18l$rFI$_9~$s$G3JG<(B */
    w_r = lookup_word(ndata, next_token());
    if (w_r == w_r_last) {
      j_printerr("Error: duplicated entry\n");
      j_printerr("at 2-gram #%d: \"%s\"\n", n2+1, pbuf);
      exit(1);
    } else if (w_r < w_r_last) {
      j_printerr("Error: entry not sorted downward\n");
      j_printerr("at 2-gram #%d: \"%s\"\n", n2+1, pbuf);
      exit(1);
    }
    ndata->n2tonid[n2] = w_r;
    ndata->p_lr[n2] = p;
    ndata->n3_bgn[n2] = NNID_INVALID;
    ndata->n3_num[n2] = 0;

    n2++;
    w_last = w_l;
    w_r_last = w_r;

    /* $B?t$r%A%'%C%/(B */
    if (n2 > ndata->ngram_num[1]) {
      j_printerr("Error: actual 2-gram num not match the header value\n");
      j_printerr("%d != %d ?\n", n2, ndata->ngram_num[1]);
      exit(1);
    }
  }
  
  /* $B:G8e$N(Bn2_num$B$r3JG<(B */
  ndata->n2_num[w_last] = n2 - ndata->n2_bgn[w_last];

}

/* $B4{B8$N(BLR 2-gram $B$K(B RL $B>pJs$r2C$($k(B */
/* rl$B>pJs$N3JG<NN0h$O4{$K(BLR$BFI$_9~$_;~E@$G(Bmalloc$B$5$l$F$$$k(B */
static void
add_bigram_rl(FILE *fp, NGRAM_INFO *ndata)
{
  WORD_ID w_l, w_r;
  LOGPROB prob, bo_wt;
  int bi_counter = 0;
  NNID n2;

  while (getl(buf, sizeof(buf), fp) != NULL && buf[0] != '\\') {
    /* p(w_l|w_r) w_r w_l bo_wt_rl */
    if ( ++bi_counter % 100000 == 0) j_printerr("  2-gram read %d\n",bi_counter);
    prob = atof(first_token(buf));
    w_r = lookup_word(ndata, next_token());
    w_l = lookup_word(ndata, next_token());
    bo_wt = atof(next_token());
    n2 = search_bigram(ndata, w_l, w_r);
    if (n2 == NNID_INVALID) {
      j_printerr("Warning: (%s,%s) not exist in LR 2-gram (ignored)\n",
	      ndata->wname[w_l], ndata->wname[w_r]);
    } else {
      ndata->p_rl[n2] = prob;
      ndata->bo_wt_rrl[n2] = bo_wt;
    }
  }
}
    

/* RL 3-gram$B$rFI$_9~$`(B */

static void
set_trigram(FILE *fp, NGRAM_INFO *ndata)
{
  int w_l, w_m, w_r;
  LOGPROB p_rl;
  int tri_counter = 0;
  int w_r_last, w_m_last, w_l_last;
  NNID n2, n2_last;
  NNID n3;

  ndata->n3tonid = (WORD_ID *)mybmalloc(sizeof(WORD_ID)*ndata->ngram_num[2]);
  ndata->p_rrl = (LOGPROB *)mybmalloc(sizeof(LOGPROB)*ndata->ngram_num[2]);
  n3 = 0;

  n2 = n2_last = NNID_INVALID;
  w_r_last = w_m_last = w_l_last = -1;
  for (;;) {

    /* $B0l9TFI$_9~$`(B */
    if (getl(buf, sizeof(buf), fp) == NULL || buf[0] == '\\') break;
    strcpy(pbuf, buf);
    if (++tri_counter % 100000 == 0) j_printerr("  3-gram read %d\n",tri_counter);

    /* N-gram$B3NN((B */
    p_rl = atof(first_token(buf));
    /* $B1&C18l$N(B id $B$r8!:w(B */
    w_r = lookup_word(ndata, next_token());
    /* $BCfC18l$N(B id $B$r8!:w(B */
    w_m = lookup_word(ndata, next_token());

    /* $B:G=i!$$b$7$/$O?7$7$$(B(w_r,w_m)$B$@$C$?$i!$%3%s%F%-%9%H@Z$jBX$((B */
    if (w_r != w_r_last || w_m != w_m_last) {
      n2 = search_bigram(ndata, w_m, w_r);
      if (n2 == NNID_INVALID) {	/* no context */
        j_printerr("Warning: context (%s,%s) not exist in 2-gram (ignored)\n",
		   ndata->wname[w_m], ndata->wname[w_r]);
        continue;
      }
      /* $BD>A0$N(B n2 $B$N(B n3_num $B$,$3$3$G7hDj(B */
      if (n2_last != NNID_INVALID) ndata->n3_num[n2_last] = n3 - ndata->n3_bgn[n2_last];
      /* $B<!$N(Bn2$B$O?75,EPO?$G$J$1$l$P%(%i!<(B */
      if (ndata->n3_bgn[n2] != NNID_INVALID) {
	j_printerr("Error: entry not sorted (same left context not sequenced)\n");
	j_printerr("at 3-gram #%d: \"%s\"\n", n3+1, pbuf);
	exit(1);
      }
      ndata->n3_bgn[n2] = n3;
      n2_last = n2;
      w_l_last = -1;
    } else {
      /* $BD>A0$HF1$8%3%s%F%-%9%H$G$+$D%9%-%C%W$9$k>l9g(B */
      if (n2 == NNID_INVALID) continue;
    }
    
    /* $B:8C18l$rFI$_9~$s$G=g<!3JG<(B */
    w_l = lookup_word(ndata, next_token());
    if (w_l == w_l_last) {
      j_printerr("Error: duplicated entry\n");
      j_printerr("at 3-gram #%d: \"%s\"\n", n3+1, pbuf);
      exit(1);
    } else if (w_l < w_l_last) {
      j_printerr("Error: entry not sorted downward\n");
      j_printerr("at 3-gram #%d: \"%s\"\n", n3+1, pbuf);
      exit(1);
    }
    ndata->n3tonid[n3] = w_l;
    ndata->p_rrl[n3] = p_rl;

    n3++;
    w_m_last = w_m;
    w_r_last = w_r;
    w_l_last = w_l;

    /* $B?t$r%A%'%C%/(B */
    if (n3 > ndata->ngram_num[2]) {
      j_printerr("Error: actual 3-gram num not match the header value\n");
      j_printerr("%d != %d ?\n", n3, ndata->ngram_num[2]);
      exit(1);
    }
  }

  /* $B:G8e$N(Bn3_num$B$r3JG<(B */
  ndata->n3_num[n2_last] = n3 - ndata->n3_bgn[n2_last];

}


static boolean LR_2gram_read = FALSE;

boolean				/* $BJV$jCM(B:TRUE on success, FALSE on failure */
ngram_read_arpa(FILE *fp, NGRAM_INFO *ndata, int direction)
{
  int n;

  if (!LR_2gram_read && direction == DIR_RL) {
    j_printerr("you should first read LR 2-gram\n");
    return FALSE;
  }


  if (direction == DIR_LR) {
    n = 2;
  } else {
    n = 3;
  }

  /* read until `\data\' found */
  while (getl(buf, sizeof(buf), fp) != NULL && strncmp(buf,"\\data\\",6) != 0);
    
  /* read n-gram total info */
  if (direction == DIR_LR) {
    set_total_info(fp, ndata);
  } else {
    set_and_check_total_info(fp, ndata);
  }
  if (ndata->ngram_num[0] > MAX_WORD_NUM) {
    j_printerr("Error: vocabulary size exceeded limit (%d)\n", MAX_WORD_NUM);
    exit(1);
  }
  ndata->max_word_num = ndata->ngram_num[0];
  
  /* read 1-gram data */
  if (!strnmatch(buf,"\\1-grams",8)) {
    j_printerr("data format error: 1-gram not found\n");
    exit(1);
  }
  j_printerr("  reading 1-gram part...\n");
  if (direction == DIR_LR) {
    set_unigram(fp, ndata);
  } else {
    add_unigram(fp, ndata);
  }
  j_printerr("  1-gram read %d end\n", ndata->ngram_num[0]);
  
  if (n >= 2) {
    /* read 2-gram data */
    if (!strnmatch(buf,"\\2-grams", 8)) {
      j_printerr("data format error: 2-gram not found\n");
      exit(1);
    }
    j_printerr("  reading 2-gram part...\n");
    if (direction == DIR_LR) {
      set_bigram(fp, ndata);
    } else {
      add_bigram_rl(fp, ndata);
    }
    j_printerr("  2-gram read %d end\n", ndata->ngram_num[1]);
  }

  if (n >= 3) {
      /* read 3-gram data */
    if (!strnmatch(buf,"\\3-grams", 8)) {
      j_printerr("data format error: 3-gram not found\n");
      exit(1);
    }
    if ( direction == DIR_LR) {
      j_printerr("should not happen..\n");
      exit(1);
    } else {
      j_printerr("  reading 3-gram part...\n");
      set_trigram(fp, ndata);
    }
    j_printerr("  3-gram read %d end\n", ndata->ngram_num[2]);
  }

  /* finished */
  
  /* set unknown (=OOV) word id */
  set_unknown_id(ndata);

  if (direction == DIR_LR) {
    LR_2gram_read = TRUE;
  }

  return TRUE;
}
