/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* mkwhmm.c --- make word(phrase) HMM from HTK_HMM_INFO */
/* initial & accept arc will be stripped */
/* trans prob to accept state will be stored in accept_ac_a */

/* $Id: mkwhmm.c,v 1.1.1.1 1998/07/16 07:41:59 ri Exp $ */

#include <sent/stddefs.h>
#include <sent/hmm.h>

static int
totalstatelen(HMM_Logical **hdseq, int hdseqlen)
{
  int i, len;

  len = 0;
  for (i=0;i<hdseqlen;i++) {
    len += hdseq[i]->def->state_num - 2;
  }
  return(len);
}

static void
add_arc(HMM_STATE *state, int arc, LOGPROB a)
{
  A_CELL *atmp;

  atmp = (A_CELL *)mymalloc(sizeof(A_CELL));
  atmp->a = a;
  atmp->arc = arc;
  atmp->next = state->ac;
  state->ac = atmp;
}

/* make word(phrase) HMM from HTK_HMM_INFO */
/* new HMM is malloced and returned */
HMM *
new_make_word_hmm_with_lm(HTK_HMM_INFO *hmminfo, HMM_Logical **hdseq, int hdseqlen, LOGPROB *lscore)
{
  HMM *new;
  int i,j,n;
  int afrom, ato;
#ifndef LOGTABLE
  PROB prob;
#endif
  LOGPROB logprob;
  HTK_HMM_Data *dt;

  new = (HMM *)mymalloc(sizeof(HMM));
  new->len = totalstatelen(hdseq, hdseqlen);
  new->state = (HMM_STATE *)mymalloc(sizeof(HMM_STATE) * new->len);
  for (i=0;i<new->len;i++) {
    new->state[i].ac = NULL;
    new->state[i].out = NULL;
  }

  /* assign state outprob info  */
  n = 0;
  for (i = 0; i < hdseqlen; i++) {
    for (j = 1; j < hdseq[i]->def->state_num - 1; j++) {
      new->state[n++].out = hdseq[i]->def->s[j];
    }
  }
  
  /* make transition arcs */
  /* initial state check */
/* 
 *   for (i=0;i<hdseq[0]->def->state_num;i++) {
 * #ifdef LOGTABLE
 *     if (i != 1 && (hdseq[0]->def->tr->a[0][i]) != LOG_ZERO) {
 * #else
 *     if (i != 1 && (hdseq[0]->def->tr->a[0][i]) != 0.0) {
 * #endif
 *	 j_printerr("initial state contains more than 1 arc.\n");
 *     }
 *   }
 */
  new->accept_ac_a = LOG_ZERO;
  n = 0;
  for (i = 0; i < hdseqlen; i++) {
    dt = hdseq[i]->def;
    for (afrom = 1; afrom < dt->state_num - 1; afrom++) {
      for (ato = 1; ato < dt->state_num; ato++) {
#ifdef LOGTABLE
	logprob = dt->tr->a[afrom][ato];
	if (logprob != LOG_ZERO) {
#else
	prob = dt->tr->a[afrom][ato];
	if (prob != 0.0) {
	  logprob = log10(prob);
#endif
	  if (ato == dt->state_num - 1 && lscore != NULL){
	    logprob += lscore[i];
	  }
	  if (n+(ato-afrom) >= new->len) { /* arc to accept node */
	    if (new->accept_ac_a != LOG_ZERO) {
	      j_printerr("more than 1 arc to accept node found.\n");
	      exit(1);
	    } else {
	      new->accept_ac_a = logprob;
	    }
	  } else {
	    add_arc(&(new->state[n]), n + (ato - afrom), logprob);
	  }
	}
      }
      n++;
    }
  }

  return (new);
}

HMM *
new_make_word_hmm(HTK_HMM_INFO *hmminfo, HMM_Logical **hdseq, int hdseqlen)
{
  return(new_make_word_hmm_with_lm(hmminfo, hdseq, hdseqlen, NULL));
}

void
free_hmm(HMM *d)
{
  A_CELL *ac, *atmp;
  int i;

  for (i=0;i<d->len;i++) {
    ac = d->state[i].ac;
    while (ac) {
      atmp = ac->next;
      free(ac);
      ac = atmp;
    }
  }
  free(d->state);
  free(d);
}

