/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* outprob.c --- calculate out probability for given paramter vector */

/* $Id: outprob.c,v 1.2 1998/08/31 05:20:48 ri Exp $ */

#include <sent/stddefs.h>
#include <sent/htk_hmm.h>
#include <sent/htk_param.h>
#include <sent/hmm.h>
#include <sent/gprune.h>

/* cache for each frame */
MIXCACHE ***mixture_cache = NULL; /* [bookid][time][#] (sequencial) */
int mix_n;		/* num of density to store in a frame */
#ifdef USE_ADDLOG_ARRAY
LOGPROB *probs = NULL;
#endif

#undef MAX_PRUNING_WITHIN_STATE	/* enable pruning in tri/mono */

/* $BJ,I[$H%Y%/%H%k$KBP$7=PNO3NN($r7W;;$7$FJV$9(B */
LOGPROB
calc_contprob(HTK_HMM_Dens *binfo, VECT *vec, short veclen)
{
  VECT tmp, x;
  VECT *mean;
  VECT *var;

  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;
  tmp = 0.0;
  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp += x * x / *(var++);
  }
  return((tmp + binfo->gconst) / -2.0);
}

/* $BF~NOCf$N$"$k%U%l!<%`$KBP$9$k=PNO3NN(7W;;(B */
/* p = a + b $B$+$i!$(Blog(p) $B$r(B x = log(a), y = log(b) $B$+$i;;=P(B
 * $BBg$-$$J}$r(B x $B$H$9$k$H!$(B
 * log(p) = x + log(1 + exp(y - x))
 * $B:8JU8e$m$N9`$,=<J,>.$5$$$r9M$($k!%(B
 * exp(LOGZERO) -> 0 $B$h$j!$(Blogscale$B$G$NOB;;$K$*$$$F(B -1/LOG_ZERO$B$rL5;k$G$-$k(B
 * $BBg$-$5$N>e8B$H$_$J$9!%(B
 *
 * $B1&JU$N8e$m$N9`$K$*$$$F!$(B1$B$N6aK5$G$O(B log(1+z)$B!A(Bz $B$J$N$G!$(B
 * exp(y-x) < -1/LOG_ZERO $B$J$iB-$5$:$K:Q$`(B
 *
 */

#ifdef LOGTABLE

#define TBLSIZE 500000
#define VRANGE 15		/* must be > - LOG_ADDMIN */
#define TMAG 33333.3333		/* TBLSIZE/VRANGE */

static LOGPROB tbl[TBLSIZE];	/* [0,1] */

void
make_log_tbl()
{
  LOGPROB f;
  int i;

  j_printerr("Generating addlog table...");
  for (i=0;i<TBLSIZE;i++){
    f = - ((float)VRANGE * (float)i / (float)TBLSIZE);
    tbl[i] = log(1 + exp(f));
    /*if (i < 10 || i > TBLSIZE - 10) j_printf("%f: %d(%f)\n", f, i, tbl[i]);*/
  }
  j_printerr("%d kb...done\n",(TBLSIZE * sizeof(LOGPROB)) / 1024);
}

LOGPROB
addlog(LOGPROB x, LOGPROB y)
{
  /* return(log(exp(x)+exp(y))) */
  LOGPROB tmp;
  unsigned int idx;
  
  if (x < y) {
    if ((tmp = x - y) < LOG_ADDMIN) return y;
    else {
      idx = (unsigned int)((- tmp) * TMAG + 0.5);
      /* j_printf("%f == %f\n",tbl[idx],log(1 + exp(tmp))); */
      return (y + tbl[idx]);
    }
  } else {
    if ((tmp = y - x) < LOG_ADDMIN) return x;
    else {
      idx =(unsigned int)((- tmp) * TMAG + 0.5);
      /* j_printf("%f == %f\n",tbl[idx],log(1 + exp(tmp))); */
      return (x + tbl[idx]);
    }
  }
}

LOGPROB
addlog_array(LOGPROB *a, int n)
{
  LOGPROB tmp;
  LOGPROB x,y;
  unsigned int idx;

  y = LOG_ZERO;
  for(n--; n >= 0; n--) {
    x = a[n];
    if (x > y) {
      tmp = x; x = y; y = tmp;
    }
    /* always y >= x */
    if ((tmp = x - y) < LOG_ADDMIN) continue;
    else {
      idx = (unsigned int)((- tmp) * TMAG + 0.5);
      y += tbl[idx];
    }
  }
  return(y);
}

#else  /* $B%F!<%V%kL5$7(B */

LOGPROB
addlog(LOGPROB x, LOGPROB y)
{
  /* return(log(exp(x)+exp(y))) */
  LOGPROB tmp;

  if (x < y) {
    if ((tmp = x - y) < LOG_ADDMIN) return y;
    else {
/*	 j_printf("aa %f\n", tmp); */
      return (y + log(1 + exp(tmp)));
    }
  } else {
    if ((tmp = y - x) < LOG_ADDMIN) return x;
    else {
/*	 j_printf("aa %f\n", tmp); */
      return( x + log(1 + exp(tmp)));
    }
  }
}

LOGPROB
addlog_array(LOGPROB *x, int n, LOGPROB max)
{
  LOGPROB tmp, y;
  int i;

  y = max;
  for(i=0;i<n;i++) {
    if ((tmp = x[i] - max) < LOG_ADDMIN) continue;
    y += log(1 + exp(tmp));
  }
  return(y);
}


#endif

#ifdef MAX_PRUNING_WITHIN_STATE
static LOGPROB
calc_contprob_with_pruning(HTK_HMM_Dens *binfo, VECT *vec, short veclen, LOGPROB thres)
{
  LOGPROB tmp, x;
  VECT *mean;
  VECT *var;
  LOGPROB fthres = thres * (-2.0);

  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;

  tmp = binfo->gconst;
  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp += x * x / *(var++);
    if ( tmp > fthres) {
      return LOG_ZERO;
    }
  }
  return(tmp / -2.0);
}

LOGPROB
calc_outprob(HTK_HMM_State *stateinfo, VECT *vec, short veclen)
{
  int i, maxi;
  LOGPROB prob;
  LOGPROB maxprob = LOG_ZERO;

  i = maxi = stateinfo->mix_num - 1;
  for (; i >= 0; i--) {
    prob = calc_contprob_with_pruning(stateinfo->b[i], vec, veclen, maxprob);
    if (prob > maxprob) {
      maxprob = prob;
      maxi = i;
    }
  }
  return((maxprob + stateinfo->bweight[maxi]) / 2.30258509); /* power e -> 10 */
}

#else  /* ~MAX_PRUNING_WITHIN_STATE */

LOGPROB
calc_outprob(HTK_HMM_State *stateinfo, VECT *vec, short veclen)
{
  int i = stateinfo->mix_num - 1;
  LOGPROB logprob = LOG_ZERO;
  HTK_HMM_Dens *mixture;

  for (; i >= 0; i--) {
    mixture = stateinfo->b[i];
    if (!mixture) continue;	/* skip mixture == NULL  */
#ifdef LOGTABLE
    logprob = addlog(logprob, calc_contprob(mixture,vec,veclen) + stateinfo->bweight[i] /* $BBP?t2=:Q$_(B(rdhmmdef_state) */);
#else
    logprob = addlog(logprob, calc_contprob(mixture,vec,veclen) + log(stateinfo->bweight[i]));
#endif
  }

  return(logprob / 2.30258509);	/* power e -> 10 */
}

#endif /* MAX_PRUNING_WITHIN_STATE */



/* returns outprob for given HMM state, after mixture calculation was done */
#ifdef USE_ADDLOG_ARRAY

LOGPROB
calc_outprob_tm(int t, HTK_HMM_State *stateinfo)
{
  int i;
  int bid = ((GCODEBOOK *)(stateinfo->b))->id;
  LOGPROB logprob = LOG_ZERO;
  MIXCACHE *mc;
  
  mc = mixture_cache[t][bid];
  for(i=0;i<mix_n;i++) {
#ifdef LOGTABLE
    probs[i] = mc[i].score + stateinfo->bweight[mc[i].id];
#else
    probs[i] = mc[i].score + log(stateinfo->bweight[mc[i].id]);
#endif
  }
  logprob = addlog_array(probs, mix_n);
  if (logprob <= LOG_ZERO) return LOG_ZERO;
  return(logprob / 2.30258509);	/* power e -> 10 */
}

#else  /* ~USE_ADDLOG_ARRAY */

LOGPROB
calc_outprob_tm(int t, HTK_HMM_State *stateinfo)
{
  int i;
  int bid = ((GCODEBOOK *)(stateinfo->b))->id;
  int mixid;
  LOGPROB mixprob;
  LOGPROB logprob = LOG_ZERO;

  for(i=0;i<mix_n;i++) {
    mixid = mixture_cache[t][bid][i].id;
    mixprob = mixture_cache[t][bid][[i].score;
#ifdef LOGTABLE
    logprob = addlog(logprob, mixprob + stateinfo->bweight[mixid] /* $BBP?t2=:Q$_(B(
rdhmmdef_state) */);
#else
    logprob = addlog(logprob, mixprob + log(stateinfo->bweight[mixid]));
#endif
  }
  if (logprob <= LOG_ZERO) return LOG_ZERO;
  return(logprob / 2.30258509); /* power e -> 10 */
}

#endif /* USE_ADDLOG_ARRAY */
