/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* outprob_cache.c */
#include <sent/stddefs.h>
#include <sent/htk_hmm.h>
#include <sent/htk_param.h>
#include <sent/hmm.h>
#include <sent/gprune.h>

#define LOG_UNDEF (LOG_ZERO - 1) /* distinguish from LOG_ZERO score and not calced */

/* $BF~NOCf$N$"$k%U%l!<%`$KBP$9$k=PNO3NN(7W;;(B($B%-%c%C%7%eIU$-(B) */
static LOGPROB **outprob_cache = NULL;
static LOGPROB *ftmp;
static int total_state_num = -1;
/* $BF1$8%U%l!<%`Fb$G$N7W;;$N9bB.2=(B */
static int last_frame = -1;
static LOGPROB *last_cache;
static VECT *last_vec;
static int last_veclen;

/* $B%-%c%C%7%e$N=i4|2=(B */
static void
outprob_cache_init(HTK_HMM_INFO *hmminfo)
{
  total_state_num = hmminfo->totalstatenum;

  outprob_cache = NULL;

  last_frame = -1;
}

/* $B%-%c%C%7%eNN0h$r3NJ](B */
void
outprob_cache_prepare(int framenum)
{
  int i;
  int size;

  if (total_state_num < 0) {
    j_printerr("outprob_cache not initialized before preparation.\n");
    exit(1);
  }

  /* cache$BNN0h$r3NJ](B */
  if (outprob_cache != NULL) {
    free(ftmp);
    free(outprob_cache);
  }
  outprob_cache = (LOGPROB **)mymalloc(sizeof(LOGPROB *) * framenum);
  if ((size = total_state_num * framenum) == 0) return;
  ftmp = (LOGPROB *)mymalloc(sizeof(LOGPROB) * size);
  for(i=0;i<framenum;i++) {
    outprob_cache[i] = &(ftmp[i * total_state_num]);
  }

  /* tell outprob() for new param */
  last_frame = -1;
}

/* $B%-%c%C%7%e$NCf?H$r%/%j%"(B */
void
outprob_cache_clear(int framenum)
{
  int s,t;
  for (t = 0; t < framenum; t++) {
    for (s = 0; s < total_state_num; s++) {
      outprob_cache[t][s] = LOG_UNDEF;
    }
  }
  /* tell outprob() for new param */
  last_frame = -1;
}    



/* $B=i4|2=(B */
void
outprob_init(HTK_HMM_INFO *hmminfo)
{
  outprob_cache_init(hmminfo);
#ifdef LOGTABLE
  make_log_tbl();
#endif
}

/* $B=PNO3NN($rF@$k(B */
LOGPROB
outprob_ss(
     int t,			/* $B%U%l!<%`;XDj(B */
     HTK_HMM_State *stateinfo,	/* $B=PNO(B */
     HTK_Param *param)		/* $BFCD'%Q%i%a!<%?%Y%/%H%k9=B$BN(B */
{
  LOGPROB outp;

  if (last_frame != t) {
    last_frame = t;
    last_cache = outprob_cache[t];
    last_vec = param->parvec[t];
    last_veclen = param->veclen;
  }
  if ((outp = last_cache[stateinfo->id]) == LOG_UNDEF) {
    last_cache[stateinfo->id] = outp = calc_outprob(stateinfo, last_vec, last_veclen);
  }
  return(outp);
}

/* $B=PNO3NN($rF@$k(B (TMix$BMQ(B) */
LOGPROB
outprob_tm(
     int t,			/* $B%U%l!<%`;XDj(B */
     HTK_HMM_State *stateinfo,	/* $B=PNO(B */
     HTK_Param *param)		/* $BFCD'%Q%i%a!<%?%Y%/%H%k9=B$BN(B */
{
  LOGPROB outp;

  if (last_frame != t) {
    last_frame = t;
    last_cache = outprob_cache[t];
    last_vec = param->parvec[t];
    last_veclen = param->veclen;
  }
  if ((outp = last_cache[stateinfo->id]) == LOG_UNDEF) {
    last_cache[stateinfo->id] = outp = calc_outprob_tm(t, stateinfo);
  }
  return(outp);
}
