/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* voca_load_htkdict.c --- read in vocabulary data */

/* $Id: voca_load_htkdict.c,v 1.2 1998/09/02 15:47:24 ri Exp $ */

/* format is HTK Dictionary format */

#include <sent/stddefs.h>
#include <sent/vocabulary.h>
#include <sent/htk_hmm.h>

/* 
 * $B""<-=q$N;EMM(B
 * 
 * $B!&0l9T$4$H$K(B1$B%l%3!<%I(B(1$BC18l(B)$B!%(B
 * 
 * $B%U%#!<%k%I!'(B $BJ8K!%(%s%H%j(B [$B=PNOJ8;zNs(B] $B2;AG(B $B2;AG(B ...
 * 
 * $B!{3F%U%#!<%k%I$N@bL@(B
 *         $BJ8K!%(%s%H%j(B
 *		   (N-gram$B$N>l9g(B)
 *		   $BC18l$N(BN-gram$B$G$N%(%s%H%jL>(B
 *	            ($BF|K\8l(B)[+($B8l44(B)]+($B7ABVAG<oN`(BID) $B$N7A(B
 *                 N-gram$B$N%(%s%H%j$H$NBP1~$r<h$k$N$KMQ$$$k(B
 *
 * 	   [$B=PNOJ8;zNs(B]
 *		   $BG'<17k2L=PNO;~$K=PNO$9$kJ8;zNs!%(B
 * 
 *	   $B2;AGI=5-(B...
 *		   $B2;AG%b%G%k$K=`$8$?2;AGI=5-(B(HTK format)
 *		   $B0l2;AG$:$D(B ' ' $B$G6h@Z$C$FI=5-(B(HTK .dict $B$K=`$8$k(B)
 */

static char trbuf[3][20];
static char chbuf[30];
static char nophone[1];
static int  trp_l, trp, trp_r;

/* return string of triphone name composed from last 3 call */
static char *
cycle_triphone(char *p)
{
  int i;
  
  if (p == NULL) {		/* initialize */
    nophone[0]='\0';
    for(i=0;i<3;i++) trbuf[i][0] = '\0';
    trp_l = 0;
    trp   = 1;
    trp_r = 2;
    return NULL;
  }

  strcpy(trbuf[trp_r],p);

  chbuf[0]='\0';
  if (trbuf[trp_l][0] != '\0') {
    strcat(chbuf,trbuf[trp_l]);
    strcat(chbuf,HMM_LC_DLIM);
  }
  if (trbuf[trp][0] == '\0') {
    i = trp_l;
    trp_l = trp;
    trp = trp_r;
    trp_r = i;
    return NULL;
  }
  strcat(chbuf, trbuf[trp]);
  if (trbuf[trp_r][0] != '\0') {
    strcat(chbuf,HMM_RC_DLIM);
    strcat(chbuf,trbuf[trp_r]);
  }
  i = trp_l;
  trp_l = trp;
  trp = trp_r;
  trp_r = i;

  return(chbuf);
}

/* parse winfo and set maximum state length */
static void
set_maxwn(WORD_INFO *winfo)
{
  int w,p,n;
  int maxwn;

  maxwn = 0;
  for (w=0;w<winfo->num;w++) {
    n = 0;
    for (p=0;p<winfo->wlen[w];p++) {
      n += (winfo->wseq[w][p])->def->state_num - 2;
    }
    if (maxwn < n) maxwn = n;
  }
  winfo->maxwn = maxwn;
}

/* parse winfo and set maximum word length */
static void
set_maxwlen(WORD_INFO *winfo)
{
  WORD_ID w;
  int maxwlen;
  maxwlen = 0;
  for(w=0;w<winfo->num;w++) {
    if (maxwlen < winfo->wlen[w]) maxwlen = winfo->wlen[w];
  }
  winfo->maxwlen = maxwlen;
}


#define PHONEMELEN_STEP  5	/* malloc base */

/* read in vocabulary file */
boolean				/* TRUE on success */
voca_load_htkdict(
     FILE *fp,
     WORD_INFO *winfo,
     HTK_HMM_INFO *hmminfo,	/* if NULL, phonemes are ignored */
     boolean ignore_tri_conv)	/* TRUE if convert to triphone should be ignored */
{
  boolean ok_flag = TRUE;
  WORD_ID vnum;
  char buf[800], tmpbuf[800];
  int j,len,tmpmaxlen;
  HMM_Logical *tmplg;
  HMM_Logical **tmpwseq;
  char *p, *lp = NULL, *ptmp;
  boolean do_conv = FALSE;

  if (hmminfo != NULL && hmminfo->is_triphone && (! ignore_tri_conv))
    do_conv = TRUE;

  winfo_init(winfo);

  vnum = 0;
  while (getl(buf, sizeof(buf), fp) != NULL) {
    /* $BJ8K!%(%s%H%j(B */
    ptmp = first_token(buf);
    winfo->wname[vnum] = strcpy((char *)mymalloc(strlen(ptmp)+1), ptmp);

    /* $B=PNOJ8;zNs(B */
    strcpy(tmpbuf, next_token());
    if (tmpbuf[0] != '[') {
      j_printerr("word output string not specified for %s\n", winfo->wname[vnum]);
      ok_flag = FALSE;
    }
    j = strlen(tmpbuf);
    if (tmpbuf[j-1] == ']') {
      tmpbuf[j-1] = '\0';
    }
    ptmp = &(tmpbuf[1]);
    winfo->woutput[vnum] = strcpy((char *)mymalloc(strlen(ptmp)+1), ptmp);
    
    /* $B2;AGI=5-(B */
    if (hmminfo == NULL) {
      /* $BFI$_9~$^$J$$(B */
      winfo->wseq[vnum] = NULL;
      winfo->wlen[vnum] = 0;
    } else {
      tmpmaxlen = PHONEMELEN_STEP;
      tmpwseq = (HMM_Logical **)mymalloc(sizeof(HMM_Logical *) * tmpmaxlen);
      len = 0;
      
      /* $B;HMQ%b%G%k$,(B triphone $B$N>l9g$O!$(Btriphone$B$r:n$j$J$,$i3JG<(B */
      if (do_conv) {
	cycle_triphone(NULL);
	lp = next_token_if_any();
	if (lp == NULL) {
	  j_printerr("word %s has no phoneme?\n", winfo->wname[vnum]);
	  exit(1);
	}
	cycle_triphone(lp);
      }

      for (;;) {
	if (do_conv) {
/*	  if (lp != NULL) j_printf(" %d%s",len,lp);*/
	  if (lp != NULL) lp = next_token_if_any();
	  if (lp != NULL) p = cycle_triphone(lp);
	  else p = cycle_triphone(nophone);
	} else {
	  p = next_token_if_any();
	}
	if (p == NULL) break;

	if ((tmplg = htk_hmmdata_lookup_logical(hmminfo, p)) == NULL) {
	  /* not found */
	  if (do_conv) {
	    j_printerr("line %d: logical triphone \"%s\" not found in logical / defined HMM\n", vnum+1, p);
	  } else {
	    j_printerr("at line %d: HMM \"%s\" not found\n", vnum+1, p);
	  }
	  ok_flag = FALSE;
	  continue;
	}

	tmpwseq[len] = tmplg;
	len++;
	if (len >= tmpmaxlen) {
	  /* expand wseq area by PHONEMELEN_STEP */
	  tmpmaxlen += PHONEMELEN_STEP;
	  tmpwseq = (HMM_Logical **)myrealloc(tmpwseq, sizeof(HMM_Logical *) * tmpmaxlen);
	}
      }
      if (len == 0) {
	j_printerr("at line %d: no phone specified.\n", vnum+1);
	ok_flag = FALSE;
      }
      winfo->wseq[vnum] = tmpwseq;
      winfo->wlen[vnum] = len;
    }
    
    vnum++;
    if (vnum >= winfo->maxnum) winfo_expand(winfo);
    
  }

  winfo->num = vnum;

  /* maxwn $B?t$rCj=P$7$F3JG<(B */
  set_maxwn(winfo);
  set_maxwlen(winfo);

  return(ok_flag);
}


/* convert monophone dictionary to word-internal triphone */
boolean
voca_mono2tri(WORD_INFO *winfo, HTK_HMM_INFO *hmminfo)
{
  WORD_ID w;
  int ph;
  char *p;
  HMM_Logical *tmplg;
  boolean ok_flag = TRUE;
  
  for (w=0;w<winfo->num;w++) {
    cycle_triphone(NULL);
    cycle_triphone(winfo->wseq[w][0]->name);

    for (ph = 0; ph < winfo->wlen[w] ; ph++) {
      if (ph == winfo->wlen[w] - 1) {
	p = cycle_triphone(nophone);
      } else {
	p = cycle_triphone(winfo->wseq[w][ph + 1]->name);
      }
      if ((tmplg = htk_hmmdata_lookup_logical(hmminfo, p)) == NULL) {
	j_printerr("voca_mono2tri: word \"%s[%s]\"(id=%d): HMM \"%s\" not found\n", winfo->wname[w], winfo->woutput[w], w, p);
	ok_flag = FALSE;
	continue;
      }
      winfo->wseq[w][ph] = tmplg;
    }
  }
  return (ok_flag);
}
	
    
