/* Copyright (c) 1991-2000 Doshita Lab. Speech Group, Kyoto University */
/*   All rights reserved   */

/* wav2mfcc-pipe.c --- split Wav2MFCC to perform per-frame-basis,
   and also realtime CMN for 1st-pass pipe-lining */
/* most codes from "wav2mfcc.c" */

#include <sent/stddefs.h>
#include <sent/mfcc.h>

static double *fbank;                    /* Filterbank */
static FBankInfo fb;

/* preparation */
void
WMP_init(Value para, float **bf)
{
  /* Get filterbank information */
  fb = InitFBank(para);
  
  if((fbank = (double *)mymalloc((para.fbank_num+1)*sizeof(double))) == NULL){
    perror(NULL); exit(1);
  }
  if((*bf = (float *)mymalloc(fb.fftN * sizeof(float))) == NULL){
    perror(NULL); exit(1);
  }
}

/* calc MFCC_E_Z for 1 frame */
void
WMP_calc(float *mfcc, float *bf, Value para)
{
  float energy;

  if (para.raw_e) {
    /* calculate log raw energy */
    energy = CalcLogRawE(bf, para.framesize);
    /* pre-emphasize */
    PreEmphasise(bf, para);
    /* hamming window */
    Hamming(bf, para.framesize);
  } else {
    /* pre-emphasize */
    PreEmphasise(bf, para);
    /* hamming window */
    Hamming(bf, para.framesize);
    /* calculate log energy */
    energy = CalcLogRawE(bf, para.framesize);
  }
  /* filterbank */
  MakeFBank(bf, fbank, fb, para);
  /* MFCC */
  MakeMFCC(fbank, mfcc, para);
  /* weight cepstrum */
  WeightCepstrum(mfcc, para);
  /* Normalise Log Energy is not implemented */
  if(para.enormal) {
    j_printerr("normalize log energy is not implemented in pipeline mode\n");
    exit(1);
  }
  mfcc[para.mfcc_dim] = energy;
}

/* calc MFCC_E_D_Z from MFCC_E_Z for 1 frame */
void WMP_Delta(float **c, int t, int frame, Value para)
{
  int theta, n, dim, B = 0;
  float A1, A2, sum;

  for(theta = 1; theta <= para.delWin; theta++)
    B += theta * theta;

  dim = para.vec_num / 2;

  for(n = 1; n <= dim; n++){
    sum = 0;
    for(theta = 1; theta <= para.delWin; theta++){
      /* Replicate the first or last vector */
      /* at the beginning and end of speech */
      if(t - theta < 0) A1 = c[0][n - 1];
      else A1 = c[t - theta][n - 1];
      if(t + theta >= frame) A2 = c[frame - 1][n - 1];
      else A2 = c[t + theta][n - 1];
      sum += theta * (A2 - A1);
    }
    c[t][n + para.mfcc_dim] = sum / (2 * B);
  }
}


/* real-time CMN */
/*#define LAST_FRAME 100*/
static float *lastcsum = NULL;
static float *mfcc_ave = NULL;
static int cnum = 0;
static int dim;

/* calc mfcc_ave[] and clear lastcsum[] */
void
CMN_realtime_update()
{
  int i;
  if (mfcc_ave == NULL) mfcc_ave = (float *)mymalloc(sizeof(float) * dim);
  for(i=0;i<dim;i++) {
    mfcc_ave[i] = lastcsum[i] / (float)cnum;
    lastcsum[i] = 0.0;
  }
  cnum = 0;
}
 
void
CMN_realtime(float *mfcc, int cdim)
{
  int i;

  if (lastcsum == NULL) {	/* initialize */
    dim = cdim;
    lastcsum = (float *)mymalloc(sizeof(float) * dim);
    for(i=0;i<dim;i++) lastcsum[i] = 0.0;
    cnum = 0;
  }
  cnum++;
  for(i=0;i<dim;i++) {
    lastcsum[i] += mfcc[i];
  }
  if (mfcc_ave != NULL) {
    for(i=0;i<dim;i++) {
      mfcc[i] -= mfcc_ave[i];
    }
  }
  /* you must update mfcc_ave manually from main */
/*  if (cnum >= LAST_FRAME) {
    CMN_realtime_update();
  }*/
}

